﻿#SingleInstance Force
#MaxThreads 255
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 작성자 : DDART
;; 배포장소 : DDART AUTOMATION
;; https://auto.ddart.net
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
global _app := new CAutosenderApplication("sktdata_autosender", "1.27", "SKT 데이터선물 자동발송기", "2023.09.01")
return
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Main program
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
class CAutosenderApplication extends CApplication
{
	InitApplication()
	{
		this.bRunnable := true
		this.lvReceiver := new CListView("수신정보")
		this.lvReceiver.ShowWindow()
		this.LoadReceiver()
		this.sizer.Add(this.lvReceiver)
		this.sizer.Add(this.tabLog)
		this.tabLog.ShowWindow()
	}
	InitSizerSplitter()
	{
		this.sizer.SetSplitterPos(1, (this.setting.spv_pos ? this.setting.spv_pos : this.h_toolbar + Floor(this.h_client / 2)))
		return true
	}
	AddHotKey()
	{
		this.hotkeys["F1"] := this.ID_HELP_HELP
		this.hotkeys["F2"] := this.ID_FILE_MAKEEXE
		this.hotkeys["F5"] := this.ID_OPTION_SETTING
		this.hotkeys["F7"] := this.ID_TOOL_EXECUTE
		this.hotkeys["F8"] := this.ID_TOOL_STOP
		this.hotkeys["^a"] := this.ID_EDIT_ADD
		this.hotkeys["^m"] := this.ID_EDIT_MODIFY
		this.hotkeys["F5"] := this.ID_EDIT_INIT
		this.hotkeys["Delete"] := this.ID_EDIT_DELETE
	}
	AddMenu()
	{
		this.menu.file := new CMenu("File")
		this.ID_FILE_MAKEEXE := this.menu.file.Add("실행파일 만들기(&M)`tF2")
		this.ID_FILE_EXIT := this.menu.file.Add("종료(&X)")
		
		this.menu.option := new CMenu("Option")
		this.ID_OPTION_SETTING := this.menu.option.Add("설정(&S)`tF5")
		this.menu.edit := new CMenu("Edit")
		this.ID_EDIT_ADD := this.menu.edit.Add("추가(&A)`tCtrl+A")
		this.ID_EDIT_MODIFY := this.menu.edit.Add("수정(&M)`tCtrl+M")
		this.ID_EDIT_DELETE := this.menu.edit.Add("삭제(&D)`tDel")
		this.ID_EDIT_UP := this.menu.edit.Add("위로(&U)")
		this.ID_EDIT_DOWN := this.menu.edit.Add("아래로")
		this.ID_EDIT_INIT := this.menu.edit.Add("초기화(&I)`tF5")
		this.menu.tool := new CMenu("Tool")
		this.ID_TOOL_EXECUTE := this.menu.tool.Add("실행(&R)`tF7") 
		this.ID_TOOL_STOP := this.menu.tool.Add("중지(&S)`tF8") 
		this.ID_TOOL_CHROMELAUNCH := this.menu.tool.Add("개발자 모드 크롬(&C)")
		
		this.menu.help := new CMenu("Help")
		this.ID_HELP_HELP := this.menu.help.Add("도움말(&H)`tF1")
		this.ID_HELP_UPDATE := this.menu.help.Add("업데이트 확인(&U)")
		this.menu.bar := new CMenu("MenuBar")
		this.menu.bar.Add("파일(&F)", this.menu.file)
		this.menu.bar.Add("편집(&E)", this.menu.edit)
		this.menu.bar.Add("도구(&T)", this.menu.tool)
		this.menu.bar.Add("옵션(&O)", this.menu.option)
		this.menu.bar.Add("도움말(&H)", this.menu.help)
		Gui, Menu, % this.menu.bar.name
	}
	AddToolbar()
	{
		buttons := Array()
		icons := Array()
		if (A_OSVersion = "WIN_7")
		{
			icons := [152, 151, 140, 16, 25, 156, 107]
		}
		else if (InStr(A_OSVersion, "WIN_8"))
		{
			icons := [123, 138, 127, 12, 18, 143, 94]
		}
		else if (this.GetWin10Version() >= 1803)
		{
			icons := [133, 36, 125, 12, 18, 136, 92]
		}
		else
		{
			icons := [119, 134, 123, 12, 18, 139, 90]
		}
		
		buttons.Push(new CToolbarButton(this.ID_EDIT_INIT, "초기화", icons[1]))
		buttons.Push(new CToolbarButton(this.ID_EDIT_DELETE, "삭 제", icons[2]))
		buttons.Push(new CToolbarButton(this.ID_EDIT_MODIFY, "수 정", icons[3]))
		buttons.Push(new CToolbarButton(this.ID_EDIT_ADD, "추 가", icons[4]))
		buttons.Push(new CToolbarButton(this.ID_EDIT_UP, "위 로", 247, "shell32.dll"))
		buttons.Push(new CToolbarButton(this.ID_EDIT_DOWN, "아래로", 248, "shell32.dll"))
		buttons.Push(new CToolbarButton(this.ID_OPTION_SETTING, "설 정", icons[5]))
		buttons.Push(new CToolbarButton(this.ID_TOOL_STOP, "중 지", icons[6]))
		buttons.Push(new CToolbarButton(this.ID_TOOL_EXECUTE, "실 행", icons[7]))
		iml := new CImageList()
		iml.Create(32, 32, CImageList.ILC_COLOR32 | CImageList.ILC_MASK, buttons.MaxIndex(), 2)
		Loop, % buttons.MaxIndex()
		{
			iml.AddIconFromModule(buttons[A_Index].icofile, buttons[A_Index].icono, true)
		}
		this.toolbar.SetImageList(0, iml.himl)
		Loop, % buttons.MaxIndex()
		{
			this.toolbar.AddButton(A_Index - 1, buttons[A_Index].id, buttons[A_Index].text)
		}
		this.toolbar.SetMaxTextRows(1)
		this.toolbar.AutoSize()
	}
	AppendLogBar(title)
	{
		this.AppendLog("_______________________________________________________________<" title ">______")
	}
	GetChrome(url_base := "")
	{
		return new CChrome(this.setting.chrome_id, this.setting.chrome_debugging_port, url_base)
	}
	LoadSetting()
	{
		this.setting := new CAutosenderSetting()
		base.LoadSetting()
		if (!this.setting.HasKey("chrome_id"))
		{
			this.setting.chrome_id := this.setting.browser > 0 ? this.setting.browser - 1 : 1
		}
	}
	LoadReceiver()
	{
		col := 1
		this.lvReceiver.InsertCol(col++, "100 sort ", "이름")
		this.lvReceiver.InsertCol(col++, "200 sort center", "수신번호")
		this.lvReceiver.InsertCol(col++, "100 sort float right", "선물 용량(GB)")
		this.lvReceiver.InsertCol(col++, "100 float right", "보낸 용량(GB)")
		this.lvReceiver.InsertCol(col++, "100 float right", "남은 용량(GB)")
		Loop, % this.setting.receivers.MaxIndex()
		{
			rv := this.setting.receivers[A_Index]
			this.lvReceiver.Add("", rv.name, rv.phone, rv.total, rv.sent, rv.left)
		}
		this.OnContentChanged()
	}
	Run()
	{
		this.SetTimer(this.ID_TIMER_RUN, "Off")
		
		if (this.lvReceiver.GetCount())
		{
			this.InitLog()
			this.AppendLogBar("실행 시작")
			datasender := new CSTKDataSender()
			if (datasender.Init(this.setting.fetcher))
			{
				this.AppendLog(datasender.win.GetBrowserInfo())
				this.AppendLogBar("자동선물 시작")
				if (datasender.AutoSend(this.lvReceiver))
				{
					this.AppendLogBar("자동선물 완료")
				}
				else
				{
					this.AppendLogBar("자동선물 중단")
					this.AppendLog(datasender.GetError())
				}
			}
			else
			{
				this.AppendLog(datasender.GetError())
			}		
			this.AppendLogBar("실행 종료")
		}
		else
		{
			this.AppendLog("수신자를 추가해주세요")
		}
		this.ResetRunState()
	}
	SaveReceiver()
	{
		this.setting.receivers := []
		Loop, % this.lvReceiver.GetCount()
		{
			row := A_Index
			receiver := {}
			col := 1
			receiver.name := this.lvReceiver.GetItemText(row, col++)
			receiver.phone := this.lvReceiver.GetItemText(row, col++)
			receiver.total := this.lvReceiver.GetItemText(row, col++)
			receiver.sent := this.lvReceiver.GetItemText(row, col++)
			receiver.left := this.lvReceiver.GetItemText(row, col++)
			this.setting.receivers.Push(receiver)
		}
	}
	SaveSetting()
	{
		this.setting.spv_pos := this.sizer.GetSplitterPos(1)
		this.SaveReceiver()
		base.SaveSetting()
	}
	UpdateButtons()
	{
		if (this.lvReceiver.GetCount() = 0)
		{
			this.bRunnable := false
		}
		base.UpdateButtons()
		this.EnableButton(this.menu.edit, this.ID_EDIT_INIT, !this.bRunning && this.bRunnable)
		row := iLast := iFirst := this.lvReceiver.GetNext(0)
		bSelected := false
		if (row)
		{
			bSelected := true
			while(row := this.lvReceiver.GetNext(row))
			{
				if (row != iLast + 1)
				{
					bSelected := false
				}
				iLast := row
			}
		}
		this.EnableButton(this.menu.edit, this.ID_EDIT_DELETE, !this.bRunning && this.bRunnable && iLast)
		this.EnableButton(this.menu.edit, this.ID_EDIT_MODIFY, !this.bRunning && this.bRunnable && this.lvReceiver.GetSelectedCount() = 1)
		this.EnableButton(this.menu.edit, this.ID_EDIT_UP, !this.bRunning && this.bRunnable && bSelected && iFirst > 1)
		this.EnableButton(this.menu.edit, this.ID_EDIT_DOWN, !this.bRunning && this.bRunnable && bSelected && iLast < this.lvReceiver.GetCount())
	}
	OnAdd()
	{
		this.dlgAx := new CReceiverInfoDialog()
		if (this.dlgAx.DoModal())
		{
			this.lvReceiver.Add("", this.dlgAx.name, this.dlgAx.phone, this.dlgAx.amount, 0.0 , Format("{1:.1f}", this.dlgAx.amount))
		}
		this.dlgAx := ""
		this.OnContentChanged()
	}
	OnChromeLaunch()
	{
		chrome := this.GetChrome()
		if (!chrome.Launch())
		{
			this.AppendLog(chrome.korname " 을 실행할수 없습니다")
			this.AppendLog("실행 명령 : " chrome.launch_cmd)
			this.AppendLog("마지막 에러 : " chrome.err)
		}
	}	
	OnCommand(wParam, lParam, nMsg, hWnd)
	{
		id := wParam & 0xFFFF
		if (id =  this.ID_EDIT_ADD)
		{
			this.OnAdd()
		}
		else if (id =  this.ID_EDIT_MODIFY)
		{
			this.OnModify()
		}
		else if (id =  this.ID_EDIT_UP)
		{
			this.OnUp()
		}
		else if (id =  this.ID_EDIT_DOWN)
		{
			this.OnDown()
		}
		else if (id =  this.ID_EDIT_DELETE)
		{
			this.OnDelete()
		}
		else if (id =  this.ID_EDIT_INIT)
		{
			this.OnInit()
		}
		else if (id = this.ID_TOOL_CHROMELAUNCH)
		{
			this.OnChromeLaunch()
		}
		else if (id =  this.ID_OPTION_SETTING)
		{
			this.OnOptionSetting()
		}
		else
		{
			base.OnCommand(wParam, lParam, nMsg, hWnd)
		}
	}
	OnContentChanged()
	{
		if (this.bRunnable && this.lvReceiver.GetCount() = 0)
		{
			this.bRunnable := false
			this.UpdateButtons()
		}
		else if (!this.bRunnable && this.lvReceiver.GetCount() > 0)
		{
			this.bRunnable := true
			this.UpdateButtons()
		}
		else
		{
		}
	}
	OnDelete()
	{
		count := 0
		row := 0
		while(row := this.lvReceiver.GetNext(row))
		{
			this.lvReceiver.Delete(row)
			count++
		}
		if (count > 0)
		{
			this.OnContentChanged()
		}
		else
		{
			this.AppendLog("삭제할 번호를 먼저 선택하세요")
		}
	}
	OnErase()
	{
		this.editLog.Clear()
	}
	OnInit()
	{
		count := this.lvReceiver.GetCount()
		if (count)
		{
			Loop, %count%
			{
				this.lvReceiver.Modify(A_Index, "Col4", 0.0)
				this.lvReceiver.Modify(A_Index, "Col5", Format("{1:.1f}", this.lvReceiver.GetItemText(A_Index, 3)))
			}
			this.AppendLog("보낸 용량, 남은 용량을 초기화했습니다")
		}
	}
	OnLoad()
	{
		base.OnLoad()
		this.PostMessage(this.WM_SETTING_CHANGED, 0, 0)
	}
	OnModify()
	{
		if (this.lvReceiver.GetCount("S") > 1)
		{
			this.AppendLog("하나만 선택하세요")
			return 
		}
		row := this.lvReceiver.GetNext(0)
		if (row > 0)
		{
			col := 1
			name := this.lvReceiver.GetItemText(row, col++)
			phone := this.lvReceiver.GetItemText(row, col++)
			amount := this.lvReceiver.GetItemText(row, col++)
			sent := this.lvReceiver.GetItemText(row, col++)
			this.dlgAx  := new CReceiverInfoDialog(name, phone, amount)
			if (this.dlgAx.DoModal())
			{
				this.lvReceiver.Modify(row, "", this.dlgAx.name, this.dlgAx.phone, this.dlgAx.amount, Format("{1:.1f}", sent), Format("{1:.1f}", this.dlgAx.amount - sent))
			}
			this.dlgAx := ""
		}
		else
		{
			this.AppendLog("수정할 번호를 먼저 선택하세요")
		}
	}
	OnUp()
	{
		iFirst := iLast := row := this.lvReceiver.GetNext(0)
		while(row := this.lvReceiver.GetNext(row))
		{
			iLast := row
		}
		if (iFirst > 1)
		{
			row_delete := iFirst - 1
			rows := Array()
			Loop, 5
			{
				rows.Push( this.lvReceiver.GetItemText(row_delete, A_Index))
			}
			this.lvReceiver.Insert(iLast + 1, "", rows*)
			this.lvReceiver.Delete(row_delete)
		}
	}
	OnDown()
	{
		iFirst := iLast := row := this.lvReceiver.GetNext(0)
		while(row := this.lvReceiver.GetNext(row))
		{
			iLast := row
		}
		if (iFirst > 0 && iLast < this.lvReceiver.GetCount())
		{
			row_delete := iLast + 1
			rows := Array()
			Loop, 5
			{
				rows.Push( this.lvReceiver.GetItemText(row_delete, A_Index))
			}
			this.lvReceiver.Delete(row_delete)
			this.lvReceiver.Insert(iFirst, "", rows*)
		}
	}
	OnNotify(wParam, lParam, nMsg, hWnd)
	{
		code := NumGet(lParam+0, 2 * A_PtrSize, "Int")
		if (code = CListViewCtrl.LVN_ITEMCHANGED)
		{
			this.UpdateButtons()
		}
	}
	
	OnOptionSetting()
	{
		this.dlgAx := new CAutoSenderSettingDialog()
		if (this.dlgAx.DoModal(this.setting))
		{
			this.setting := this.dlgAx.setting
		}
		this.dlgAx := ""
		this.PostMessage(this.WM_SETTING_CHANGED, 0, 0)
	}
	OnSettingChanged(wParam, lParam, nMsg, hWnd)
	{
		this.InitLog()
		this.UpdateButtons()
	}
	OnWebSocketDisconnected(wParam, lParam, nMsg, hWnd)
	{
	}	
}

class CApplication extends CWindow
{
	__New(name, version, korname, release_date := "")
	{
		this.AddErrorHandler()
		SetWorkingDir %A_ScriptDir%
		SetWinDelay, -1
		FileEncoding, UTF-8
		; GUI
		Gui, +Resize     +HWNDhWnd
		base.__New(hWnd)
		this.name := name
		this.version := version
		this.korname := korname
		this.release_date := release_date
		this.titletext := korname " v" version (release_date ? " (" release_date ")" : "")
		this.cmd_exit := ""
		this.m := A_ScreenDPI / 96
		this.toolbar := new CToolbarCtrl()
		this.toolbar_pos := 0
		this.menu := Object()
		this.AddMenu()
		this.AddToolbar()
		this.AddHotkeyHandler()
		this.AddStatusBar()
		; Init
		this.InitIEEmulation()
		this.dlgAx := ""
		this.bRunnable := false
		this.bRunning := false
		this.setting := { width : 800, height : 600}
		; AutoSizer
		this.sizer := new CAutoSizer(true, 2)
	
		this.hCursorNS := DllCall("LoadCursor", "ptr", 0, "ptr", CWU.IDC_SIZENS)
		this.hCursorWE := DllCall("LoadCursor", "ptr", 0, "ptr", CWU.IDC_SIZEWE)
		this.hCursorArrow := DllCall("LoadCursor", "ptr", 0, "ptr", CWU.IDC_ARROW)
		
		; 윈도우메세지핸들러
		this.WM_SETTING_CHANGED := 0x10000
		this.WM_WEBSOCKET_DISCONNECTED := 0x10000 + 1
		this.fn_wm_handler := Object()
		this.AddMessageHandler(CWU.WM_SIZE, "OnSize")
		this.AddMessageHandler(CWU.WM_COMMNOTIFY, "OnCommNotify")
		this.AddMessageHandler(CWU.WM_NOTIFY, "OnNotify")
		this.AddMessageHandler(CWU.WM_KEYDOWN, "OnKeyDown")
		this.AddMessageHandler(CWU.WM_COMMAND, "OnCommand")
		this.AddMessageHandler(CWU.WM_SYSCOMMAND, "OnSysCommand")
		this.AddMessageHandler(CWU.WM_MOUSEMOVE, "OnMouseMove")
		this.AddMessageHandler(CWU.WM_LBUTTONDOWN, "OnLButtonDown")
		this.AddMessageHandler(CWU.WM_LBUTTONUP, "OnLButtonUp")
		this.AddMessageHandler(this.WM_SETTING_CHANGED, "OnSettingChanged")
		this.AddMessageHandler(this.WM_WEBSOCKET_DISCONNECTED, "OnWebSocketDisconnected")
		; 타이머
		this.fn_tm_handler := Array()
		this.ID_TIMER_RUN := this.AddTimer("Run")
		this.ID_TIMER_MSGBOX := this.AddTimer("HandleMsgBox")
		this.OnUpdateCheck(false)
		this.LoadIcon()
		this.LoadSetting()
		this.tabLog := new CTabView()
		this.editLog := new CEditCtrl("로그창", true)
		this.tabLog.Add(this.editLog)
		this.idRun := 0
		this.bRunning := false
		this.InitApplication()
		; 마지막
		this.Show()
		this.OnLoad()
	}
	__Delete()
	{
	}
	AddErrorHandler()
	{
		this.fnOnError := this.OnError.Bind(this)
		OnError(this.fnOnError)
	}
	AddHotkeyHandler()
	{
		this.hotkeys := Object()
		this.AddHotkey()
		fn := this.OnHotKey.Bind(this)
		Hotkey, IfWinActive, % this.title
		for key, val in % this.hotkeys
		{
			HotKey, %key%, %fn%
		}
		this.fnHotKey := fn
	}
	AddMessageHandler(wm, handler)
	{
		if (IsFunc(this[handler]))
		{
			this.fn_wm_handler[wm] := this[handler].Bind(this)
			OnMessage(wm, this.fn_wm_handler[wm])
		}
	}
	AddTimer(handler)
	{
		if (IsFunc(this[handler]))
		{
			return this.fn_tm_handler.Push(this[handler].Bind(this))
		}
		return 0
	}
	AddHotKey()
	{
	}
	AddMenu()
	{
	}
	AddStatusBar()
	{
		this.statusbar := new CStatusBarCtrl()
		this.statusbar.SetParts(-1, -1, -1, -1)
		this.statusbar.SetText(0, this.titletext)
		this.statusbar.SetText(1, A_YYYY "-" A_MM "-" A_DD)
		this.statusbar.SetText(2, "OK")		
	}
	AddToolbar()
	{
	}
	AppendLog(msg)
	{
		FormatTime, TimeString, , HH:mm:ss
		msg := "[" TimeString "] " msg "`r`n"
		this.editLog.AppendText(msg)
		this.editLog.Scroll()
	}
	AppendLogBar(title)
	{
		FormatTime, TimeString, , HH:mm:ss
		this.UpdateSB(title)
		this.UpdateSB(TimeString, 2)
		this.AppendLog("__________________________________________________________________________<" title ">______")
	}
	EnableButton(ByRef menu, id, bEnable := true)
	{
		menu.Enable(id, bEnable)
		this.toolbar.EnableButton(id, bEnable)
	}
	ExitIfStop()
	{
		if (!this.bRunnable)
		{
			this.UpdateSB("사용자에 의해 중단되었습니다")
			this.AppendLog("사용자에 의해 중단되었습니다")
			this.ResetRunState()
			Exit, 1
		}
	}
	GetWin10Version()
	{
		if (!InStr(A_OSVersion, "10."))
		{
			return 0
		}
		RegRead, DisplayVersion , HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion, DisplayVersion
		return DisplayVersion
	}
	InitApplication()
	{
		this.bRunnable := true
		this.InitLog()
	}
	InitIEEmulation()
	{
		if (!InStr(A_OSVersion, "WIN_") || SubStr(A_OSVersion, 5, 1) >= 7)
		{
            be := {"Default" : 0, "IE7" : 7000, "IE8" : 8000, "IE8STD" : 8888, "IE9" : 9000, "IE9STD" : 9999, "IE10" :10000, "IE10STD" : 10001, "IE11" : 11000, "IE11Edge" : 11001}
			emulation := be.IE11Edge
			RegRead, emul_current, HKEY_CURRENT_USER\Software\Microsoft\Internet Explorer\Main\FeatureControl\FEATURE_BROWSER_EMULATION, AutoHotKey.exe
			if (emul_current != emulation)
			{
				RegWrite, REG_DWORD, HKEY_CURRENT_USER\Software\Microsoft\Internet Explorer\Main\FeatureControl\FEATURE_BROWSER_EMULATION, AutoHotKey.exe, %emulation%
			}
		}
	}
	InitLog(iSel := 0)
	{
		this.tabLog.SetCurSel(iSel)
		if (!this.tabLog.IsWindowVisible())
		{
			this.tabLog.ShowWindow()
			this.UpdateMenus()
			this.Size()
		}
	}
	InitSizerSplitter()
	{
		return true
	}
	LoadIcon()
	{
		ico := this.name ".ico"
		if (InStr(A_ScriptName, ".exe") || !FileExist(ico))
		{
			return
		}
		hIcon := LoadPicture(ico, "Icon1 w32 h32", type)
		this.SendMessage(CWU.WM_SETICON, 0, hIcon)
	}
	LoadSetting()
	{
		this.setting_file := this.name ".setting.json"
		if (FileExist(this.setting_file))
		{
			FileRead, str, % this.setting_file
			setting := JSON.Load(str)
			for key, val in % setting
			{
				if (!IsObject(val) && RegExMatch("^\d+$", val))
				{
					val := val + 0
				}
				this.setting[key] := val
			}
			if (!this.setting.width || !this.setting.height)
			{
				this.setting.width := 900
				this.setting.height := 1000
			}
		}
	}
	ReleaseNote()
	{
		RegRead, ProductName , HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion, ProductName
		RegRead, CurrentBuildNumber , HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion, CurrentBuildNumber
		if (CurrentBuildNumber >= 22000)
		{
			ProductName := StrReplace(ProductName, "Windows 10", "Windows 11")
		}
		win10ver := this.GetWin10Version()
		if (win10ver > 0)
		{
			os := ProductName " 버전 " win10ver " (" A_OSVersion ")"
		}
		else
		{
			os := ProductName
		}
		this.AppendLog("운영체제 : " os ", 오토핫키 : " A_AhkVersion " " (A_IsUnicode ? "Unicode" : "Ansi") " " (A_PtrSize = 8 ? "64" : "32") "-bit" ", " this.korname " : " this.version)
		if (A_Args.Length() > 1)
		{
			if (InStr(A_Args[1], "ahk_pid"))
			{
				if (this.ID_HELP_RELEASENOTE)
				{
					this.OnReleaseNote()
				}
				if (InStr(A_Args[5], ".exe"))
				{
					;; 업데이트 전에 exe 파일로 이용중이었음
					this.OnMakeExe()
				}
			}
		}
	}
	RemoveErrorHandler()
	{
		OnError(this.fnOnError, 0)
		this.fnOnError := ""
	}
	RemoveHotkeyHandler()
	{
		for key in % this.hotkeys
		{
			HotKey, %key%, _RemoveHotkey
		}
		this.fnHotKey := ""
	}
	RemoveMessageHandlers()
	{
		for wm, fn in this.fn_wm_handler
		{
			OnMessage(wm, fn, 0)
			this.fn_wm_handler[wm] := ""
		}
	}
	RemoveTimers()
	{
		Loop, % this.fn_tm_handler.MaxIndex()
		{
			this.SetTimer(A_Index, "Delete")
			this.fn_tm_handler[A_Index] := ""
		}
	}
	ResetRunState()
	{
		this.bRunning := false
		this.bRunnable := true
		this.UpdateButtons()
	}
	SetTimer(name, option)
	{
		fn := this.fn_tm_handler[name]
		if (fn)
		{
			SetTimer, %fn%, %option%
		}
	}
	Size()
	{
		if (!this.w || !this.h)
		{
			this.w := this.setting.width
			this.h := this.setting.height
		}
		else
		{
			this.setting.width := this.w
			this.setting.height := this.h
		}
		this.toolbar.Size()
		pos := this.toolbar.GetPos()
		this.h_toolbar := pos.h
		spw := Floor(120 * this.m)
		this.statusbar.SetParts(this.w - 3 * spw, this.w - 2 * spw, this.w - spw, -1)
		this.statusbar.Size()
		pos := this.statusbar.GetPos()
		this.h_statusbar := pos.h
		this.h_client := this.h - this.h_toolbar - this.h_statusbar
		if (this.toolbar_pos = 0)
		{
			this.sizer.SetRect(new CRect(0, this.h_toolbar, this.w, this.h - this.h_statusbar))
		}
		else if (this.toolbar_pos = 1)
		{
			this.sizer.SetRect(new CRect(0, 0, this.w, this.h_client))
			this.toolbar.Move(0, this.h_client, this.w, this.h_toolbar)
		}
		if (!this.sizer.bInitSplitter)
		{
			this.sizer.bInitSplitter := this.InitSizerSplitter()
		}
		this.sizer.Size()
	}
	Run()
	{
		this.SetTimer(this.ID_TIMER_RUN, "Off")
		this.ResetRunState()
	}
	SaveSetting()
	{
		if (FileExist(this.setting_file))
		{
			FileDelete, % this.setting_file
		}
		rc := this.GetWindowRect()
		this.setting.left := rc.left
		this.setting.top := rc.top
		rc := this.GetClientRect()
		this.setting.width := rc.right - rc.left
		this.setting.height := rc.bottom - rc.top
		FileAppend, % JSON.Dump(this.setting), % this.setting_file, UTF-8
	}
	Show()
	{
		if (!this.setting.width || !this.setting.height)
		{
			this.setting.width := 900
			this.setting.height := 1000
		}
		this.w := this.setting.width
		this.h := this.setting.height
		pos := "w" Floor(this.w / this.m) " h" Floor(this.h / this.m)
		if (this.setting.HasKey("left"))
		{
			pos .= " x" Floor(this.setting.left / this.m)
		}
		if (this.setting.HasKey("top"))
		{
			pos .= " y" Floor(this.setting.top / this.m)
		}
		Gui, 1:Show, %pos%, % this.titletext
	}
	UpdateApplication(version, sha256)
	{
		this.InitLog()
		backup_dir := "backup_versions"
		if (!FileExist(backup_dir))
		{
			FileCreateDir, %backup_dir%
		}
		if (RegExMatch(A_ScriptName, "O)(.+)\.(\w{3})", matches))
		{
			name := this.name
			
			dst := A_ScriptDir
			ahk_src := A_ScriptDir "\" name ".ahk"
			exe_src := A_ScriptDir "\" name ".exe"
			ahk_dst := A_ScriptDir "\" backup_dir "\" name "_v" this.version ".ahk"
			exe_dst := A_ScriptDir "\" backup_dir "\" name "_v" this.version ".exe"
			if (matches[1] = name)
			{
				FileMove, %ahk_src%, %ahk_dst%, true
				FileMove, %exe_src%, %exe_dst%, true
			}
			zip := A_ScriptDir "\" name ".zip"
			try
			{
				UrlDownloadToFile, https://auto.ddart.net/download/%name%/%version%, %zip%
			}
			catch e
			{
				this.AppendLog(e.Message)
			}
			if (FileExist(zip))
			{
				ret := CmdRun("certutil -hashfile " zip " sha256")
				if (RegExMatch(ret, "Oi)([A-F0-9]{64})", matches))
				{
					download_sha256 := matches[1]
					StringLower, download_sha256, download_sha256

					if (download_sha256 != sha256)
					{
						MsgBox, 다운로드된 파일의 해쉬(CheckSum)값이 원본버전과 맞지 않습니다.
						FileDelete, %zip%
						return
					}
				}
				this.AppendLog(zip)
				obj := ComObjCreate("Shell.Application")
				obj.Namespace(dst).CopyHere(obj.Namespace(zip).items, 4|16)
				ObjRelease(obj)
				FileDelete, %zip%
				if (InStr(A_ScriptName, ".exe"))
				{
					this.OnMakeExe(false)
					this.cmd_exit := exe_src " ahk_pid " this.GetPID() " version " this.version " " A_ScriptName
				}
				else
				{
					this.cmd_exit := ahk_src " ahk_pid " this.GetPID() " version " this.version " " A_ScriptName
				}
				
				MsgBox, 4, 업데이트 확인, 업데이트 완료되었습니다. 업데이트반영을 위해 지금 재실행하시겠습니까? , 10
				IfMsgBox, Yes
				{
					this.OnFileExit()
				}			
			}
			else
			{
				this.AppendLog("업데이트 다운로드 실패")
			}
		}
	}
	UpdateButtons()
	{
		this.EnableButton(this.menu.tool, this.ID_TOOL_EXECUTE, !this.bRunning && this.bRunnable)
		this.EnableButton(this.menu.tool, this.ID_TOOL_STOP, this.bRunning && this.bRunnable)
	}
	UpdateMenus()
	{
	}
	UpdateSB(msg, iPart := 0)
	{
		this.statusbar.SetText(iPart, msg)
	}
	OnClose()
	{
		this.SaveSetting()
		return false
	}
	OnCommand(wParam, lParam, nMsg, hWnd)
	{
		id := wParam & 0xFFFF
		notification_code := wParam >> 16
		if (id = this.ID_FILE_EXIT)
		{
			this.OnFileExit()
		}
		else if (id = this.ID_TOOL_EXECUTE)
		{
			this.OnExecute()
		}
		else if (id = this.ID_TOOL_STOP)
		{
			this.OnStop()
		}
		else if (id = this.ID_HELP_HELP)
		{
			this.OnHelp()
		}
		else if (id = this.ID_HELP_RELEASENOTE)
		{
			this.OnReleaseNote()
		}
		else if (id = this.ID_HELP_UPDATE)
		{
			this.OnUpdateCheck()
		}
		else if (id = this.ID_FILE_MAKEEXE)
		{
			this.OnMakeExe()
		}
		else if (id = this.editLog.id)
		{
		}
		else
		{
		}
	}
	HandleMsgBox()
	{
		this.SetTimer(this.ID_TIMER_MSGBOX, "Off")
		pos := this.GetPos()
		wnd := new CWindow(this.WaitPopupWindow(500))
		if (wnd.Exist())
		{
			mbpos := wnd.GetPos()
			wnd.Move(pos.x + pos.w / 2 - mbpos.w/ 2, pos.y + pos.h / 2 - mbpos.h / 2)
		}
	}
	OnCommNotify(wParam, lParam, nMsg, hWnd)
	{
		if (wParam = 1027)
		{
			this.SetTimer(this.ID_TIMER_MSGBOX, -100)
		}
	}
	OnError(error)
	{
		msg := "오류발생 : " error.File " " error.Line "번째줄`r`n"
		msg .= error.Extra " : " error.Message
		this.AppendLog(msg)
		if (this.IsAlwaysOnTop())
		{
			this.AlwaysOnTop(false)
			MsgBox, %msg%
			this.AlwaysOnTop(true)
		}
		else
		{
			MsgBox, %msg%
		}
		this.ResetRunState()
		return true
	}
	OnExecute(id := 0)
	{
		if (this.bRunnable && !this.bRunning)
		{
			this.idRun := id
			this.bRunning := true
			this.UpdateButtons()
			this.SetTimer(this.ID_TIMER_RUN, -10)
		}		
	}
	OnFileExit()
	{
		this.PostMessage(CWU.WM_SYSCOMMAND, 0xF060, 0)
	}
	OnHelp()
	{
		Run, https://auto.ddart.net/xe/help
	}
	OnHotKey()
	{
		cmd := this.hotkeys[A_ThisHotkey]
		if (cmd)
		{
			this.PostMessage(CWU.WM_COMMAND, cmd, 0)
		}
	}
	OnKeyDown(wParam, lParam, nMsg, hWnd)
	{
		if (this.dlgAx && hWnd = this.dlgAx.hWndIE && this.dlgAx.iaxie.pi)
		{
			; AxIE에 일부 키 전달안되는 문제. TAB=9, ENTER=13, DELETE=46
			if (wParam = 9 || wParam = 13 || wParam = 46)
			{
				VarSetCapacity(msg, 7*A_PtrSize)
				for i,val in [hWnd, nMsg, wParam, lParam, A_EventInfo, A_GuiX, A_GuiY]
					NumPut(val, msg, (i-1)*A_PtrSize)
				this.dlgAx.iaxie.TranslateAccelerator(&msg)
				; 메세지 루프탈출
				return 0 
			}
		}
	}
	OnLButtonDown(wParam, lParam, nMsg, hWnd)
	{
		if (hWnd != this.hWnd)
		{
			return
		}
		x := (lParam & 0xFFFF) << 48 >> 48
		y := lParam >> 16
		this.cur_sizer := this.sizer.HitTest(x, y)
		if (this.cur_sizer)
		{
			this.SetCapture()
			this.nSplitterPos := -1
		}
	}
	OnLButtonUp(wParam, lParam, nMsg, hWnd)
	{
		if (hWnd != this.hWnd)
		{
			return
		}
		if (this.cur_sizer && this.nSplitterPos >= 0)
		{
			this.ReleaseCapture()
			this.DrawFocusRect(-1)
			this.cur_sizer.SetCurSplitterPos(this.nSplitterPos)
			this.nSplitterPos := -1
			this.cur_sizer := 0
			this.Size()
		}
	}
	OnLoad()
	{
		this.ReleaseNote()
		this.UpdateSB("프로그램이 시작되었습니다")
	}
	DrawFocusRect(pos)
	{
		hdc := DllCall("GetDC", "ptr", this.hWnd, "ptr")
		VarSetCapacity(rc, 4 * 4, 0)
		if (this.nSplitterPos >= 0)
		{
			rc_f := this.cur_sizer.GetFocusRect(this.nSplitterPos)
			NumPut(rc_f.left, rc, 0, "Int")
			NumPut(rc_f.top, rc, 4, "Int")
			NumPut(rc_f.right, rc, 8, "Int")
			NumPut(rc_f.bottom, rc, 12, "Int")
			DllCall("DrawFocusRect", "ptr", hdc, "ptr", &rc)
		}
		if (pos >= 0)
		{
			rc_f := this.cur_sizer.GetFocusRect(pos)
			NumPut(rc_f.left, rc, 0, "Int")
			NumPut(rc_f.top, rc, 4, "Int")
			NumPut(rc_f.right, rc, 8, "Int")
			NumPut(rc_f.bottom, rc, 12, "Int")
			DllCall("DrawFocusRect", "ptr", hdc, "ptr", &rc)
		}
		DllCall("ReleaseDC", "ptr", this.hWnd, "ptr", hdc)
	}
	OnMakeExe(bRuntimeSrc := true)
	{
		file := A_ScriptDir "\" this.name ".exe"
		src := bRuntimeSrc ? A_ScriptDir "\" A_ScriptName : A_ScriptDir "\" this.name ".ahk"
		ico := A_ScriptDir "\" this.name ".ico"
		compiler := RegExReplace(A_AhkPath, "[^\\]+\.exe$") "Compiler\Ahk2Exe.exe"
		this.InitLog()
		if (bRuntimeSrc && InStr(A_ScriptName, ".exe"))
		{
			this.AppendLog("이미 실행파일로 실행되었습니다.")
			return
		}
		if (!FileExist(src))
		{
			this.AppendLog(src "를 찾을수 없습니다")
			return 
		}
		if (!FileExist(compiler))
		{
			this.AppendLog(compiler "를 찾을수 없습니다")
			return 
		}
		if (FileExist(file))
		{
			FileDelete, %file%
			this.AppendLog("기존 " file " 를 지웁니다")
		}
		icon_option := ""
		if (FileExist(ico))
		{
			icon_option = /icon "%ico%"
		}
		RunWait, "%compiler%" /in "%src%" /out  "%file%" %icon_option%
		if (FileExist("bin\rcedit.exe"))
		{
			file_version := this.file_version
			RunWait, %A_ScriptDir%\bin\rcedit.exe %file% --set-file-version %file_version%
			cmd = %A_ScriptDir%\bin\rcedit.exe %file% --set-file-version %file_version%
			this.AppendLog(cmd)
		}
		this.AppendLog(file " 가 생성되었습니다")
	}
	OnMouseMove(wParam, lParam, nMsg, hWnd)
	{
		static bSetCursor := false
		if (hWnd != this.hWnd)
		{
			return
		}
		x := (lParam & 0xFFFF) << 48 >> 48
		y := lParam >> 16
		if (this.cur_sizer)
		{
			pos := this.cur_sizer.GetCurSplitterPos(x, y)
			this.DrawFocusRect(pos)
			this.nSplitterPos := pos
		}
		else if (ret := this.sizer.HitTest(x, y))
		{
			DllCall("SetCursor", "ptr", ret.bVertical ? this.hCursorNS : this.hCursorWE)
			bSetCursor := true
		}
		else if (bSetCursor)
		{
			DllCall("SetCursor", "ptr", this.hCursorArrow)
			bSetCursor := false
		}
	}
	OnReleaseNote()
	{
		http := new CHttp()
		response := JSON.Load(http.Get("https://auto.ddart.net/download/" this.name "/note"))
		this.AppendLog(StrReplace(response.note, "&#39;", "'"))
		this.InitLog(0)
	}
	OnNotify(wParam, lParam, nMsg, hWnd)
	{
		id := NumGet(lParam + 0, 1 * A_PtrSize, "UPtr")
		if (id = this.tabLog.id)
		{
			code := NumGet(lParam+0, 2 * A_PtrSize, "Int")
			if (code = CTabCtrl.TCN_SELCHANGE)
			{
			}
		}
	}
	OnSettingChanged(wParam, lParam, nMsg, hWnd)
	{
	}
	OnSize(wParam, lParam, nMsg, hWnd)
	{
		if (hWnd = this.hwnd && wParam != CWU.SIZE_MINIMIZED)
		{
			w := lParam & 0xFFFF
			h := lParam >> 16
			this.w := w
			this.h := h
			this.Size()
		}
	}
	OnStop()
	{
		if (this.bRunning && this.bRunnable)
		{
			this.bRunnable := false
			this.UpdateButtons()
		}
	}	
	OnSysCommand(wParam, lParam, nMsg, hWnd)
	{
		if (wParam = CWU.SC_CLOSE)
		{
			if (hWnd = this.hwnd)
			{
				this.OnClose()
				this.RemoveMessageHandlers()
				this.RemoveTimers()
				this.RemoveHotkeyHandler()
				this.RemoveErrorHandler()
				if (this.cmd_exit)
				{
					cmd := this.cmd_exit
					Run, %cmd%
				}
				this := ""
				ExitApp
			}
		}
	}
	OnUpdateCheck(bUser := true)
	{
		timestamp := A_Now
		timestamp -= 19700101000000, seconds
		name := this.name
		if (bUser)
		{
			this.InitLog()
		}
		else
		{
			RegRead, DisableAutoUpdate , HKEY_CURRENT_USER\Software\DDART\%name%, DisableAutoUpdate
			if (DisableAutoUpdate)
			{
				return
			}
			RegRead, LastUpdate , HKEY_CURRENT_USER\Software\DDART\%name%, LastUpdate
			if (LastUpdate && timestamp - LastUpdate < 12 * 60 * 60)
			{
				return 
			}
		}
		http := new CHttp()
		response := JSON.Load(http.Get("https://auto.ddart.net/download/" name "/version"))
		version := response.version
		RegWrite, REG_DWORD, HKEY_CURRENT_USER\Software\DDART\%name%, LastUpdate, %timestamp%
		if (this.version = version)
		{
			this.AppendLog("현재 " this.version " 버전은 최신버전입니다")
		}
		else if (this.version < version)
		{
			MsgBox, 4, 업데이트확인, 새로운 버전(%version%) 이 있습니다. 다운받으시겠습니까?, 10
			IfMsgBox, Yes
			{
				RegWrite, REG_DWORD , HKEY_CURRENT_USER\Software\DDART\%name%, DisableAutoUpdate, 0
				this.UpdateApplication(version, response.sha256)
			}
			else if (!bUser)
			{
				RegWrite, REG_DWORD , HKEY_CURRENT_USER\Software\DDART\%name%, DisableAutoUpdate, 1
			}
		}
		else if (!version)
		{
			this.AppendLog("현재 서버문제로 최신버전 확인을 할수가 없습니다")
		}
		else
		{
			this.AppendLog("개발자?")
		}
	}
	OnWebSocketDisconnected(wParam, lParam, nMsg, hWnd)
	{
		if (wParam)
		{
			this.ResetRunState()
		}
	}
}
_RemoveHotkey()
{
	; 아무것도 안함. Hotkey 에 등록된 binded function object 삭제를 위한 것
}
class CWU
{
	static WM_NULL := 0x0000
	static WM_CREATE := 0x0001
	static WM_DESTROY := 0x0002
	static WM_MOVE := 0x0003
	static WM_SIZE := 0x0005
	static WM_ACTIVATE := 0x0006
	static WM_SETFOCUS := 0x0007
	static WM_KILLFOCUS := 0x0008
	static WM_ENABLE := 0x000A
	static WM_SETREDRAW := 0x000B
	static WM_SETTEXT := 0x000C
	static WM_GETTEXT := 0x000D
	static WM_GETTEXTLENGTH := 0x000E
	static WM_PAINT := 0x000F
	static WM_CLOSE := 0x0010
	static WM_QUERYENDSESSION := 0x0011
	static WM_QUIT := 0x0012
	static WM_QUERYOPEN := 0x0013
	static WM_ERASEBKGND := 0x0014
	static WM_SYSCOLORCHANGE := 0x0015
	static WM_ENDSESSION := 0x0016
	static WM_SHOWWINDOW := 0x0018
	static WM_WININICHANGE := 0x001A
	static WM_SETTINGCHANGE := 0x001A
	static WM_DEVMODECHANGE := 0x001B
	static WM_ACTIVATEAPP := 0x001C
	static WM_FONTCHANGE := 0x001D
	static WM_TIMECHANGE := 0x001E
	static WM_CANCELMODE := 0x001F
	static WM_SETCURSOR := 0x0020
	static WM_MOUSEACTIVATE := 0x0021
	static WM_CHILDACTIVATE := 0x0022
	static WM_QUEUESYNC := 0x0023
	static WM_GETMINMAXINFO := 0x0024
	static WM_PAINTICON := 0x0026
	static WM_ICONERASEBKGND := 0x0027
	static WM_NEXTDLGCTL := 0x0028
	static WM_SPOOLERSTATUS := 0x002A
	static WM_DRAWITEM := 0x002B
	static WM_MEASUREITEM := 0x002C
	static WM_DELETEITEM := 0x002D
	static WM_VKEYTOITEM := 0x002E
	static WM_CHARTOITEM := 0x002F
	static WM_SETFONT := 0x0030
	static WM_GETFONT := 0x0031
	static WM_SETHOTKEY := 0x0032
	static WM_GETHOTKEY := 0x0033
	static WM_QUERYDRAGICON := 0x0037
	static WM_COMPAREITEM := 0x0039
	static WM_GETOBJECT := 0x003D
	static WM_COMPACTING := 0x0041
	static WM_COMMNOTIFY := 0x0044
	static WM_WINDOWPOSCHANGING := 0x0046
	static WM_WINDOWPOSCHANGED := 0x0047
	static WM_POWER := 0x0048
	static WM_COPYDATA := 0x004A
	static WM_CANCELJOURNAL := 0x004B
	static WM_NOTIFY := 0x004E
	static WM_INPUTLANGCHANGEREQUEST := 0x0050
	static WM_INPUTLANGCHANGE := 0x0051
	static WM_TCARD := 0x0052
	static WM_HELP := 0x0053
	static WM_USERCHANGED := 0x0054
	static WM_NOTIFYFORMAT := 0x0055
	static WM_CONTEXTMENU := 0x007B
	static WM_STYLECHANGING := 0x007C
	static WM_STYLECHANGED := 0x007D
	static WM_DISPLAYCHANGE := 0x007E
	static WM_GETICON := 0x007F
	static WM_SETICON := 0x0080
	static WM_NCCREATE := 0x0081
	static WM_NCDESTROY := 0x0082
	static WM_NCCALCSIZE := 0x0083
	static WM_NCHITTEST := 0x0084
	static WM_NCPAINT := 0x0085
	static WM_NCACTIVATE := 0x0086
	static WM_GETDLGCODE := 0x0087
	static WM_SYNCPAINT := 0x0088
	static WM_NCMOUSEMOVE := 0x00A0
	static WM_NCLBUTTONDOWN := 0x00A1
	static WM_NCLBUTTONUP := 0x00A2
	static WM_NCLBUTTONDBLCLK := 0x00A3
	static WM_NCRBUTTONDOWN := 0x00A4
	static WM_NCRBUTTONUP := 0x00A5
	static WM_NCRBUTTONDBLCLK := 0x00A6
	static WM_NCMBUTTONDOWN := 0x00A7
	static WM_NCMBUTTONUP := 0x00A8
	static WM_NCMBUTTONDBLCLK := 0x00A9
	static WM_NCXBUTTONDOWN := 0x00AB
	static WM_NCXBUTTONUP := 0x00AC
	static WM_NCXBUTTONDBLCLK := 0x00AD
	static WM_INPUT_DEVICE_CHANGE := 0x00FE
	static WM_INPUT := 0x00FF
	static WM_KEYFIRST := 0x0100
	static WM_KEYDOWN := 0x0100
	static WM_KEYUP := 0x0101
	static WM_CHAR := 0x0102
	static WM_DEADCHAR := 0x0103
	static WM_SYSKEYDOWN := 0x0104
	static WM_SYSKEYUP := 0x0105
	static WM_SYSCHAR := 0x0106
	static WM_SYSDEADCHAR := 0x0107
	static WM_UNICHAR := 0x0109
	static WM_KEYLAST := 0x0109
	static WM_IME_STARTCOMPOSITION := 0x010D
	static WM_IME_ENDCOMPOSITION := 0x010E
	static WM_IME_COMPOSITION := 0x010F
	static WM_IME_KEYLAST := 0x010F
	static WM_INITDIALOG := 0x0110
	static WM_COMMAND := 0x0111
	static WM_SYSCOMMAND := 0x0112
	static WM_TIMER := 0x0113
	static WM_HSCROLL := 0x0114
	static WM_VSCROLL := 0x0115
	static WM_INITMENU := 0x0116
	static WM_INITMENUPOPUP := 0x0117
	static WM_GESTURE := 0x0119
	static WM_GESTURENOTIFY := 0x011A
	static WM_MENUSELECT := 0x011F
	static WM_MENUCHAR := 0x0120
	static WM_ENTERIDLE := 0x0121
	static WM_MENURBUTTONUP := 0x0122
	static WM_MENUDRAG := 0x0123
	static WM_MENUGETOBJECT := 0x0124
	static WM_UNINITMENUPOPUP := 0x0125
	static WM_MENUCOMMAND := 0x0126
	static WM_CHANGEUISTATE := 0x0127
	static WM_UPDATEUISTATE := 0x0128
	static WM_QUERYUISTATE := 0x0129
	static WM_CTLCOLORMSGBOX := 0x0132
	static WM_CTLCOLOREDIT := 0x0133
	static WM_CTLCOLORLISTBOX := 0x0134
	static WM_CTLCOLORBTN := 0x0135
	static WM_CTLCOLORDLG := 0x0136
	static WM_CTLCOLORSCROLLBAR := 0x0137
	static WM_CTLCOLORSTATIC := 0x0138
	static MN_GETHMENU := 0x01E1
	static WM_MOUSEFIRST := 0x0200
	static WM_MOUSEMOVE := 0x0200
	static WM_LBUTTONDOWN := 0x0201
	static WM_LBUTTONUP := 0x0202
	static WM_LBUTTONDBLCLK := 0x0203
	static WM_RBUTTONDOWN := 0x0204
	static WM_RBUTTONUP := 0x0205
	static WM_RBUTTONDBLCLK := 0x0206
	static WM_MBUTTONDOWN := 0x0207
	static WM_MBUTTONUP := 0x0208
	static WM_MBUTTONDBLCLK := 0x0209
	static WM_MOUSEWHEEL := 0x020A
	static WM_XBUTTONDOWN := 0x020B
	static WM_XBUTTONUP := 0x020C
	static WM_XBUTTONDBLCLK := 0x020D
	static WM_MOUSEHWHEEL := 0x020E
	static WM_PARENTNOTIFY := 0x0210
	static WM_ENTERMENULOOP := 0x0211
	static WM_EXITMENULOOP := 0x0212
	static WM_NEXTMENU := 0x0213
	static WM_SIZING := 0x0214
	static WM_CAPTURECHANGED := 0x0215
	static WM_MOVING := 0x0216
	static WM_POWERBROADCAST := 0x0218
	static WM_DEVICECHANGE := 0x0219
	static WM_MDICREATE := 0x0220
	static WM_MDIDESTROY := 0x0221
	static WM_MDIACTIVATE := 0x0222
	static WM_MDIRESTORE := 0x0223
	static WM_MDINEXT := 0x0224
	static WM_MDIMAXIMIZE := 0x0225
	static WM_MDITILE := 0x0226
	static WM_MDICASCADE := 0x0227
	static WM_MDIICONARRANGE := 0x0228
	static WM_MDIGETACTIVE := 0x0229
	static WM_MDISETMENU := 0x0230
	static WM_ENTERSIZEMOVE := 0x0231
	static WM_EXITSIZEMOVE := 0x0232
	static WM_DROPFILES := 0x0233
	static WM_MDIREFRESHMENU := 0x0234
	static WM_POINTERDEVICECHANGE := 0x238
	static WM_POINTERDEVICEINRANGE := 0x239
	static WM_POINTERDEVICEOUTOFRANGE := 0x23A
	static WM_TOUCH := 0x0240
	static WM_NCPOINTERUPDATE := 0x0241
	static WM_NCPOINTERDOWN := 0x0242
	static WM_NCPOINTERUP := 0x0243
	static WM_POINTERUPDATE := 0x0245
	static WM_POINTERDOWN := 0x0246
	static WM_POINTERUP := 0x0247
	static WM_POINTERENTER := 0x0249
	static WM_POINTERLEAVE := 0x024A
	static WM_POINTERACTIVATE := 0x024B
	static WM_POINTERCAPTURECHANGED := 0x024C
	static WM_TOUCHHITTESTING := 0x024D
	static WM_POINTERWHEEL := 0x024E
	static WM_POINTERHWHEEL := 0x024F
	static DM_POINTERHITTEST := 0x0250
	static WM_POINTERROUTEDTO := 0x0251
	static WM_POINTERROUTEDAWAY := 0x0252
	static WM_POINTERROUTEDRELEASED := 0x0253
	static WM_IME_SETCONTEXT := 0x0281
	static WM_IME_NOTIFY := 0x0282
	static WM_IME_CONTROL := 0x0283
	static WM_IME_COMPOSITIONFULL := 0x0284
	static WM_IME_SELECT := 0x0285
	static WM_IME_CHAR := 0x0286
	static WM_IME_REQUEST := 0x0288
	static WM_IME_KEYDOWN := 0x0290
	static WM_IME_KEYUP := 0x0291
	static WM_MOUSEHOVER := 0x02A1
	static WM_MOUSELEAVE := 0x02A3
	static WM_NCMOUSEHOVER := 0x02A0
	static WM_NCMOUSELEAVE := 0x02A2
	static WM_WTSSESSION_CHANGE := 0x02B1
	static WM_TABLET_FIRST := 0x02c0
	static WM_TABLET_LAST := 0x02df
	static WM_DPICHANGED := 0x02E0
	static WM_DPICHANGED_BEFOREPARENT := 0x02E2
	static WM_DPICHANGED_AFTERPARENT := 0x02E3
	static WM_GETDPISCALEDSIZE := 0x02E4
	static WM_CUT := 0x0300
	static WM_COPY := 0x0301
	static WM_PASTE := 0x0302
	static WM_CLEAR := 0x0303
	static WM_UNDO := 0x0304
	static WM_RENDERFORMAT := 0x0305
	static WM_RENDERALLFORMATS := 0x0306
	static WM_DESTROYCLIPBOARD := 0x0307
	static WM_DRAWCLIPBOARD := 0x0308
	static WM_PAINTCLIPBOARD := 0x0309
	static WM_VSCROLLCLIPBOARD := 0x030A
	static WM_SIZECLIPBOARD := 0x030B
	static WM_ASKCBFORMATNAME := 0x030C
	static WM_CHANGECBCHAIN := 0x030D
	static WM_HSCROLLCLIPBOARD := 0x030E
	static WM_QUERYNEWPALETTE := 0x030F
	static WM_PALETTEISCHANGING := 0x0310
	static WM_PALETTECHANGED := 0x0311
	static WM_HOTKEY := 0x0312
	static WM_PRINT := 0x0317
	static WM_PRINTCLIENT := 0x0318
	static WM_APPCOMMAND := 0x0319
	static WM_THEMECHANGED := 0x031A
	static WM_CLIPBOARDUPDATE := 0x031D
	static WM_DWMCOMPOSITIONCHANGED := 0x031E
	static WM_DWMNCRENDERINGCHANGED := 0x031F
	static WM_DWMCOLORIZATIONCOLORCHANGED := 0x0320
	static WM_DWMWINDOWMAXIMIZEDCHANGE := 0x0321
	static WM_DWMSENDICONICTHUMBNAIL := 0x0323
	static WM_DWMSENDICONICLIVEPREVIEWBITMAP := 0x0326
	static WM_GETTITLEBARINFOEX := 0x033F
	static WM_HANDHELDFIRST := 0x0358
	static WM_HANDHELDLAST := 0x035F
	static WM_AFXFIRST := 0x0360
	static WM_AFXLAST := 0x037F
	static WM_PENWINFIRST := 0x0380
	static WM_PENWINLAST := 0x038F
	static WM_APP := 0x8000
	static WM_USER := 0x0400

	static WS_OVERLAPPED := 0x00000000
	static WS_POPUP := 0x80000000
	static WS_CHILD := 0x40000000
	static WS_MINIMIZE := 0x20000000
	static WS_VISIBLE := 0x10000000
	static WS_DISABLED := 0x08000000
	static WS_CLIPSIBLINGS := 0x04000000
	static WS_CLIPCHILDREN := 0x02000000
	static WS_MAXIMIZE := 0x01000000
	static WS_CAPTION := 0x00C00000
	static WS_BORDER := 0x00800000
	static WS_DLGFRAME := 0x00400000
	static WS_VSCROLL := 0x00200000
	static WS_HSCROLL := 0x00100000
	static WS_SYSMENU := 0x00080000
	static WS_THICKFRAME := 0x00040000
	static WS_GROUP := 0x00020000
	static WS_TABSTOP := 0x00010000
	static WS_MINIMIZEBOX := 0x00020000
	static WS_MAXIMIZEBOX := 0x00010000
	static WS_TILED := CWU.WS_OVERLAPPED
	static WS_ICONIC := CWU.WS_MINIMIZE
	static WS_SIZEBOX := CWU.WS_THICKFRAME
	static WS_OVERLAPPEDWINDOW := CWU.WS_OVERLAPPED |  CWU.WS_CAPTION | CWU.WS_SYSMENU |  CWU.WS_THICKFRAME | CWU.WS_MINIMIZEBOX | CWU.WS_MAXIMIZEBOX
	static WS_POPUPWINDOW := (CWU.WS_POPUP | CWU.WS_BORDER | CWU.WS_SYSMENU)
	static WS_CHILDWINDOW := CWU.WS_CHILD
	static WS_TILEDWINDOW := CWU.WS_OVERLAPPEDWINDOW
	static WS_EX_DLGMODALFRAME := 0x00000001
	static WS_EX_NOPARENTNOTIFY := 0x00000004
	static WS_EX_TOPMOST := 0x00000008
	static WS_EX_ACCEPTFILES := 0x00000010
	static WS_EX_TRANSPARENT := 0x00000020
	static WS_EX_MDICHILD := 0x00000040
	static WS_EX_TOOLWINDOW := 0x00000080
	static WS_EX_WINDOWEDGE := 0x00000100
	static WS_EX_CLIENTEDGE := 0x00000200
	static WS_EX_CONTEXTHELP := 0x00000400
	static WS_EX_RIGHT := 0x00001000
	static WS_EX_LEFT := 0x00000000
	static WS_EX_RTLREADING := 0x00002000
	static WS_EX_LTRREADING := 0x00000000
	static WS_EX_LEFTSCROLLBAR := 0x00004000
	static WS_EX_RIGHTSCROLLBAR := 0x00000000
	static WS_EX_CONTROLPARENT := 0x00010000
	static WS_EX_STATICEDGE := 0x00020000
	static WS_EX_APPWINDOW := 0x00040000
	static WS_EX_OVERLAPPEDWINDOW := (CWU.WS_EX_WINDOWEDGE | CWU.WS_EX_CLIENTEDGE)
	static WS_EX_PALETTEWINDOW := (CWU.WS_EX_WINDOWEDGE | CWU.WS_EX_TOOLWINDOW | CWU.WS_EX_TOPMOST)
	static WS_EX_LAYERED := 0x00080000
	static WS_EX_NOINHERITLAYOUT := 0x00100000
	static WS_EX_NOREDIRECTIONBITMAP := 0x00200000
	static WS_EX_LAYOUTRTL := 0x00400000
	static WS_EX_COMPOSITED := 0x02000000
	static WS_EX_NOACTIVATE := 0x08000000
	
	static CW_USEDEFAULT := 0x80000000
	
	static CTLCOLOR_MSGBOX := 0
	static CTLCOLOR_EDIT := 1
	static CTLCOLOR_LISTBOX := 2
	static CTLCOLOR_BTN := 3
	static CTLCOLOR_DLG := 4
	static CTLCOLOR_SCROLLBAR := 5
	static CTLCOLOR_STATIC := 6
	static CTLCOLOR_MAX := 7
	static COLOR_SCROLLBAR := 0
	static COLOR_BACKGROUND := 1
	static COLOR_ACTIVECAPTION := 2
	static COLOR_INACTIVECAPTION := 3
	static COLOR_MENU := 4
	static COLOR_WINDOW := 5
	static COLOR_WINDOWFRAME := 6
	static COLOR_MENUTEXT := 7
	static COLOR_WINDOWTEXT := 8
	static COLOR_CAPTIONTEXT := 9
	static COLOR_ACTIVEBORDER := 10
	static COLOR_INACTIVEBORDER := 11
	static COLOR_APPWORKSPACE := 12
	static COLOR_HIGHLIGHT := 13
	static COLOR_HIGHLIGHTTEXT := 14
	static COLOR_BTNFACE := 15
	static COLOR_BTNSHADOW := 16
	static COLOR_GRAYTEXT := 17
	static COLOR_BTNTEXT := 18
	static COLOR_INACTIVECAPTIONTEXT := 19
	static COLOR_BTNHIGHLIGHT := 20
	static COLOR_3DDKSHADOW := 21
	static COLOR_3DLIGHT := 22
	static COLOR_INFOTEXT := 23
	static COLOR_INFOBK := 24
	static COLOR_HOTLIGHT := 26
	static COLOR_GRADIENTACTIVECAPTION := 27
	static COLOR_GRADIENTINACTIVECAPTION := 28
	static COLOR_MENUHILIGHT := 29
	static COLOR_MENUBAR := 30
	static COLOR_DESKTOP := CWU.COLOR_BACKGROUND
	static COLOR_3DFACE := CWU.COLOR_BTNFACE
	static COLOR_3DSHADOW := CWU.COLOR_BTNSHADOW
	static COLOR_3DHIGHLIGHT := CWU.COLOR_BTNHIGHLIGHT
	static COLOR_3DHILIGHT := CWU.COLOR_BTNHIGHLIGHT
	static COLOR_BTNHILIGHT := CWU.COLOR_BTNHIGHLIGHT	
	
	static RT_CURSOR := 1
	static RT_BITMAP := 2
	static RT_ICON := 3
	static RT_MENU := 4
	static RT_DIALOG := 5
	static RT_STRING := 6
	static RT_FONTDIR := 7
	static RT_FONT := 8
	static RT_ACCELERATOR := 9
	static RT_RCDATA := 10
	static RT_MESSAGETABLE := 11
	static DIFFERENCE := 11
	static RT_GROUP_CURSOR := (CWU.RT_CURSOR + CWU.DIFFERENCE)
	static RT_GROUP_ICON := (CWU.RT_ICON + CWU.DIFFERENCE)
	static RT_VERSION := 16
	static RT_DLGINCLUDE := 17
	static RT_PLUGPLAY := 19
	static RT_VXD := 20
	static RT_ANICURSOR := 21
	static RT_ANIICON := 22
	static RT_HTML := 23
	static RT_MANIFEST := 24
	static CREATEPROCESS_MANIFEST_RESOURCE_ID := 1
	static ISOLATIONAWARE_MANIFEST_RESOURCE_ID := 2
	static ISOLATIONAWARE_NOSTATICIMPORT_MANIFEST_RESOURCE_ID := 3
	static ISOLATIONPOLICY_MANIFEST_RESOURCE_ID := 4
	static ISOLATIONPOLICY_BROWSER_MANIFEST_RESOURCE_ID := 5
	static MINIMUM_RESERVED_MANIFEST_RESOURCE_ID := 1
	static MAXIMUM_RESERVED_MANIFEST_RESOURCE_ID := 16
	static SWP_NOSIZE := 0x0001
	static SWP_NOMOVE := 0x0002
	static SWP_NOZORDER := 0x0004
	static SWP_NOREDRAW := 0x0008
	static SWP_NOACTIVATE := 0x0010
	static SWP_FRAMECHANGED := 0x0020
	static SWP_SHOWWINDOW := 0x0040
	static SWP_HIDEWINDOW := 0x0080
	static SWP_NOCOPYBITS := 0x0100
	static SWP_NOOWNERZORDER := 0x0200
	static SWP_NOSENDCHANGING := 0x0400
	static SWP_DRAWFRAME := CWU.SWP_FRAMECHANGED
	static SWP_NOREPOSITION := CWU.SWP_NOOWNERZORDER
	static SWP_DEFERERASE := 0x2000
	static SWP_ASYNCWINDOWPOS := 0x4000
	static HWND_TOP := 0
	static HWND_BOTTOM := 1
	static HWND_TOPMOST := -1
	static HWND_NOTOPMOST := -2
	
	static IMAGE_BITMAP := 0
	static IMAGE_ICON := 1
	static IMAGE_CURSOR := 2
	static IMAGE_ENHMETAFILE := 3
	static LR_DEFAULTCOLOR := 0x00000000
	static LR_MONOCHROME := 0x00000001
	static LR_COLOR := 0x00000002
	static LR_COPYRETURNORG := 0x00000004
	static LR_COPYDELETEORG := 0x00000008
	static LR_LOADFROMFILE := 0x00000010
	static LR_LOADTRANSPARENT := 0x00000020
	static LR_DEFAULTSIZE := 0x00000040
	static LR_VGACOLOR := 0x00000080
	static LR_LOADMAP3DCOLORS := 0x00001000
	static LR_CREATEDIBSECTION := 0x00002000
	static LR_COPYFROMRESOURCE := 0x00004000
	static LR_SHARED := 0x00008000
	
	static GW_HWNDFIRST := 0
	static GW_HWNDLAST := 1
	static GW_HWNDNEXT := 2
	static GW_HWNDPREV := 3
	static GW_OWNER := 4
	static GW_CHILD := 5
	static GW_ENABLEDPOPUP := 6
	static GW_MAX := 6	
	static GWL_WNDPROC := (-4)
	static GWL_HINSTANCE := (-6)
	static GWL_HWNDPARENT := (-8)
	static GWL_STYLE := (-16)
	static GWL_EXSTYLE := (-20)
	static GWL_USERDATA := (-21)
	static GWL_ID := (-12)
	
	static SC_SIZE := 0xF000
	static SC_MOVE := 0xF010
	static SC_MINIMIZE := 0xF020
	static SC_MAXIMIZE := 0xF030
	static SC_NEXTWINDOW := 0xF040
	static SC_PREVWINDOW := 0xF050
	static SC_CLOSE := 0xF060
	static SC_VSCROLL := 0xF070
	static SC_HSCROLL := 0xF080
	static SC_MOUSEMENU := 0xF090
	static SC_KEYMENU := 0xF100
	static SC_ARRANGE := 0xF110
	static SC_RESTORE := 0xF120
	static SC_TASKLIST := 0xF130
	static SC_SCREENSAVE := 0xF140
	static SC_HOTKEY := 0xF150
	static SC_DEFAULT := 0xF160
	static SC_MONITORPOWER := 0xF170
	static SC_CONTEXTHELP := 0xF180
	static SC_SEPARATOR := 0xF00F
	static SM_CXSCREEN := 0
	static SM_CYSCREEN := 1
	static SM_CXVSCROLL := 2
	static SM_CYHSCROLL := 3
	static SM_CYCAPTION := 4
	static SM_CXBORDER := 5
	static SM_CYBORDER := 6
	static SM_CXDLGFRAME := 7
	static SM_CYDLGFRAME := 8
	static SM_CYVTHUMB := 9
	static SM_CXHTHUMB := 10
	static SM_CXICON := 11
	static SM_CYICON := 12
	static SM_CXCURSOR := 13
	static SM_CYCURSOR := 14
	static SM_CYMENU := 15
	static SM_CXFULLSCREEN := 16
	static SM_CYFULLSCREEN := 17
	static SM_CYKANJIWINDOW := 18
	static SM_MOUSEPRESENT := 19
	static SM_CYVSCROLL := 20
	static SM_CXHSCROLL := 21
	static SM_DEBUG := 22
	static SM_SWAPBUTTON := 23
	static SM_RESERVED1 := 24
	static SM_RESERVED2 := 25
	static SM_RESERVED3 := 26
	static SM_RESERVED4 := 27
	static SM_CXMIN := 28
	static SM_CYMIN := 29
	static SM_CXSIZE := 30
	static SM_CYSIZE := 31
	static SM_CXFRAME := 32
	static SM_CYFRAME := 33
	static SM_CXMINTRACK := 34
	static SM_CYMINTRACK := 35
	static SM_CXDOUBLECLK := 36
	static SM_CYDOUBLECLK := 37
	static SM_CXICONSPACING := 38
	static SM_CYICONSPACING := 39
	static SM_MENUDROPALIGNMENT := 40
	static SM_PENWINDOWS := 41
	static SM_DBCSENABLED := 42
	static SM_CMOUSEBUTTONS := 43
	static SM_CXFIXEDFRAME := CWU.SM_CXDLGFRAME
	static SM_CYFIXEDFRAME := CWU.SM_CYDLGFRAME
	static SM_CXSIZEFRAME := CWU.SM_CXFRAME
	static SM_CYSIZEFRAME := CWU.SM_CYFRAME
	static SM_SECURE := 44
	static SM_CXEDGE := 45
	static SM_CYEDGE := 46
	static SM_CXMINSPACING := 47
	static SM_CYMINSPACING := 48
	static SM_CXSMICON := 49
	static SM_CYSMICON := 50
	static SM_CYSMCAPTION := 51
	static SM_CXSMSIZE := 52
	static SM_CYSMSIZE := 53
	static SM_CXMENUSIZE := 54
	static SM_CYMENUSIZE := 55
	static SM_ARRANGE := 56
	static SM_CXMINIMIZED := 57
	static SM_CYMINIMIZED := 58
	static SM_CXMAXTRACK := 59
	static SM_CYMAXTRACK := 60
	static SM_CXMAXIMIZED := 61
	static SM_CYMAXIMIZED := 62
	static SM_NETWORK := 63
	static SM_CLEANBOOT := 67
	static SM_CXDRAG := 68
	static SM_CYDRAG := 69
	static SM_SHOWSOUNDS := 70
	static SM_CXMENUCHECK := 71
	static SM_CYMENUCHECK := 72
	static SM_SLOWMACHINE := 73
	static SM_MIDEASTENABLED := 74
	static SM_MOUSEWHEELPRESENT := 75
	static SM_XVIRTUALSCREEN := 76
	static SM_YVIRTUALSCREEN := 77
	static SM_CXVIRTUALSCREEN := 78
	static SM_CYVIRTUALSCREEN := 79
	static SM_CMONITORS := 80
	static SM_SAMEDISPLAYFORMAT := 81
	static SM_IMMENABLED := 82
	static SM_CXFOCUSBORDER := 83
	static SM_CYFOCUSBORDER := 84
	static SM_TABLETPC := 86
	static SM_MEDIACENTER := 87
	static SM_STARTER := 88
	static SM_SERVERR2 := 89
	static SM_MOUSEHORIZONTALWHEELPRESENT := 91
	static SM_CXPADDEDBORDER := 92
	static SM_DIGITIZER := 94
	static SM_MAXIMUMTOUCHES := 95
	static SM_CMETRICS := 97
	static SM_REMOTESESSION := 0x1000
	static SM_SHUTTINGDOWN := 0x2000
	static SM_REMOTECONTROL := 0x2001
	static SM_CARETBLINKINGENABLED := 0x2002
	static SM_CONVERTIBLESLATEMODE := 0x2003
	static SM_SYSTEMDOCKED := 0x2004	
	static IDC_ARROW := 32512
	static IDC_IBEAM := 32513
	static IDC_WAIT := 32514
	static IDC_CROSS := 32515
	static IDC_UPARROW := 32516
	static IDC_SIZE := 32640
	static IDC_ICON := 32641
	static IDC_SIZENWSE := 32642
	static IDC_SIZENESW := 32643
	static IDC_SIZEWE := 32644
	static IDC_SIZENS := 32645
	static IDC_SIZEALL := 32646
	static IDC_NO := 32648
	static IDC_HAND := 32649
	static IDC_APPSTARTING := 32650
	static IDC_HELP := 32651
	static IDC_PIN := 32671
	static IDC_PERSON := 32672
	static MK_LBUTTON := 0x0001
	static MK_RBUTTON := 0x0002
	static MK_SHIFT := 0x0004
	static MK_CONTROL := 0x0008
	static MK_MBUTTON := 0x0010
	static MK_XBUTTON1 := 0x0020
	static MK_XBUTTON2 := 0x0040
	
	static SIZE_RESTORED  := 0
	static SIZE_MINIMIZED := 1
	static SIZE_MAXIMIZED := 2
	static SIZE_MAXSHOW := 3
	static SIZE_MAXHIDE := 4
	
	GetSystemMetrics(nIndex)
	{
		return DllCall("GetSystemMetrics", "int", nIndex)
	}
}

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Windows & Controls
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
class CRect
{
	__New(l :=0 , t := 0, r := 0, b := 0)
	{
		this.left := l + 0
		this.top := t + 0
		this.right := r + 0
		this.bottom := b + 0
	}
}
class CPoint
{
	__New(x, y)
	{
		this.x := x + 0
		this.y := y + 0
	}
}
class CSize
{
	__New(w, h)
	{
		this.w := w + 0
		this.h := h + 0
	}
}
class CPosition
{
	__New(x, y, w, h)
	{
		this.x := x + 0
		this.y := y + 0
		this.w := w + 0
		this.h := h + 0
	}
}
class CSplitter
{
	__New(min)
	{
		this.split := Array(0, 0)
		this.min := min
	}
	Add(pos := 0)
	{
		return this.split.InsertAt(this.split.MaxIndex(), pos)
	}
	Adjust(i, v)
	{
		if (v < this.split[i] + this.min)
		{
			return this.split[i] + this.min
		}
		else if (v > this.split[i + 2] - this.min)
		{
			return this.split[i + 2] - this.min
		}
		return v
	}
	Exist()
	{
		return this.split.Length() > 2
	}
	GetIndex(pos)
	{
		if (!this.Exist())
		{
			return 0
		}
		Loop, % this.split.Length() - 2
		{
			v := this.Get(A_Index)
			if (Abs(v - pos) <= 2)
			{
				return A_Index
			}
			else if (v > pos)
			{
				return 0
			}
		}
		return 0
	}
	Get(i)
	{
		return this.split[i + 1]
	}
	GetStart(i)
	{
		return this.split[i]
	}
	GetDistance(i)
	{
		return this.Get(i) - this.GetStart(i)
	}
	Set(i, v)
	{
		this.split[i + 1] := v
	}
	SetRange(min, max)
	{
		this.split[1] := min
		this.split[this.split.MaxIndex()] := max
		if (this.Exist())
		{
			Loop, % this.split.MaxIndex() - 2
			{
				this.Set(A_Index, this.Adjust(A_Index, this.Get(A_Index)))
			}
		}
	}
}
class CAutoSizer
{
	__New(bVertical := true, spacing := 0, min := 25)
	{
		this.bVertical := bVertical
		this.children := Array()
		this.splitter := new CSplitter(min)
		this.spacing := spacing
		this.rc := new CRect()
		this.nCurSplitter := 0
		this.bInitSplitter := false
	}
	Add(child)
	{
		if (this.children.Push(child) > 1)
		{
			this.splitter.Add()
		}
	}
	Count()
	{
		return this.children.Length()
	}
	GetCurSplitterPos(x, y)
	{
		pos := this.bVertical ? (y - this.rc.top) : (x - this.rc.left)
		return this.splitter.Adjust(this.nCurSplitter, pos)
	}
	GetChildPos(i)
	{
		if (this.bVertical)
		{
			return new CPosition(this.rc.left + this.spacing
			, this.rc.top + this.splitter.GetStart(i)+ this.spacing
			, this.rc.right - this.rc.left - 2 * this.spacing
			, this.splitter.GetDistance(i) - 2 * this.spacing)
		}
		else
		{
			return new CPosition(this.rc.left + this.splitter.GetStart(i)  + this.spacing
			, this.rc.top + this.spacing
			, this.splitter.GetDistance(i) - 2 * this.spacing
			, this.rc.bottom - this.rc.top - 2 * this.spacing)
		}
	}
	GetFocusRect(pos)
	{
		if (this.bVertical)
		{
			return new CRect(this.rc.left, this.rc.top + pos - 2, this.rc.right, this.rc.top + pos + 2)
		}
		else
		{
			return new CRect(this.rc.left + pos - 2, this.rc.top, this.rc.left + pos + 2, this.rc.bottom)
		}
	}
	GetSplitterPos(i)
	{
		return this.splitter.Get(i)
	}
	HitTest(x, y)
	{
		pos := this.bVertical ? (y - this.rc.top) : (x - this.rc.left)
		this.nCurSplitter := this.splitter.GetIndex(pos)
		if (this.nCurSplitter)
		{
			return this
		}
		Loop, % this.children.Length()
		{
			if (pos > this.splitter.Get(A_Index))
			{
				return 0
			}
			if (this.children[A_Index].splitter && ret := this.children[A_Index].HitTest(x, y))
			{
				return ret
			}
		}
		return 0
	}
	InsertAt(i, child, pos := 25)
	{
		this.children.InsertAt(i, child)
		if (this.children.Length() > 1)
		{
			this.splitter.split.InsertAt(i + 1, pos)
		}
	}
	RemoveAt(i)
	{
		if (i > 0 && i <= this.children.MaxIndex())
		{
			this.children.RemoveAt(i)
			this.splitter.split.RemoveAt(i + 1)
		}
	}
	ResizeChildSizer()
	{
		Loop, % this.children.Length()
		{
			if (this.children[A_Index].splitter)
			{
				if (this.bVertical)
				{
					this.children[A_Index].SetRect(new CRect(this.rc.left, this.rc.top + this.splitter.GetStart(A_Index), this.rc.right, this.rc.top + this.splitter.GetDistance(A_Index)))
				}
				else
				{
					this.children[A_Index].SetRect(new CRect(this.rc.left + this.splitter.GetStart(A_Index), this.rc.top, this.rc.left + this.splitter.GetDistance(A_Index), this.rc.bottom))
				}
			}
		}
	}
	SetCurSplitterPos(pos)
	{
		this.SetSplitterPos(this.nCurSplitter, pos)
	}
	SetRect(rc)
	{
		this.rc := rc
		if (this.bVertical)
		{
			this.splitter.SetRange(0, rc.bottom - rc.top)
		}
		else
		{
			this.splitter.SetRange(0, rc.right - rc.left)
		}
		this.ResizeChildSizer()
	}
	SetSplitterPos(i, v)
	{
		if (this.Count() >= i + 1)
		{
			this.splitter.Set(i, this.splitter.Adjust(i, v))
			this.ResizeChildSizer()
		}
	}
	Size()
	{
		Loop, % this.children.Length()
		{
			if (this.children[A_Index].splitter)
			{
				this.children[A_Index].Size()
			}
			else
			{
				pos := this.GetChildPos(A_Index)
				this.children[A_Index].Move(pos.x, pos.y, pos.w, pos.h)
			}
		}
	}
}
class CWNDCLASSEX
{
	static CS_VREDRAW := 0x0001
	static CS_HREDRAW := 0x0002
	static CS_DBLCLKS := 0x0008
	static CS_OWNDC := 0x0020
	static CS_CLASSDC := 0x0040
	static CS_PARENTDC := 0x0080
	static CS_NOCLOSE := 0x0200
	static CS_SAVEBITS := 0x0800
	static CS_BYTEALIGNCLIENT := 0x1000
	static CS_BYTEALIGNWINDOW := 0x2000
	static CS_GLOBALCLASS := 0x4000
	__New(szClassName, szWindowName)
	{
		this.style := this.CS_VREDRAW | this.CS_HREDRAW
		this.szClassName := szClassName
		this.szWindowName := szWindowName
		this.cbClsExtra := 0
		this.cbWndExtra := 0
		this.hInstance := 0
		this.hIcon := 0
		this.hCursor := 0
		this.hbrBackground := 0
		this.hIconSm := 0
	}
}
class CCWEX
{
	__New(szClassName, szWindowName)
	{
		this.szClassName := szClassName
		this.szWindowName := szWindowName
		this.dwStyle := 0
		this.dwExStyle := 0
		this.x := CWU.CW_USEDEFAULT
		this.y :=CWU.CW_USEDEFAULT
		this.nWidth := CWU.CW_USEDEFAULT
		this.nHeight := CWU.CW_USEDEFAULT
		this.hMenu := 0
		this.hInstance := 0
		this.lpParam := 0
	}
}
class CWindow
{
	static _inst := Object()
	static _inst_create := ""
	static _wndproc := RegisterCallback(CWindow.WndProc, "" , 4)
	__New(hWnd)
	{
		if (RegExMatch(hWnd, "i)^0x[0-9a-f]+$") || RegExMatch(hWnd, "^\d+$"))
		{
			this.hWnd := hWnd
			this.title := "ahk_id " hWnd
		}
		else
		{
			this.title := hWnd
			this.hWnd := WinExist(this.title)
		}
	}
	Activate()
	{
		WinActivate, % this.title
	}
	Active()
	{
		return WinActive(this.title)
	}
	AlwaysOnTop(bOn := true)
	{
		onoff := bOn ? "On" : "Off" 
		WinSet, Alwaysontop, %onoff%, % this.title
	}
	Click(x, y)
	{
		lParam := this.MakeLong(x, y)
        this.PostMessage(CWU.WM_LBUTTONDOWN, CWU.MK_LBUTTON, lParam)
        this.PostMessage(CWU.WM_LBUTTONUP, CWU.MK_LBUTTON, lParam)
	}
	ClientToScreen(x, y)
	{
		VarSetCapacity(pt, 8, 0)
		NumPut(x, pt, 0, "int")
		NumPut(y, pt, 4, "int")
		if (DllCall("ClientToScreen", "ptr", this.hWnd, "ptr", &pt))
		{
			return this.GetPOINTObject(&pt)
		}
		return false
	}
	Close()
	{
		if (this.Exist())
		{
			WinClose, % this.title
			return true
		}
		return false
	}
	Control(name, subcmd, value :="")
	{
		Control, %subcmd%, %value%, %name%, % this.title
	}
	ControlClick(name := "", x := 0, y := 0)
	{
		pos := "x" x " y" y
		ControlClick, %name%, % this.title, , , , %pos%
	}
	ControlFocus(name := "")
	{
		ControlFocus, %name%, % this.title
	}
	ControlGet(name, subcmd := "hwnd", value := "")
	{
		ControlGet, outvar, %subcmd%, %value%, %name%, % this.title
		return outvar
	}
	ControlGetFocus()
	{
		ControlGetFocus, var, % this.title
		return var
	}
	ControlGetPos(name)
	{
		ControlGetPos, x, y, w, h, %name%, % this.title
		return new CPosition(x, y, w, h)
	}
	ControlGetText(name = "")
	{
		ControlGetText, text, %name%, % this.title
		return text
	}
	ControlMove(name, x, y, w := "", h := "")
	{
		ControlMove, %name%, x, y, w, h, % this.title
	}
	ControlPostMessage(uMsg, wParam := 0, lParam :=0, Control := "")
	{
		PostMessage, %uMsg%, %wParam%, %lParam%, %Control%, % this.title
	}
	ControlSend(name, keys)
	{
		ControlSend, %name%, %keys%, % this.title
	}
	ControlSendMessage(uMsg, wParam :=0, lParam := 0, Control := "", timeout := 5000)
	{
		SendMessage, %uMsg%, %wParam%, %lParam%, %Control% , % this.title, , , , %timeout%
		return ErrorLevel
	}
	ControlSendRaw(name, keys)
	{
		ControlSendRaw, %name%, %keys%, % this.title
	}
	
	ControlSetText(name, txt)
	{
		ControlSetText, %name%, %txt%, % this.title
	}
	DestroyWindow()
	{
		return DllCall("DestroyWindow", "ptr", this.hWnd)
	}
	Exist()
	{
		return WinExist(this.title)
	}
	Get(subcmd)
	{
		WinGet, var, %subcmd%, % this.title
		return var
	}
	GetClass()
	{
		WinGetClass, var, % this.title
		return var
	}
	GetClientRect()
	{
		VarSetCapacity(rect, 16, 0)
		if (DllCall("GetClientRect", "ptr", this.hWnd, "ptr", &rect))
		{
			return this.GetRECTObject(&rect)
		}
		return 0
	}
	GetControl(name)
	{
		return new CWindow(this.ControlGet(name))
	}
	GetControlList()
	{
		return this.Get("ControlList")
	}
	GetDialogWindow()
	{
		WinGet, windows, List, % this.title
		Loop, %windows%
		{
			win := new CWindow(windows%A_Index%)
			if (win.GetClass() = "#32770")
			{
				return win.hWnd
			}
		}
		return ""
	}
	GetParent()
	{
		SetFormat, Integer, h
		parent := DllCall("GetParent", "ptr", this.hWnd)
		SetFormat, Integer, d
		return parent
	}
	GetPID()
	{
		return this.Get("PID")
	}
	GetPixelColor(x, y, bCoordWindow := true)
	{
		if (bCoordWindow)
		{	
			this.WindowToScreen(x, y)
		}
		PixelGetColor, clr, x, y
		return clr
	}
	GetPOINTObject(ptr)
	{
		return new CPoint(NumGet(ptr + 0, 0, "Int"), NumGet(ptr + 0, 4, "Int"))
	}
	GetPos()
	{
		WinGetPos, x, y, w, h, % this.title
		return new CPosition(x, y, w, h)
	}	
	GetRECTObject(ptr)
	{
		return new CRect(NumGet(ptr + 0, 0, "Int"), NumGet(ptr + 0, 4, "Int"), NumGet(ptr + 0, 8, "Int"), NumGet(ptr + 0, 12, "Int"))
	}
	GetProcessName()
	{
		return this.Get("ProcessName")
	}
	GetText()
	{
		nBuffer := this.GetWindowTextLength() + 1
		VarSetCapacity(szText , nBuffer * (A_IsUnicode ? 2 : 1),0)
		length := this.SendMessage(CWU.WM_GETTEXT, nBuffer, &szText)
		return StrGet(&szText, length)
	}
	GetTitle()
	{
		WinGetTitle, var, % this.title
		return var
	}
	GetWindow(uCmd)
	{
		return DllCall("GetWindow", "ptr", this.hWnd, "uint", uCmd, "ptr")
	}
	GetWindowLong(nIndex)
	{
		if (A_PtrSize = 8)
		{
			return DllCall("GetWindowLongPtr", "ptr", this.hWnd, "int", nIndex, "ptr")
		}
		else
		{
			return DllCall("GetWindowLong", "ptr", this.hWnd, "int", nIndex)
		}
	}
	GetWindowRect()
	{
		VarSetCapacity(rect, 16, 0)
		if (DllCall("GetWindowRect", "ptr", this.hWnd, "ptr", &rect))
		{
			return this.GetRECTObject(&rect)
		}
		return 0
	}
	GetWindowTextLength()
	{
		return this.SendMessage(CWU.WM_GETTEXTLENGTH, 0, 0)
	}
	Hide()
	{
		WinHide, % this.title
	}
	HIWORD(dword)
	{
		return (dword >> 16) & 0xffff
	}
	InvalidateRect(rc := 0, bErase := true)
	{
		lpRect := 0
		if (rc)
		{
			VarSetCapacity(rect, 4 * 4, 0)
			NumPut(rc.left, rect, 0, "Int")
			NumPut(rc.top, rect, 4, "Int")
			NumPut(rc.right, rect, 8, "Int")
			NumPut(rc.bottom, rect, 12, "Int")
			lpRect := &rect
		}
		return DllCall("InvalidateRect", "ptr", this.hWnd, "ptr", lpRect, "int", bErase)
	}
	IsAlwaysOnTop()
	{
		if (this.Get("ExStyle") & 0x08)
		{
			return true
		}
		return false
	}
	IsWindow()
	{
		return DllCall("IsWindow", "ptr", this.hWnd)
	}
	IsWindowVisible()
	{
		return DllCall("IsWindowVisible", "ptr", this.hWnd)
	}
	Kill()
	{
		WinKill, % this.title
	}
	LogChildControl(x := 0, y = 0)
	{
		Loop, Parse, % this.GetControlList(), `n
		{
			child := this.GetControl(A_LoopField)
			pos := this.ControlGetPos(A_LoopField)
			l := pos.x
			r := pos.x + pos.w
			t := pos.y
			b := pos.y + pos.h
			if ((x = 0 && y = 0) || (x >= l && x <= r && y >= t && y <= b))
			{
				;this.ControlClick(A_LoopField, x - pos.x, y - pos.y)
				;Sleep, 1000
			}
		}
	}
	LOWORD(dword)
	{
		return dword & 0xffff
	}
	MakeLong(lo, hi)
	{
		return (hi << 16) | (lo & 0xffff)
	}
	Minimize()
	{
		this.SendMessage(CWU.WM_SYSCOMMAND, CWU.SC_MINIMIZE, 0)
	}
	Maximize()
	{
		this.SendMessage(CWU.WM_SYSCOMMAND, CWU.SC_MAXIMIZE, 0)
	}
	Restore()
	{
		this.SendMessage(CWU.WM_SYSCOMMAND, CWU.SC_RESTORE, 0)
	}
	MouseClick(x, y)
	{
		rc := this.GetWindowRect()
		x += rc.left
		y += rc.top
		MouseClick, left, x, y
	}
	MouseMove(x, y)
	{
		rc := this.GetWindowRect()
		return DllCall("SetCursorPos", "int", rc.left + x, "int", rc.top + y)
	}
	Move(x, y, w := "", h := "")
	{
		if (!w || !h)
		{
			WinMove, % this.title, , x, y
		}
		else
		{
			WinMove, % this.title, , x, y, w, h
		}
	}
	MoveWindow(x, y, nWidth, nHeight, bRepaint := true)
	{
		return DllCall("MoveWindow", "ptr", this.hWnd, "int", x, "int", y, "int", nWidth, "int", nHeight, "int", bRepaint)
	}
	PopupWindowExist()
	{
		WinGet, windows, List
		Loop, %windows%
		{
			win := new CWindow(windows%A_Index%)
			if (win.GetClass() = "#32770" && win.GetProcessName() = this.GetProcessName())
			{
				return win.hWnd
			}
		}
		return 0
	}
	PostMessage(uMsg, wParam := 0, lParam :=0)
	{
		return DllCall("PostMessage", "ptr", this.hWnd, "uint", uMsg, "uptr", wParam, "ptr", lParam)
	}
	ReleaseCapture()
	{
		return DllCall("ReleaseCapture")
	}
	Reset()
	{
		this.hWnd := this.Exist()
		this.title := "ahk_id " this.hWnd
	}
	ScreenToClient(x, y)
	{
		VarSetCapacity(pt, 8, 0)
		NumPut(x, pt, 0, "int")
		NumPut(y, pt, 4, "int")
		if (DllCall("ScreenToClient", "ptr", this.hWnd, "ptr", &pt))
		{
			return this.GetPOINTObject(&pt)
		}
		return false
	}
	Send(keys)
	{
		this.Activate()
		Send, %keys%
	}
	SendMessage(uMsg, wParam :=0, lParam := 0)
	{
		return DllCall("SendMessage", "ptr", this.hWnd, "uint", uMsg, "uptr", wParam, "ptr", lParam)
	}
	SetActiveWindow()
	{
		return DllCall("SetActiveWindow", "ptr", this.hWnd, "ptr")
	}
	SetCapture()
	{
		return DllCall("SetCapture", "ptr", this.hWnd, "ptr")
	}
	SetFocus()
	{
		return DllCall("SetFocus", "ptr", this.hWnd, "ptr")
	}
	SetText(szText)
	{
		return this.SendMessage(CWU.WM_SETTEXT, 0, &szText)
	}
	SetTitle(title)
	{
		WinSetTitle, %title%
	}
	SetWindowLong(nIndex, value)
	{
		if (A_PtrSize = 8)
		{
			return DllCall("SetWindowLongPtr", "ptr", this.hWnd, "int", nIndex, "ptr", value)
		}
		else
		{
			return DllCall("SetWindowLong", "ptr", this.hWnd, "int", nIndex, "int", value)
		}
	}
	ShowWindow(bShow := true)
	{
		return DllCall("ShowWindow", "ptr", this.hWnd, "int", bShow)
	}
	SetWindowPos(flags := 3, hWndInsertAfter := 0, x := 0, y := 0, cx := 0, cy := 0)
	{
		return DllCall("SetWindowPos", "ptr", this.hWnd, "ptr", hWndInsertAfter, "int", x, "int", y, "int", cx, "int", cy,  "uint", flags)
	}
	UpdateWindow()
	{
		return DllCall("UpdateWindow", "ptr", this.hWnd)
	}
	WaitControlFocus(name, timeout := 2000)
	{
		tc_start := A_TickCount
		while (1)
		{
			this.ControlClick(name)
			Sleep, 500
			focus := this.ControlGetFocus() 
			if (focus = name)
			{
				return true
			}
			if (A_TickCount - tc_start >= timeout)
			{
				return false
			}
		}
	}
	WaitPopupWindow(timeout := 2000)
	{
		tc_start := A_TickCount
		while (1)
		{
			hWnd := this.PopupWindowExist()
			if (hWnd)
			{
				return hWnd
			}
			Sleep, 100
			if (A_TickCount - tc_start >= timeout)
			{
				return 0
			}
		}
	}
	WindowToScreen(ByRef x, ByRef y)
	{
		pos := this.GetPos()
		x += pos.x
		y += pos.y
	}
	WinWait(title, timeout := 3, text := "")
	{
		WinWait, %title%, %text%, %timeout%
		return ErrorLevel = 0
	}
	;--------------------------------------------------------------
	; WinAPI Window creation support
	Create(hWndParent := 0)
	{
		this.cwex := new CCWEX(this.szClassName, this.szWindowName)
		this.SetCreateWindowInfo()
		return this.CreateWindowEx(this.cwex.dwExStyle, this.cwex.szClassName, this.cwex.szWindowName, this.cwex.dwStyle
			, this.cwex.x, this.cwex.y, this.cwex.nWidth, this.cwex.nHeight, hWndParent, this.cwex.hMenu, this.cwex.hInstance, this.cwex.lpParam)
	}
	CreateWindowEx(dwExStyle, lpClassName,lpWindowName,dwStyle,x,y,nWidth,nHeight,hWndParent,hMenu,hInstance,lpParam)
	{
		CWindow._inst_create := this
		this.hWnd := DllCall("CreateWindowEx",  "Uint",   dwExStyle, "ptr",   &lpClassName, "ptr",   &lpWindowName, "Uint",   dwStyle
		  , "int",   x, "int",   y, "int",   nWidth, "int",   nHeight
		  , "ptr",   hWndParent, "ptr",   hMenu, "ptr",   hInstance, "ptr",   lpParam)	 
		 return this.hWnd
	}
	DefaultWindowProc(hWnd, uMsg, wParam, lParam)
	{
		return DllCall("DefWindowProc", "Ptr", hWnd, "UInt" ,uMsg, "UPtr", wParam, "Ptr",lParam)
	}
	ListenMessage(hWnd)
	{
		Critical, On
		this._inst[hWnd] := this
		Critical, Off
	}
	RegisterClassEx(style, cbClsExtra, cbWndExtra, hInstance, hIcon, hCursor, hbrBackground, lpszMenuName, lpszClassName, hIconSm)
	{
		lpfnWndProc := this._wndproc
		cbSize := VarSetCapacity(wcex, A_PtrSize == 8 ? 80 : 48, 0)
		NumPut(cbSize, wcex, 0, "UInt")
		NumPut(style, wcex, 4, "UInt")
		NumPut(lpfnWndProc, wcex, 8, "Ptr")
		NumPut(cbClsExtra, wcex, A_PtrSize == 8 ? 16 : 12, "Int")
		NumPut(cbWndExtra, wcex, A_PtrSize == 8 ? 20 : 16, "Int")
		NumPut(hInstance, wcex, A_PtrSize == 8 ? 24 : 20, "Ptr")
		NumPut(hIcon, wcex, A_PtrSize == 8 ? 32 : 24, "Ptr")
		NumPut(hCursor, wcex, A_PtrSize == 8 ? 40 : 28, "Ptr")
		NumPut(hbrBackground, wcex, A_PtrSize == 8 ? 48 : 32, "Ptr")
		NumPut(&lpszMenuName, wcex, A_PtrSize == 8 ? 56 : 36, "Ptr")
		NumPut(&lpszClassName, wcex, A_PtrSize == 8 ? 64 : 40, "Ptr")
		NumPut(hIconSm, wcex, A_PtrSize == 8 ? 72 : 44, "Ptr")
		return DllCall("RegisterClassEx", "ptr", &wcex)
	}
	RegisterEx()
	{
		this.wcex := new CWNDCLASSEX(this.szClassName, this.szWindowName)
		this.SetWindowClassInfo()
		return this.RegisterClassEx(this.wcex.style, this.wcex.cbClsExtra, this.wcex.cbWndExtra, this.wcex.hInstance
			, this.wcex.hIcon, this.wcex.hCursor, this.wcex.hbrBackground, this.wcex.szMenuName, this.wcex.szClassName, this.wcex.hIconSm)
	}
	SetCreateWindowInfo()
	{
	}
	SubclassWindow()
	{
		this.ListenMessage(this.hWnd)
		GWL_WNDPROC := -4
		if (A_PtrSize = 8)
		{
			this.pWndProcOld := DllCall("SetWindowLongPtr", "ptr", this.hwnd, "int", GWL_WNDPROC, "Ptr", this._wndproc, "Ptr")
		}
		else
		{
			this.pWndProcOld := DllCall("SetWindowLongPtr", "int", this.hwnd, "int", GWL_WNDPROC, "Ptr", this._wndproc, "Ptr")
		}	
		return this.pWndProcOld
	}	
	WndProc(uMsg, wParam, lParam)
	{
		Critical, On
		hWnd := this
		if (uMsg = CWU.WM_CREATE)
		{
			_this := CWindow._inst_create
			if (_this)
			{
				CWindow._inst_create := ""
				_this.ListenMessage(hWnd)
				return _this.OnCreate(hWnd, uMsg, wParam, lParam)
			}
			return 0
		}
		instance := CWindow._inst[hWnd]
		if (instance)
		{
			if (uMsg = CWU.WM_DESTROY)
			{
				CWindow._inst[hWnd] := ""
			}
			if (instance.OnMessage(hWnd, uMsg, wParam, lParam))
			{
				return 0
			}
			if (instance.pWndProcOld)
			{
				; SubclassWindow
				return DllCall("CallWindowProc", "Ptr" , instance.pWndProcOld, "Ptr", hWnd, "UInt" ,uMsg, "UPtr",wParam, "Ptr",lParam)
			}
			return instance.DefaultWindowProc(hWnd, uMsg, wParam, lParam)
		}
		return CWindow.DefaultWindowProc(hWnd, uMsg, wParam, lParam)
	}
	OnCreate(hWnd, uMsg, wParam, lParam)
	{
		return 0
	}
	OnMessage(hWnd, uMsg, wParam, lParam)
	{
		return 0
	}
	
}

class CMenu
{
	static MF_INSERT := 0x00000000
	static MF_CHANGE := 0x00000080
	static MF_APPEND := 0x00000100
	static MF_DELETE := 0x00000200
	static MF_REMOVE := 0x00001000
	static MF_BYCOMMAND := 0x00000000
	static MF_BYPOSITION := 0x00000400
	static MF_SEPARATOR := 0x00000800
	static MF_ENABLED := 0x00000000
	static MF_GRAYED := 0x00000001
	static MF_DISABLED := 0x00000002
	static MF_UNCHECKED := 0x00000000
	static MF_CHECKED := 0x00000008
	static MF_USECHECKBITMAPS := 0x00000200
	static MF_STRING := 0x00000000
	static MF_BITMAP := 0x00000004
	static MF_OWNERDRAW := 0x00000100
	static MF_POPUP := 0x00000010
	static MF_MENUBARBREAK := 0x00000020
	static MF_MENUBREAK := 0x00000040
	static MF_UNHILITE := 0x00000000
	static MF_HILITE := 0x00000080
	static MF_DEFAULT := 0x00001000
	static MF_SYSMENU := 0x00002000
	static MF_HELP := 0x00004000
	static MF_RIGHTJUSTIFY := 0x00004000
	static MF_MOUSESELECT := 0x00008000
	__New(name)
	{
		this.name := name
	}
	Add(item, sub := "", bEnable := true)
	{
		if (item)
		{
			handler := sub ? ":" sub.name : CMenu.Handler
			Menu, % this.name, Add, %item%, %handler%
			return this.GetLastId()
		}
		else
		{
			Menu, % this.name, Add
		}
	}
	Check(id, bCheck := true)
	{
		hMenu := MenuGetHandle(this.name)
		uCheck := this.MF_BYCOMMAND | bCheck ? this.MF_CHECKED : this.MF_UNCHECKED
		return DllCall("CheckMenuItem", "ptr", hMenu, "uint", id, "uint", uCheck)
	}
	Enable(id, bEnable := true)
	{
		hMenu := MenuGetHandle(this.name)
		uEnable := this.MF_BYCOMMAND | bEnable ? this.MF_ENABLED : this.MF_DISABLED
		return DllCall("EnableMenuItem", "ptr", hMenu, "uint", id, "uint", uEnable)
	}
	EnableAll(bEnable := true)
	{
		count := this.GetMenuItemCount()
		hMenu := MenuGetHandle(this.name)
		Loop, %count%
		{
			pos := A_Index - 1
			uEnable := 0x0400
			uEnable |= bEnable ? 0x00 : 0x01
			DllCall("EnableMenuItem", "ptr", hMenu, "uint", pos, "uint", uEnable)
		}
	}
	GetMenuItemCount()
	{
		return DllCall("GetMenuItemCount", "ptr", MenuGetHandle(this.name))
	}
	GetMenuItemID(pos)
	{
		return DllCall("GetMenuItemID", "ptr", MenuGetHandle(this.name), "int", pos)
	}
	GetLastId()
	{
		return this.GetMenuItemID(this.GetMenuItemCount() - 1)
	}
	GetState(id, flags := 0)
	{
		hMenu := MenuGetHandle(this.name)
		return DllCall("GetMenuState", "ptr", hMenu, "uint", id, "uint", flags)		
	}
	Handler(ItemName, ItemPos, MenuName)
	{
	}
	IsChecked(id)
	{
		return this.GetState(id)  & this.MF_CHECKED
	}
	IsDisabled(id)
	{
		return this.GetState(id)  & this.MF_DISABLED
	}
	IsEnabled(id)
	{
		return ! this.IsDisabled(id)
	}
	IsUnchecked(id)
	{
		return ! this.IsChecked(id)
	}
	Rename(item, name)
	{
		Menu, % this.name, Rename, %item%, %name%
	}
}
class CControl extends CWindow
{
	static WND_CLASS := ""
	static CCS_TOP := 0x00000001
	static CCS_NOMOVEY := 0x00000002
	static CCS_BOTTOM := 0x00000003
	static CCS_NORESIZE := 0x00000004
	static CCS_NOPARENTALIGN := 0x00000008
	static CCS_ADJUSTABLE := 0x00000020
	static CCS_NODIVIDER := 0x00000040
	static CCS_VERT := 0x00000080
	static CCS_LEFT := (CControl.CCS_VERT | CControl.CCS_TOP)
	static CCS_RIGHT := (CControl.CCS_VERT | CControl.CCS_BOTTOM)
	static CCS_NOMOVEX := (CControl.CCS_VERT | CControl.CCS_NOMOVEY)
	
	static LVM_FIRST := 0x1000
	static DTM_FIRST := 0x1000
	static TV_FIRST := 0x1100
	static HDM_FIRST := 0x1200
	static TCM_FIRST := 0x1300
	static PGM_FIRST := 0x1400
	static ECM_FIRST := 0x1500
	static BCM_FIRST := 0x1600
	static CBM_FIRST := 0x1700
	static CCM_FIRST := 0x2000  
	static CCM_LAST := (CControl.CCM_FIRST + 0x200)
	static CCM_SETBKCOLOR := (CControl.CCM_FIRST + 1)
	static CCM_SETCOLORSCHEME := (CControl.CCM_FIRST + 2)
	static CCM_GETCOLORSCHEME := (CControl.CCM_FIRST + 3)
	static CCM_GETDROPTARGET := (CControl.CCM_FIRST + 4)
	static CCM_SETUNICODEFORMAT := (CControl.CCM_FIRST + 5)
	static CCM_GETUNICODEFORMAT := (CControl.CCM_FIRST + 6)
	static CCM_SETVERSION := (CControl.CCM_FIRST + 0x7)
	static CCM_GETVERSION := (CControl.CCM_FIRST + 0x8)
	static CCM_SETNOTIFYWINDOW := (CControl.CCM_FIRST + 0x9)
	static CCM_SETWINDOWTHEME := (CControl.CCM_FIRST + 0xb)
	static CCM_DPISCALE := (CControl.CCM_FIRST + 0xc)
	static INFOTIPSIZE := 1024
	static NM_FIRST := 0
	static NM_LAST := - 99
	static LVN_FIRST := -100
	static LVN_LAST := -199
	static HDN_FIRST := -300
	static HDN_LAST := -399
	static TVN_FIRST := -400
	static TVN_LAST := -499
	static TTN_FIRST := -520
	static TTN_LAST := -549
	static TCN_FIRST := -550
	static TCN_LAST := -580
	static CDN_FIRST := -601
	static CDN_LAST := -699
	static TBN_FIRST := -700
	static TBN_LAST := -720
	static UDN_FIRST := -721
	static UDN_LAST := -729
	static DTN_FIRST := -740
	static DTN_LAST := -745
	static MCN_FIRST := -746
	static MCN_LAST := -752
	static DTN_FIRST2 := -753
	static DTN_LAST2 := -799
	static CBEN_FIRST := -800
	static CBEN_LAST := -830
	static RBN_FIRST := -831
	static RBN_LAST := -859
	static IPN_FIRST := -860
	static IPN_LAST := -879
	static SBN_FIRST := -880
	static SBN_LAST := -899
	static PGN_FIRST := -900
	static PGN_LAST := -950
	static WMN_FIRST := -1000
	static WMN_LAST := -1200
	static BCN_FIRST := -1250
	static BCN_LAST := -1350
	static TRBN_FIRST := -1501
	static TRBN_LAST := -1519
	static EN_FIRST := -1520
	static EN_LAST := -1540
	static NM_OUTOFMEMORY := (CControl.NM_FIRST-1)
	static NM_CLICK := (CControl.NM_FIRST-2)
	static NM_DBLCLK := (CControl.NM_FIRST-3)
	static NM_RETURN := (CControl.NM_FIRST-4)
	static NM_RCLICK := (CControl.NM_FIRST-5)
	static NM_RDBLCLK := (CControl.NM_FIRST-6)
	static NM_SETFOCUS := (CControl.NM_FIRST-7)
	static NM_KILLFOCUS := (CControl.NM_FIRST-8)
	static NM_CUSTOMDRAW := (CControl.NM_FIRST-12)
	static NM_HOVER := (CControl.NM_FIRST-13)
	static NM_NCHITTEST := (CControl.NM_FIRST-14)
	static NM_KEYDOWN := (CControl.NM_FIRST-15)
	static NM_RELEASEDCAPTURE := (CControl.NM_FIRST-16)
	static NM_SETCURSOR := (CControl.NM_FIRST-17)
	static NM_CHAR := (CControl.NM_FIRST-18)
	static NM_TOOLTIPSCREATED := (CControl.NM_FIRST-19)
	static NM_LDOWN := (CControl.NM_FIRST-20)
	static NM_RDOWN := (CControl.NM_FIRST-21)
	static NM_THEMECHANGED := (CControl.NM_FIRST-22)
	static NM_FONTCHANGED := (CControl.NM_FIRST-23)
	static NM_CUSTOMTEXT := (CControl.NM_FIRST-24)
	static NM_TVSTATEIMAGECHANGING := (CControl.NM_FIRST-24)
	static MSGF_COMMCTRL_BEGINDRAG := 0x4200
	static MSGF_COMMCTRL_SIZEHEADER := 0x4201
	static MSGF_COMMCTRL_DRAGSELECT := 0x4202
	static MSGF_COMMCTRL_TOOLBARCUST := 0x4203
	static CDRF_DODEFAULT := 0x00000000
	static CDRF_NEWFONT := 0x00000002
	static CDRF_SKIPDEFAULT := 0x00000004
	static CDRF_DOERASE := 0x00000008
	static CDRF_SKIPPOSTPAINT := 0x00000100
	static CDRF_NOTIFYPOSTPAINT := 0x00000010
	static CDRF_NOTIFYITEMDRAW := 0x00000020
	static CDRF_NOTIFYSUBITEMDRAW := 0x00000020
	static CDRF_NOTIFYPOSTERASE := 0x00000040
	static CDDS_PREPAINT := 0x00000001
	static CDDS_POSTPAINT := 0x00000002
	static CDDS_PREERASE := 0x00000003
	static CDDS_POSTERASE := 0x00000004
	static CDDS_ITEM := 0x00010000
	static CDDS_ITEMPREPAINT := (CControl.CDDS_ITEM | CControl.CDDS_PREPAINT)
	static CDDS_ITEMPOSTPAINT := (CControl.CDDS_ITEM | CControl.CDDS_POSTPAINT)
	static CDDS_ITEMPREERASE := (CControl.CDDS_ITEM | CControl.CDDS_PREERASE)
	static CDDS_ITEMPOSTERASE := (CControl.CDDS_ITEM | CControl.CDDS_POSTERASE)
	static CDDS_SUBITEM := 0x00020000
	static CDIS_SELECTED := 0x0001
	static CDIS_GRAYED := 0x0002
	static CDIS_DISABLED := 0x0004
	static CDIS_CHECKED := 0x0008
	static CDIS_FOCUS := 0x0010
	static CDIS_DEFAULT := 0x0020
	static CDIS_HOT := 0x0040
	static CDIS_MARKED := 0x0080
	static CDIS_INDETERMINATE := 0x0100
	static CDIS_SHOWKEYBOARDCUES := 0x0200
	static CDIS_NEARHOT := 0x0400
	static CDIS_OTHERSIDEHOT := 0x0800
	static CDIS_DROPHILITED := 0x1000
	static ICC_LISTVIEW_CLASSES := 0x00000001
	static ICC_TREEVIEW_CLASSES := 0x00000002
	static ICC_BAR_CLASSES := 0x00000004
	static ICC_TAB_CLASSES := 0x00000008
	static ICC_UPDOWN_CLASS := 0x00000010
	static ICC_PROGRESS_CLASS := 0x00000020
	static ICC_HOTKEY_CLASS := 0x00000040
	static ICC_ANIMATE_CLASS := 0x00000080
	static ICC_WIN95_CLASSES := 0x000000FF
	static ICC_DATE_CLASSES := 0x00000100
	static ICC_USEREX_CLASSES := 0x00000200
	static ICC_COOL_CLASSES := 0x00000400
	static ICC_INTERNET_CLASSES := 0x00000800
	static ICC_PAGESCROLLER_CLASS := 0x00001000
	static ICC_NATIVEFNTCTL_CLASS := 0x00002000
	static ICC_STANDARD_CLASSES := 0x00004000
	static ICC_LINK_CLASS := 0x00008000
	__New(hWnd)
	{
		base.__New(hWnd)
		this.id := this.GetCtrlID()
	}
	Destroy()
	{
		this.SendMessage(CWU.WM_SYSCOMMAND, CWU.SC_CLOSE, 0)
	}
	GetCtrlID()
	{
		return this.GetWindowLong(CWU.GWL_ID)
	}
	SetCtrlID(id)
	{
		this.id := id
		return this.SetWindowLong(CWU.GWL_ID, id)
	}
}
class CButtonCtrl extends CControl
{
	__New(name)
	{
		Gui, Add, Button,  HWNDhWnd, %name%
		base.__New(hWnd)
	}
}
class CStaticCtrl extends CControl
{
	static SS_BITMAP := 0x0000000E
	static STM_SETICON := 0x0170
	static STM_GETICON := 0x0171
	static STM_SETIMAGE := 0x0172
	static STM_GETIMAGE := 0x0173
	__New(dwStyle, dwExStyle := 0)
	{
		Gui, Add, Text, HWNDhWnd %dwStyle% E%dwExStyle%
		base.__New(hWnd)
	}
	SetImage(hBitmap)
	{
		return this.SendMessage(this.STM_SETIMAGE, CWU.IMAGE_BITMAP, hBitmap)
	}
	SetIcon(hIcon)
	{
		return this.SendMessage(this.STM_SETICON, hIcon, 0)
	}
	
}
class CEditCtrl extends CControl
{
	static EM_SCROLL := 0x00B5
	__New(name, bReadOnly := false)
	{
		this.name := name
		ES_MULTILINE := 0x0004, ES_AUTOVSCROLL := 0x0040, ES_READONLY := 0x0800
		style := ES_MULTILINE | ES_AUTOVSCROLL
		if (bReadOnly)
		{
			style |= ES_READONLY
		}
		Gui, Add, Edit,   HWNDhWnd %style% hidden
		base.__New(hWnd)
		EM_SETEXTENDEDSTYLE := CControl.ECM_FIRST + 10
		ES_EX_ALLOWEOL_CR  := 0x0001, ES_EX_ALLOWEOL_LF := 0x0002, ES_EX_ALLOWEOL_ALL := 0x0003
		ES_EX_CONVERT_EOL_ON_PASTE := 0x0004
		dwExStyle := ES_EX_ALLOWEOL_ALL | ES_EX_CONVERT_EOL_ON_PASTE
		this.SendMessage(EM_SETEXTENDEDSTYLE, 0x000F, dwExStyle)
	}
	Clear()
	{
		this.SetSel(0, -1)
		text := ""
		this.ReplaceSel(text)
	}
	SetSel(nStartChar, nEndChar)
	{
		EM_SETSEL := 0xB1
		this.SendMessage(EM_SETSEL, nStartChar, nEndChar)
	}
	ReplaceSel(ByRef text, bCanUndo := false)
	{
		EM_REPLACESEL := 0xC2
		return this.SendMessage(EM_REPLACESEL, bCanUndo, &text)
	}
	InsertText(nInsertAfterChar, ByRef text, bCanUndo := false)
	{
		this.SetSel(nInsertAfterChar, nInsertAfterChar)
		return this.ReplaceSel(text, bCanUndo)
	}	
	AppendText(text, bCanUndo :=false)
	{
		if (A_OsVersion = "WIN_7")
		{
			text := StrReplace(StrReplace(text, "`r`n", "`n"), "`n", "`r`n")
		}
		this.InsertText(this.GetWindowTextLength(), text, bCanUndo)
	}
	GetLimitText()
	{
		EM_GETLIMITTEXT := 0x00D5
		return this.SendMessage(EM_GETLIMITTEXT, 0, 0)
	}
	SetLimitText(nMax)
	{
		EM_SETLIMITTEXT := 0x00C5
		this.SendMessage(EM_SETLIMITTEXT, nMax, 0)
	}
	Scroll(iAction := 3) ; 3 = CScrollBarCtrl.SB_PAGEDOWN
	{
		ret := this.SendMessage(CEditCtrl.EM_SCROLL, iAction, 0)
		if (this.HIWORD(ret))
		{
			return this.LOWORD(ret)
		}
		return 0
	}
	
}
class CScrollBarCtrl extends CControl
{
	static WND_CLASS := "ScrollBar"
	static SBS_HORZ := 0x0000
	static SBS_VERT := 0x0001
	static SB_HORZ := 0
	static SB_VERT := 1
	static SB_CTL := 2
	static SB_BOTH := 3
	static SB_LINEUP := 0
	static SB_LINELEFT := 0
	static SB_LINEDOWN := 1
	static SB_LINERIGHT := 1
	static SB_PAGEUP := 2
	static SB_PAGELEFT := 2
	static SB_PAGEDOWN := 3
	static SB_PAGERIGHT := 3
	static SB_THUMBPOSITION := 4
	static SB_THUMBTRACK := 5
	static SB_TOP := 6
	static SB_LEFT := 6
	static SB_BOTTOM := 7
	static SB_RIGHT := 7
	static SB_ENDSCROLL := 8
	static SBS_TOPALIGN := 0x0002
	static SBS_LEFTALIGN := 0x0002
	static SBS_BOTTOMALIGN := 0x0004
	static SBS_RIGHTALIGN := 0x0004
	static SBS_SIZEBOXTOPLEFTALIGN := 0x0002
	static SBS_SIZEBOXBOTTOMRIGHTALIGN := 0x0004
	static SBS_SIZEBOX := 0x0008
	static SBS_SIZEGRIP := 0x0010
	static SBM_SETPOS := 0x00E0
	static SBM_GETPOS := 0x00E1
	static SBM_SETRANGE := 0x00E2
	static SBM_SETRANGEREDRAW := 0x00E6
	static SBM_GETRANGE := 0x00E3
	static SBM_ENABLE_ARROWS := 0x00E4
	static SBM_SETSCROLLINFO := 0x00E9
	static SBM_GETSCROLLINFO := 0x00EA
	static SBM_GETSCROLLBARINFO := 0x00EB
	static SIF_RANGE := 0x0001
	static SIF_PAGE := 0x0002
	static SIF_POS := 0x0004
	static SIF_DISABLENOSCROLL := 0x0008
	static SIF_TRACKPOS := 0x0010
	static SIF_ALL := (CScrollBarCtrl.SIF_RANGE | CScrollBarCtrl.SIF_PAGE | CScrollBarCtrl.SIF_POS | CScrollBarCtrl.SIF_TRACKPOS)
	
	GetPos()
	{
		return this.SendMessage(this.SBM_GETPOS, 0, 0)
	}
}
class CStatusBarCtrl extends CControl
{
	__New()
	{
		Gui, Add, StatusBar, HWNDhWnd
		base.__New(hWnd)
	}
	SetParts(parts*)
	{
		static SB_SETPARTS := (CWU.WM_USER+4)
		VarSetCapacity(lParam, 4 * parts.Length(), 0)
		Loop, % parts.Length()
		{
			NumPut(parts[A_Index], lParam, 4 * (A_Index - 1))
		}
		return this.SendMessage(SB_SETPARTS, parts.Length(), &lParam)
	}
	SetText(iPart, szText)
	{
		static SB_SETTEXTA := (CWU.WM_USER+1)
		static SB_SETTEXTW := (CWU.WM_USER+11)
		static SB_SETTEXT := A_IsUnicode ? SB_SETTEXTW : SB_SETTEXTA
		iPart &= 0xFF
		return this.SendMessage(SB_SETTEXT, iPart, &szText)
	}
	Size()
	{
		return this.SendMessage(CWU.WM_SIZE, 0, 0)
	}
}

class CToolbarCtrl extends CControl
{
	static WND_CLASS := "ToolbarWindow32"
	static CMB_MASKED := 0x02
	static TBSTATE_CHECKED := 0x01
	static TBSTATE_PRESSED := 0x02
	static TBSTATE_ENABLED := 0x04
	static TBSTATE_HIDDEN := 0x08
	static TBSTATE_INDETERMINATE := 0x10
	static TBSTATE_WRAP := 0x20
	static TBSTATE_ELLIPSES := 0x40
	static TBSTATE_MARKED := 0x80
	static TBSTYLE_BUTTON := 0x0000
	static TBSTYLE_SEP := 0x0001
	static TBSTYLE_CHECK := 0x0002
	static TBSTYLE_GROUP := 0x0004
	static TBSTYLE_CHECKGROUP := CToolBarCtrl.TBSTYLE_GROUP | CToolBarCtrl.TBSTYLE_CHECK
	static TBSTYLE_DROPDOWN := 0x0008
	static TBSTYLE_AUTOSIZE := 0x0010
	static TBSTYLE_NOPREFIX := 0x0020
	static TBSTYLE_TOOLTIPS := 0x0100
	static TBSTYLE_WRAPABLE := 0x0200
	static TBSTYLE_ALTDRAG := 0x0400
	static TBSTYLE_FLAT := 0x0800
	static TBSTYLE_LIST := 0x1000
	static TBSTYLE_CUSTOMERASE := 0x2000
	static TBSTYLE_REGISTERDROP := 0x4000
	static TBSTYLE_TRANSPARENT := 0x8000	
	static TBSTYLE_EX_DRAWDDARROWS := 0x00000001
	static BTNS_BUTTON := CToolBarCtrl.TBSTYLE_BUTTON
	static BTNS_SEP := CToolBarCtrl.TBSTYLE_SEP
	static BTNS_CHECK := CToolBarCtrl.TBSTYLE_CHECK
	static BTNS_GROUP := CToolBarCtrl.TBSTYLE_GROUP
	static BTNS_CHECKGROUP := CToolBarCtrl.TBSTYLE_CHECKGROUP
	static BTNS_DROPDOWN := CToolBarCtrl.TBSTYLE_DROPDOWN
	static BTNS_AUTOSIZE := CToolBarCtrl.TBSTYLE_AUTOSIZE
	static BTNS_NOPREFIX := CToolBarCtrl.TBSTYLE_NOPREFIX
	static BTNS_SHOWTEXT := 0x0040
	static BTNS_WHOLEDROPDOWN := 0x0080
	static TBSTYLE_EX_MIXEDBUTTONS := 0x00000008
	static TBSTYLE_EX_HIDECLIPPEDBUTTONS := 0x00000010
	static TBSTYLE_EX_MULTICOLUMN := 0x00000002
	static TBSTYLE_EX_VERTICAL := 0x00000004
	static TBSTYLE_EX_DOUBLEBUFFER := 0x00000080
	static TB_ADDBUTTONSA := (CWU.WM_USER + 20)
	static TB_INSERTBUTTONA := (CWU.WM_USER + 21)
	static TB_DELETEBUTTON := (CWU.WM_USER + 22)
	static TB_GETBUTTON := (CWU.WM_USER + 23)
	static TB_BUTTONCOUNT := (CWU.WM_USER + 24)
	static TB_COMMANDTOINDEX := (CWU.WM_USER + 25)	
	static TB_SAVERESTOREA := (CWU.WM_USER + 26)
	static TB_SAVERESTOREW := (CWU.WM_USER + 76)
	static TB_CUSTOMIZE := (CWU.WM_USER + 27)
	static TB_ADDSTRINGA := (CWU.WM_USER + 28)
	static TB_ADDSTRINGW := (CWU.WM_USER + 77)
	static TB_GETITEMRECT := (CWU.WM_USER + 29)
	static TB_BUTTONSTRUCTSIZE := (CWU.WM_USER + 30)
	static TB_SETBUTTONSIZE := (CWU.WM_USER + 31)
	static TB_SETBITMAPSIZE := (CWU.WM_USER + 32)
	static TB_AUTOSIZE := (CWU.WM_USER + 33)
	static TB_GETTOOLTIPS := (CWU.WM_USER + 35)
	static TB_SETTOOLTIPS := (CWU.WM_USER + 36)
	static TB_SETPARENT := (CWU.WM_USER + 37)
	static TB_SETROWS := (CWU.WM_USER + 39)
	static TB_GETROWS := (CWU.WM_USER + 40)
	static TB_SETCMDID := (CWU.WM_USER + 42)
	static TB_CHANGEBITMAP := (CWU.WM_USER + 43)
	static TB_GETBITMAP := (CWU.WM_USER + 44)
	static TB_GETBUTTONTEXTA := (CWU.WM_USER + 45)
	static TB_GETBUTTONTEXTW := (CWU.WM_USER + 75)
	static TB_REPLACEBITMAP := (CWU.WM_USER + 46)
	static TB_SETINDENT := (CWU.WM_USER + 47)
	static TB_SETIMAGELIST := (CWU.WM_USER + 48)
	static TB_GETIMAGELIST := (CWU.WM_USER + 49)
	static TB_LOADIMAGES := (CWU.WM_USER + 50)
	static TB_GETRECT := (CWU.WM_USER + 51)
	static TB_SETHOTIMAGELIST := (CWU.WM_USER + 52)
	static TB_GETHOTIMAGELIST := (CWU.WM_USER + 53)
	static TB_SETDISABLEDIMAGELIST := (CWU.WM_USER + 54)
	static TB_GETDISABLEDIMAGELIST := (CWU.WM_USER + 55)
	static TB_SETSTYLE := (CWU.WM_USER + 56)
	static TB_GETSTYLE := (CWU.WM_USER + 57)
	static TB_GETBUTTONSIZE := (CWU.WM_USER + 58)
	static TB_SETBUTTONWIDTH := (CWU.WM_USER + 59)
	static TB_SETMAXTEXTROWS := (CWU.WM_USER + 60)
	static TB_GETTEXTROWS := (CWU.WM_USER + 61)
	static TB_GETOBJECT := (CWU.WM_USER + 62)
	static TB_GETBUTTONINFOW := (CWU.WM_USER + 63)
	static TB_SETBUTTONINFOW := (CWU.WM_USER + 64)
	static TB_GETBUTTONINFOA := (CWU.WM_USER + 65)
	static TB_SETBUTTONINFOA := (CWU.WM_USER + 66)	
	static TB_INSERTBUTTONW := (CWU.WM_USER + 67)
	static TB_ADDBUTTONSW := (CWU.WM_USER + 68)
	static TB_HITTEST := (CWU.WM_USER + 69)	
	static TB_GETHOTITEM := (CWU.WM_USER + 71)
	static TB_SETHOTITEM := (CWU.WM_USER + 72)
	static TB_SETANCHORHIGHLIGHT := (CWU.WM_USER + 73)
	static TB_GETANCHORHIGHLIGHT := (CWU.WM_USER + 74)
	static TB_MAPACCELERATORA := (CWU.WM_USER + 78)
	static TB_GETINSERTMARK := (CWU.WM_USER + 79)
	static TB_SETINSERTMARK := (CWU.WM_USER + 80)
	static TB_INSERTMARKHITTEST := (CWU.WM_USER + 81)
	static TB_MOVEBUTTON := (CWU.WM_USER + 82)
	static TB_GETMAXSIZE := (CWU.WM_USER + 83)
	static TB_SETEXTENDEDSTYLE := (CWU.WM_USER + 84)
	static TB_GETEXTENDEDSTYLE := (CWU.WM_USER + 85)
	static TB_GETPADDING := (CWU.WM_USER + 86)
	static TB_SETPADDING := (CWU.WM_USER + 87)
	static TB_SETINSERTMARKCOLOR := (CWU.WM_USER + 88)
	static TB_GETINSERTMARKCOLOR := (CWU.WM_USER + 89)
	static TB_MAPACCELERATORW := (CWU.WM_USER + 90)
	static TB_SETCOLORSCHEME := CControl.CCM_SETCOLORSCHEME
	static TB_GETCOLORSCHEME := CControl.CCM_GETCOLORSCHEME
	static TB_SETUNICODEFORMAT := CControl.CCM_SETUNICODEFORMAT
	static TB_GETUNICODEFORMAT := CControl.CCM_GETUNICODEFORMAT	
	static TB_GETBUTTONTEXT := A_IsUnicode ? CToolBarCtrl.TB_GETBUTTONTEXTW : CToolBarCtrl.TB_GETBUTTONTEXTA
	static TB_GETBUTTONINFO :=A_IsUnicode ? CToolBarCtrl.TB_GETBUTTONINFOW : CToolBarCtrl.TB_GETBUTTONINFOA
	static TB_SETBUTTONINFO := A_IsUnicode ? CToolBarCtrl.TB_SETBUTTONINFOW : CToolBarCtrl.TB_SETBUTTONINFOA
	static TB_INSERTBUTTON := A_IsUnicode ? CToolBarCtrl.TB_INSERTBUTTONW : CToolBarCtrl.TB_INSERTBUTTONA
	static TB_ADDBUTTONS := A_IsUnicode ? CToolBarCtrl.TB_ADDBUTTONSW : CToolBarCtrl.TB_ADDBUTTONSA
	static TB_TB_SAVERESTORE := A_IsUnicode ? CToolBarCtrl.TB_SAVERESTOREW : CToolBarCtrl.TB_SAVERESTOREA
	static TB_ADDSTRING := A_IsUnicode ? CToolBarCtrl.TB_ADDSTRINGW : CToolBarCtrl.TB_ADDSTRINGA
	static TB_MAPACCELERATOR := A_IsUnicode ? CToolBarCtrl.TB_MAPACCELERATORW : CToolBarCtrl.TB_MAPACCELERATORA
	static TBBF_LARGE := 0x0001
	static TB_GETBITMAPFLAGS := (CWU.WM_USER + 41)
	static TBIF_IMAGE := 0x00000001
	static TBIF_TEXT := 0x00000002
	static TBIF_STATE := 0x00000004
	static TBIF_STYLE := 0x00000008
	static TBIF_LPARAM := 0x00000010
	static TBIF_COMMAND := 0x00000020
	static TBIF_SIZE := 0x00000040
	static TBIF_BYINDEX := 0x80000000	
	static TBCDRF_NOEDGES := 0x00010000
	static TBCDRF_HILITEHOTTRACK := 0x00020000
	static TBCDRF_NOOFFSET := 0x00040000
	static TBCDRF_NOMARK := 0x00080000
	static TBCDRF_NOETCHEDEFFECT := 0x00100000
	static TBCDRF_BLENDICON := 0x00200000
	static TBCDRF_NOBACKGROUND := 0x00400000
	static TBCDRF_USECDCOLORS := 0x00800000
	static TB_ENABLEBUTTON := (CWU.WM_USER + 1)
	static TB_CHECKBUTTON := (CWU.WM_USER + 2)
	static TB_PRESSBUTTON := (CWU.WM_USER + 3)
	static TB_HIDEBUTTON := (CWU.WM_USER + 4)
	static TB_INDETERMINATE := (CWU.WM_USER + 5)
	static TB_MARKBUTTON := (CWU.WM_USER + 6)
	static TB_ISBUTTONENABLED := (CWU.WM_USER + 9)
	static TB_ISBUTTONCHECKED := (CWU.WM_USER + 10)
	static TB_ISBUTTONPRESSED := (CWU.WM_USER + 11)
	static TB_ISBUTTONHIDDEN := (CWU.WM_USER + 12)
	static TB_ISBUTTONINDETERMINATE := (CWU.WM_USER + 13)
	static TB_ISBUTTONHIGHLIGHTED := (CWU.WM_USER + 14)
	static TB_SETSTATE := (CWU.WM_USER + 17)
	static TB_GETSTATE := (CWU.WM_USER + 18)
	static TB_ADDBITMAP := (CWU.WM_USER + 19)
	__New()
	{
		dwStyle := this.TBSTYLE_AUTOSIZE | this.TBSTYLE_WRAPABLE |  this.TBSTYLE_FLAT | this.TBSTYLE_LIST |  this.TBSTYLE_TRANSPARENT | this.BTNS_SHOWTEXT
		Gui, Add, Custom, ClassToolbarWindow32 HWNDhWnd %dwStyle%
		base.__new(hWnd)
	}
	AddButton(iBitmap, idCommand, szText := "", fsState := 0x04, fsStyle := 0)
	{
		iString := szText ? this.AddString(szText) : -1
		VarSetCapacity(TBBUTTON, A_PtrSize == 8 ? 32 : 20, 0)
		NumPut(iBitmap, TBBUTTON, 0, "Int")
		NumPut(idCommand, TBBUTTON, 4, "Int")
		NumPut(fsState, TBBUTTON, 8, "UChar")
		NumPut(fsStyle, TBBUTTON, 9, "UChar")
		;NumPut(bReserved, TBBUTTON, 10, "UChar")
		;NumPut(dwData, TBBUTTON, A_PtrSize == 8 ? 16 : 12, "UPtr")
		NumPut(iString, TBBUTTON, A_PtrSize == 8 ? 24 : 16, "Ptr")
		return this.SendMessage(this.TB_ADDBUTTONS, 1, &TBBUTTON)
	}
	AddString(szText)
	{
		VarSetCapacity(buffer, ((StrPut(szText)+1) * (A_IsUnicode ? 2 : 1), 0))
		StrPut(szText, &buffer, StrPut(szText) * (A_IsUnicode ? 2 : 1))
		return this.SendMessage(this.TB_ADDSTRING, 0, &buffer)
	}
    AutoSize()
    {
        return this.SendMessage(this.TB_AUTOSIZE, 0, 0)
    }
	ButtonStructSize()
	{
		this.SendMessage(this.TB_BUTTONSTRUCTSIZE, A_PtrSize == 8 ? 32 : 20, 0)
	}
	EnableButton(idCommand, bEnable := true)
	{
		return this.SendMessage(this.TB_ENABLEBUTTON, idCommand, bEnable)
	}
 	GetButtonSize()
	{
		return this.SendMessage(this.TB_GETBUTTONSIZE, 0, 0)
	}
	GetButtonText(idCommand)
	{
		VarSetCapacity(buffer , 128 * (A_IsUnicode ? 2 : 1),0)
		length := this.SendMessage(this.TB_GETBUTTONTEXT, idCommand, &buffer)
		return StrGet(&buffer, length)
	}
	SetBitmapSize(width, height)
	{
		return this.SendMessage(this.TB_SETBITMAPSIZE, 0, this.MakeLong(width, height))
	}
	SetButtonSize(width, height)
	{
		return this.SendMessage(this.TB_SETBUTTONSIZE, 0, this.MakeLong(width, height))
	}
	SetButtonText(idCommand, szText)
	{
		cbSize := VarSetCapacity(TBBUTTONINFO, A_PtrSize == 8 ? 48 : 32, 0)
		dwMask := this.TBIF_TEXT
		;fsStyle := this.BTNS_SHOWTEXT
		cchText := StrLen(szText)
		NumPut(cbSize, TBBUTTONINFO, 0, "UInt")
		NumPut(dwMask, TBBUTTONINFO, 4, "UInt")
		NumPut(idCommand, TBBUTTONINFO, 8, "Int")
		;NumPut(iImage, TBBUTTONINFO, 12, "Int")
		;NumPut(fsState, TBBUTTONINFO, 16, "UChar")
		;NumPut(fsStyle, TBBUTTONINFO, 17, "UChar")
		;NumPut(cx, TBBUTTONINFO, 18, "UShort")
		;NumPut(lParam, TBBUTTONINFO, A_PtrSize == 8 ? 24 : 20, "UPtr")
		NumPut(&szText, TBBUTTONINFO, A_PtrSize == 8 ? 32 : 24, "Ptr")
		NumPut(cchText, TBBUTTONINFO, A_PtrSize == 8 ? 40 : 28, "Int")
		return this.SendMessage(this.TB_SETBUTTONINFO, idCommand, &TBBUTTONINFO)
	}
 	SetButtonWidth(min, max)
	{
		return this.SendMessage(this.TB_SETBUTTONWIDTH, 0, this.MakeLong(min, max))
	}
	SetImageList(i, hImageList)
    {
        return this.SendMessage(this.TB_SETIMAGELIST, i, hImageList)
    }
	SetMaxTextRows(MaxRows := 0)
    {
		return this.SendMessage(this.TB_SETMAXTEXTROWS, MaxRows, 0)
    }	
    Size()
    {
        return this.SendMessage(CWU.WM_SIZE, 0, 0)
    }
}
class CToolbarButton
{
	__New(id, text, icono, icofile := "wmploc.dll")
	{
		this.id := id
		this.text := text
		this.icono := icono
		this.icofile := icofile
	}
}
class CTabCtrl extends CControl
{
	static WND_CLASS := "SysTabControl32"
	static TCS_SCROLLOPPOSITE := 0x0001
	static TCS_BOTTOM := 0x0002
	static TCS_RIGHT := 0x0002
	static TCS_MULTISELECT := 0x0004
	static TCS_FLATBUTTONS := 0x0008
	static TCS_FORCEICONLEFT := 0x0010
	static TCS_FORCELABELLEFT := 0x0020
	static TCS_HOTTRACK := 0x0040
	static TCS_VERTICAL := 0x0080
	static TCS_TABS := 0x0000
	static TCS_BUTTONS := 0x0100
	static TCS_SINGLELINE := 0x0000
	static TCS_MULTILINE := 0x0200
	static TCS_RIGHTJUSTIFY := 0x0000
	static TCS_FIXEDWIDTH := 0x0400
	static TCS_RAGGEDRIGHT := 0x0800
	static TCS_FOCUSONBUTTONDOWN := 0x1000
	static TCS_OWNERDRAWFIXED := 0x2000
	static TCS_TOOLTIPS := 0x4000
	static TCS_FOCUSNEVER := 0x8000
	static TCS_EX_FLATSEPARATORS := 0x00000001
	static TCS_EX_REGISTERDROP := 0x00000002
	static TCM_GETIMAGELIST := (CControl.TCM_FIRST + 2)
	static TCM_SETIMAGELIST := (CControl.TCM_FIRST + 3)
	static TCM_GETITEMCOUNT := (CControl.TCM_FIRST + 4)
	static TCIF_TEXT := 0x0001
	static TCIF_IMAGE := 0x0002
	static TCIF_RTLREADING := 0x0004
	static TCIF_PARAM := 0x0008
	static TCIF_STATE := 0x0010
	static TCIS_BUTTONPRESSED := 0x0001
	static TCIS_HIGHLIGHTED := 0x0002
	static TCM_GETITEMA := (CControl.TCM_FIRST + 5)
	static TCM_GETITEMW := (CControl.TCM_FIRST + 60)
	static TCM_SETITEMA := (CControl.TCM_FIRST + 6)
	static TCM_SETITEMW := (CControl.TCM_FIRST + 61)
	static TCM_INSERTITEMA := (CControl.TCM_FIRST + 7)
	static TCM_INSERTITEMW := (CControl.TCM_FIRST + 62)
	static TCM_GETITEM := A_IsUnicode ? CTabCtrl.TCM_GETITEMW : CTabCtrl.TCM_GETITEMA
	static TCM_SETITEM := A_IsUnicode ? CTabCtrl.TCM_SETITEMW : CTabCtrl.TCM_SETITEMA
	static TCM_INSERTITEM := A_IsUnicode ? CTabCtrl.TCM_INSERTITEMW : CTabCtrl.TCM_INSERTITEMA
	static TCM_DELETEITEM := (CControl.TCM_FIRST + 8)
	static TCM_DELETEALLITEMS := (CControl.TCM_FIRST + 9)
	static TCM_GETITEMRECT := (CControl.TCM_FIRST + 10)
	static TCM_GETCURSEL := (CControl.TCM_FIRST + 11)
	static TCM_SETCURSEL := (CControl.TCM_FIRST + 12)
	static TCHT_NOWHERE := 0x0001
	static TCHT_ONITEMICON := 0x0002
	static TCHT_ONITEMLABEL := 0x0004
	static TCHT_ONITEM := (CTabCtrl.TCHT_ONITEMICON | CTabCtrl.TCHT_ONITEMLABEL)
	static TCM_HITTEST := (CControl.TCM_FIRST + 13)
	static TCM_SETITEMEXTRA := (CControl.TCM_FIRST + 14)
	static TCM_ADJUSTRECT := (CControl.TCM_FIRST + 40)
	static TCM_SETITEMSIZE := (CControl.TCM_FIRST + 41)
	static TCM_REMOVEIMAGE := (CControl.TCM_FIRST + 42)
	static TCM_SETPADDING := (CControl.TCM_FIRST + 43)
	static TCM_GETROWCOUNT := (CControl.TCM_FIRST + 44)
	static TCM_GETTOOLTIPS := (CControl.TCM_FIRST + 45)
	static TCM_SETTOOLTIPS := (CControl.TCM_FIRST + 46)
	static TCM_GETCURFOCUS := (CControl.TCM_FIRST + 47)
	static TCM_SETCURFOCUS := (CControl.TCM_FIRST + 48)
	static TCM_SETMINTABWIDTH := (CControl.TCM_FIRST + 49)
	static TCM_DESELECTALL := (CControl.TCM_FIRST + 50)
	static TCM_HIGHLIGHTITEM := (CControl.TCM_FIRST + 51)
	static TCM_SETEXTENDEDSTYLE := (CControl.TCM_FIRST + 52)
	static TCM_GETEXTENDEDSTYLE := (CControl.TCM_FIRST + 53)
	static TCM_SETUNICODEFORMAT := CControl.CCM_SETUNICODEFORMAT
	static TCM_GETUNICODEFORMAT := CControl.CCM_GETUNICODEFORMAT
	static TCN_KEYDOWN := (CControl.TCN_FIRST - 0)
	static TCN_SELCHANGE := (CControl.TCN_FIRST - 1)
	static TCN_SELCHANGING := (CControl.TCN_FIRST - 2)
	static TCN_GETOBJECT := (CControl.TCN_FIRST - 3)
	static TCN_FOCUSCHANGE := (CControl.TCN_FIRST - 4)
	__New(dwStyle := 0, dwExStyle := 0)
	{
		Gui, Add, Custom, ClassSysTabControl32 HWNDhWnd %dwStyle% E%dwExStyle% hidden
		base.__New(hWnd)
	}
	AdjustRect(ByRef rc, bLarger := false)
	{
		VarSetCapacity(prc, 4 * 4, 0)
		NumPut(rc.left, prc, 0, "Int")
		NumPut(rc.top, prc, 4, "Int")
		NumPut(rc.right, prc, 8, "Int")
		NumPut(rc.bottom, prc, 12, "Int")
		this.SendMessage(this.TCM_ADJUSTRECT, bLarger, &prc)
		rc := this.GetRECTObject(&prc)
	}
	DeleteItem(iItem)
	{
		return this.SendMessage(this.TCM_DELETEITEM, iItem, 0)
	}
	GetCurSel()
	{
		return this.SendMessage(this.TCM_GETCURSEL, 0, 0)
	}
	HitTest(x, y, flags)
	{
		VarSetCapacity(TCHITTESTINFO, 12, 0)
		NumPut(x, TCHITTESTINFO, 0, "Int")
		NumPut(y, TCHITTESTINFO, 4, "Int")
		NumPut(flags, TCHITTESTINFO, 8, "UInt")
		return this.SendMessage(this.TCM_HITTEST, 0, &TCHITTESTINFO)
	}
	HighlightItem(iItem, fHighlight := 1)
	{
		return this.SendMessage(this.TCM_HIGHLIGHTITEM, iItem, fHighlight)
	}
	InsertItem(iItem, str)
	{
		VarSetCapacity(pitem, A_PtrSize = 8 ? 40 : 28, 0)
		VarSetCapacity(pstr , StrPut(str)*(A_IsUnicode ? 2:1),0)
		StrPut(str, &pstr, StrPut(str) * (A_IsUnicode ? 2 : 1))
		NumPut(this.TCIF_TEXT,pitem,0,"UInt")                   
		NumPut(&pstr,  pitem,A_PtrSize=8 ? 16:12,"Ptr") 
		NumPut(StrPut(str), pitem,A_PtrSize=8 ? 24:16,"Int")
		return this.SendMessage(this.TCM_INSERTITEM, iItem, &pitem)
	}
	SetCurFocus(iItem)
	{
		this.SendMessage(this.TCM_SETCURFOCUS, iItem, 0)
	}
	SetCurSel(iItem)
	{
		this.SendMessage(this.TCM_SETCURSEL, iItem, 0)		
	}
}
class CBitmapView extends CStaticCtrl
{
	__New(name)
	{
		this.name := name
		base.__New(CStaticCtrl.SS_BITMAP)
	}
}
class CTabView extends CTabCtrl
{
	__New(dwStyle := 0, dwExStyle := 0)
	{
		base.__New(dwStyle, dwExStyle)
		this.children := Array()
	}
	Add(child)
	{
		i := this.children.Length()
		this.InsertItem(i, child.name)
		this.children.Push(child)
		return this.children.Length()
	}
	Exist(name)
	{
		Loop, % this.children.Length()
		{
			if (this.children[A_Index].name = name)
			{
				return A_Index
			}
		}
		return 0
	}
	GetCurSelName()
	{
		i := this.GetCurSel() + 1
		return this.children[i].name
	}
	GetCurChild()
	{
		return this.children[this.GetCurSel() + 1]
	}
	GetCurChildId()
	{
		return this.children[this.GetCurSel() + 1].id
	}
	Move(x, y, w, h)
	{
		base.Move(x, y, w, h)
		rc := new CRect(x, y, x+ w, y +h)
		this.AdjustRect(rc)
		Loop, % this.children.MaxIndex()
		{
			if (this.children[A_Index].hWnd)
			{
				this.children[A_Index].Move(rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top)
			}
		}
	}
	RemoveAt(iItem)
	{
		child := this.children.RemoveAt(iItem + 1)
		if (child)
		{
			this.DeleteItem(iItem)
			return child
		}
		return ""
	}
	Select(name)
	{
		Loop, % this.children.MaxIndex()
		{
			if (name = this.children[A_Index].name)
			{
				this.SetCurSel(A_Index - 1)
				return this.children[A_Index]
			}
		}
	}
	SetCurSel(iItem)
	{
		base.SetCurSel(iItem)
		if (this.IsWindowVisible())
		{
			this.ShowChild()
		}
	}
	ShowWindow(bShow := true)
	{
		base.ShowWindow(bShow)
		this.ShowChild(bShow)
	}
	ShowChild(bShow := true)
	{
		iSel := this.GetCurSel()
		Loop, % this.children.MaxIndex()
		{
			if (A_Index != iSel + 1)
			{
				this.children[A_Index].ShowWindow(false)
			}
		}
		this.children[iSel + 1].ShowWindow(bShow)
		if (bShow)
		{
			this.children[iSel + 1].SetFocus()
		}
	}
}
class CImageList
{
	static CLR_NONE := 0xFFFFFFFF
	static CLR_DEFAULT := 0xFF000000
	static ILC_MASK := 0x00000001
	static ILC_COLOR := 0x00000000
	static ILC_COLORDDB := 0x000000FE
	static ILC_COLOR4 := 0x00000004
	static ILC_COLOR8 := 0x00000008
	static ILC_COLOR16 := 0x00000010
	static ILC_COLOR24 := 0x00000018
	static ILC_COLOR32 := 0x00000020
	static ILC_PALETTE := 0x00000800
	static ILC_MIRROR := 0x00002000
	static ILC_PERITEMMIRROR := 0x00008000
	static ILC_ORIGINALSIZE := 0x00010000
	static ILC_HIGHQUALITYSCALE := 0x00020000
	static ILD_NORMAL := 0x00000000
	static ILD_TRANSPARENT := 0x00000001
	static ILD_MASK := 0x00000010
	static ILD_IMAGE := 0x00000020
	static ILD_ROP := 0x00000040
	static ILD_BLEND25 := 0x00000002
	static ILD_BLEND50 := 0x00000004
	static ILD_OVERLAYMASK := 0x00000F00
	static ILD_PRESERVEALPHA := 0x00001000
	static ILD_SCALE := 0x00002000
	static ILD_DPISCALE := 0x00004000
	static ILD_ASYNC := 0x00008000
	static ILD_SELECTED := CImageList.ILD_BLEND50
	static ILD_FOCUS := CImageList.ILD_BLEND25
	static ILD_BLEND := CImageList.ILD_BLEND50
	static CLR_HILIGHT := CImageList.CLR_DEFAULT
	static ILS_NORMAL := 0x00000000
	static ILS_GLOW := 0x00000001
	static ILS_SHADOW := 0x00000002
	static ILS_SATURATE := 0x00000004
	static ILS_ALPHA := 0x00000008
	static ILGT_NORMAL := 0x00000000
	static ILGT_ASYNC := 0x00000001
	static ILCF_MOVE := (0x00000000)
	static ILCF_SWAP := (0x00000001)
	static ILP_NORMAL := 0
	static ILP_DOWNLEVEL := 1	
	__New(himl := "")
	{
		this.himl := himl
	}
	Add(hbmImage, hbmMask)
	{
		return DllCall("ImageList_Add", "ptr", this.himl, "ptr", hbmImage, "ptr", hbmMask)
	}
	AddIconFromModule(lpszModuleName, index, bLarge := false)
	{
		nIcons := DllCall("shell32\ExtractIconEx" (A_IsUnicode ? "W" : "A"), "str", lpszModuleName, "int", index - 1, "ptr*", hIconLarge := 0, "ptr*", hIconSmall := 0, "uint", 1, "uint")
		if (bLarge && hIconLarge)
		{
			DllCall("DestroyIcon", "ptr", hIconSmall)
			return this.AddIcon(hIconLarge)
		}
		else if (hIconSmall)
		{
			DllCall("DestroyIcon", "ptr", hIconLarge)
			return this.AddIcon(hIconSmall)
		}
		return false
	}
	AddIcon(hIcon)
	{
		return this.ReplaceIcon(-1, hIcon)
	}
	AddMask(hbmImage, crMask)
	{
		return DllCall("ImageList_AddMasked", "ptr", this.himl, "ptr", hbmImage, "uint", crMask)
	}
	BeginDrag(iTrack, dxHotspot, dyHotspot)
	{
		return DllCall("ImageList_BeginDrag", "ptr", this.himl, "int", iTrack, "int", dxHotspot, "int", dyHotspot)
	}
	Copy(himlDst, iDst, iSrc, uFlags)
	{
		return DllCall("ImageList_Copy", "ptr", himlDst, "int", iDst, "ptr", this.himl, "int", iSrc, "uint", uFlags)
	}
	Create(cx, cy, flags, cInitial, cGrow)
	{
		this.cx := cx
		this.cy := cy
		this.himl := DllCall("ImageList_Create", "int", cx, "int", cy, "uint", flags, "int", cInitial, "int", cGrow, "ptr")
		return (this.himl) ? true : false
	}
	DragEnter(hwndLock, x, y)
	{
		return DllCall("ImageList_DragEnter", "ptr", hwndLock, "int", x, "int", y)
	}
	DragLeave(hwndLock)
	{
		return DllCall("ImageList_DragLeave", "ptr", hwndLock)
	}
	DragMove(x, y)
	{
		return DllCall("ImageList_DragMove", "int", x, "int", y)
	}
	DragShowNolock(bShow)
	{
		return DllCall("ImageList_DragShowNolock", "int", bShow)
	}
	Destroy()
	{
		return DllCall("ImageList_Destroy", "ptr", this.himl)
	}
	GetImageCount()
	{
		return DllCall("ImageList_GetImageCount", "ptr", this.himl)
	}
	EndDrag()
	{
		DllCall("ImageList_EndDrag")
	}
	ReplaceIcon(i, hIcon)
	{
		return DllCall("ImageList_ReplaceIcon", "ptr", this.himl, "int", i, "ptr", hIcon)
	}
}
class CTreeViewCtrl extends CControl
{
	static WND_CLASS := "SysTreeView32"
	static TVS_HASBUTTONS := 0x0001
	static TVS_HASLINES := 0x0002
	static TVS_LINESATROOT := 0x0004
	static TVS_EDITLABELS := 0x0008
	static TVS_DISABLEDRAGDROP := 0x0010
	static TVS_SHOWSELALWAYS := 0x0020
	static TVS_RTLREADING := 0x0040
	static TVS_NOTOOLTIPS := 0x0080
	static TVS_CHECKBOXES := 0x0100
	static TVS_TRACKSELECT := 0x0200
	static TVS_SINGLEEXPAND := 0x0400
	static TVS_INFOTIP := 0x0800
	static TVS_FULLROWSELECT := 0x1000
	static TVS_NOSCROLL := 0x2000
	static TVS_NONEVENHEIGHT := 0x4000
	static TVS_NOHSCROLL := 0x8000
	static TVS_EX_NOSINGLECOLLAPSE := 0x0001
	static TVS_EX_MULTISELECT := 0x0002
	static TVS_EX_DOUBLEBUFFER := 0x0004
	static TVS_EX_NOINDENTSTATE := 0x0008
	static TVS_EX_RICHTOOLTIP := 0x0010
	static TVS_EX_AUTOHSCROLL := 0x0020
	static TVS_EX_FADEINOUTEXPANDOS := 0x0040
	static TVS_EX_PARTIALCHECKBOXES := 0x0080
	static TVS_EX_EXCLUSIONCHECKBOXES := 0x0100
	static TVS_EX_DIMMEDCHECKBOXES := 0x0200
	static TVS_EX_DRAWIMAGEASYNC := 0x0400
	static TVIF_TEXT := 0x0001
	static TVIF_IMAGE := 0x0002
	static TVIF_PARAM := 0x0004
	static TVIF_STATE := 0x0008
	static TVIF_HANDLE := 0x0010
	static TVIF_SELECTEDIMAGE := 0x0020
	static TVIF_CHILDREN := 0x0040
	static TVIF_INTEGRAL := 0x0080
	static TVIF_STATEEX := 0x0100
	static TVIF_EXPANDEDIMAGE := 0x0200
	static TVIS_SELECTED := 0x0002
	static TVIS_CUT := 0x0004
	static TVIS_DROPHILITED := 0x0008
	static TVIS_BOLD := 0x0010
	static TVIS_EXPANDED := 0x0020
	static TVIS_EXPANDEDONCE := 0x0040
	static TVIS_EXPANDPARTIAL := 0x0080
	static TVIS_OVERLAYMASK := 0x0F00
	static TVIS_STATEIMAGEMASK := 0xF000
	static TVIS_USERMASK := 0xF000
	static TVIS_EX_FLAT := 0x0001
	static TVIS_EX_DISABLED := 0x0002
	static TVIS_EX_ALL := 0x0002
	static I_CHILDRENCALLBACK := (-1)
	static I_CHILDRENAUTO := (-2)
	static TVI_ROOT := (-0x10000)
	static TVI_FIRST := (-0x0FFFF)
	static TVI_LAST := (-0x0FFFE)
	static TVI_SORT := (-0x0FFFD)
	static TVM_INSERTITEMA := (CControl.TV_FIRST + 0)
	static TVM_INSERTITEMW := (CControl.TV_FIRST + 50)
	static TVM_DELETEITEM := (CControl.TV_FIRST + 1)
	static TVM_EXPAND := (CControl.TV_FIRST + 2)
	static TVE_COLLAPSE := 0x0001
	static TVE_EXPAND := 0x0002
	static TVE_TOGGLE := 0x0003
	static TVE_EXPANDPARTIAL := 0x4000
	static TVE_COLLAPSERESET := 0x8000
	static TVM_GETITEMRECT := (CControl.TV_FIRST + 4)
	static TVM_GETCOUNT := (CControl.TV_FIRST + 5)
	static TVM_GETINDENT := (CControl.TV_FIRST + 6)
	static TVM_SETINDENT := (CControl.TV_FIRST + 7)
	static TVM_GETIMAGELIST := (CControl.TV_FIRST + 8)
	static TVSIL_NORMAL := 0
	static TVSIL_STATE := 2
	static TVM_SETIMAGELIST := (CControl.TV_FIRST + 9)
	static TVM_GETNEXTITEM := (CControl.TV_FIRST + 10)
	static TVGN_ROOT := 0x0000
	static TVGN_NEXT := 0x0001
	static TVGN_PREVIOUS := 0x0002
	static TVGN_PARENT := 0x0003
	static TVGN_CHILD := 0x0004
	static TVGN_FIRSTVISIBLE := 0x0005
	static TVGN_NEXTVISIBLE := 0x0006
	static TVGN_PREVIOUSVISIBLE := 0x0007
	static TVGN_DROPHILITE := 0x0008
	static TVGN_CARET := 0x0009
	static TVGN_LASTVISIBLE := 0x000A
	static TVGN_NEXTSELECTED := 0x000B
	static TVSI_NOSINGLEEXPAND := 0x8000
	static TVM_SELECTITEM := (CControl.TV_FIRST + 11)
	static TVM_GETITEMA := (CControl.TV_FIRST + 12)
	static TVM_GETITEMW := (CControl.TV_FIRST + 62)
	static TVM_SETITEMA := (CControl.TV_FIRST + 13)
	static TVM_SETITEMW := (CControl.TV_FIRST + 63)
	static TVM_EDITLABELA := (CControl.TV_FIRST + 14)
	static TVM_EDITLABELW := (CControl.TV_FIRST + 65)
	static TVM_GETEDITCONTROL := (CControl.TV_FIRST + 15)
	static TVM_GETVISIBLECOUNT := (CControl.TV_FIRST + 16)
	static TVM_HITTEST := (CControl.TV_FIRST + 17)
	static TVHT_NOWHERE := 0x0001
	static TVHT_ONITEMICON := 0x0002
	static TVHT_ONITEMLABEL := 0x0004
	static TVHT_ONITEM := (CTreeViewCtrl.TVHT_ONITEMICON | CTreeViewCtrl.TVHT_ONITEMLABEL | CTreeViewCtrl.TVHT_ONITEMSTATEICON)
	static TVHT_ONITEMINDENT := 0x0008
	static TVHT_ONITEMBUTTON := 0x0010
	static TVHT_ONITEMRIGHT := 0x0020
	static TVHT_ONITEMSTATEICON := 0x0040
	static TVHT_ABOVE := 0x0100
	static TVHT_BELOW := 0x0200
	static TVHT_TORIGHT := 0x0400
	static TVHT_TOLEFT := 0x0800
	static TVM_CREATEDRAGIMAGE := (CControl.TV_FIRST + 18)
	static TVM_SORTCHILDREN := (CControl.TV_FIRST + 19)
	static TVM_ENSUREVISIBLE := (CControl.TV_FIRST + 20)
	static TVM_SORTCHILDRENCB := (CControl.TV_FIRST + 21)
	static TVM_ENDEDITLABELNOW := (CControl.TV_FIRST + 22)
	static TVM_GETISEARCHSTRINGA := (CControl.TV_FIRST + 23)
	static TVM_GETISEARCHSTRINGW := (CControl.TV_FIRST + 64)
	static TVM_SETTOOLTIPS := (CControl.TV_FIRST + 24)
	static TVM_GETTOOLTIPS := (CControl.TV_FIRST + 25)
	static TVM_SETINSERTMARK := (CControl.TV_FIRST + 26)
	static TVM_SETUNICODEFORMAT := CControl.CCM_SETUNICODEFORMAT
	static TVM_GETUNICODEFORMAT := CControl.CCM_GETUNICODEFORMAT
	static TVM_SETITEMHEIGHT := (CControl.TV_FIRST + 27)
	static TVM_GETITEMHEIGHT := (CControl.TV_FIRST + 28)
	static TVM_SETBKCOLOR := (CControl.TV_FIRST + 29)
	static TVM_SETTEXTCOLOR := (CControl.TV_FIRST + 30)
	static TVM_GETBKCOLOR := (CControl.TV_FIRST + 31)
	static TVM_GETTEXTCOLOR := (CControl.TV_FIRST + 32)
	static TVM_SETSCROLLTIME := (CControl.TV_FIRST + 33)
	static TVM_GETSCROLLTIME := (CControl.TV_FIRST + 34)
	static TVM_SETINSERTMARKCOLOR := (CControl.TV_FIRST + 37)
	static TVM_GETINSERTMARKCOLOR := (CControl.TV_FIRST + 38)
	static TVM_SETBORDER := (CControl.TV_FIRST + 35)
	static TVSBF_XBORDER := 0x00000001
	static TVSBF_YBORDER := 0x00000002
	static TVM_GETITEMSTATE := (CControl.TV_FIRST + 39)
	static TVM_SETLINECOLOR := (CControl.TV_FIRST + 40)
	static TVM_GETLINECOLOR := (CControl.TV_FIRST + 41)
	static TVM_MAPACCIDTOHTREEITEM := (CControl.TV_FIRST + 42)
	static TVM_MAPHTREEITEMTOACCID := (CControl.TV_FIRST + 43)
	static TVM_SETEXTENDEDSTYLE := (CControl.TV_FIRST + 44)
	static TVM_GETEXTENDEDSTYLE := (CControl.TV_FIRST + 45)
	static TVM_SETAUTOSCROLLINFO := (CControl.TV_FIRST + 59)
	static TVM_SETHOT := (CControl.TV_FIRST + 58)
	static TVM_GETSELECTEDCOUNT := (CControl.TV_FIRST + 70)
	static TVM_SHOWINFOTIP := (CControl.TV_FIRST + 71)
	static TVM_GETITEMPARTRECT := (CControl.TV_FIRST + 72)
	static TVN_SELCHANGINGA := (CControl.TVN_FIRST-1)
	static TVN_SELCHANGINGW := (CControl.TVN_FIRST-50)
	static TVN_SELCHANGEDA := (CControl.TVN_FIRST-2)
	static TVN_SELCHANGEDW := (CControl.TVN_FIRST-51)
	static TVC_UNKNOWN := 0x0000
	static TVC_BYMOUSE := 0x0001
	static TVC_BYKEYBOARD := 0x0002
	static TVN_GETDISPINFOA := (CControl.TVN_FIRST-3)
	static TVN_GETDISPINFOW := (CControl.TVN_FIRST-52)
	static TVN_SETDISPINFOA := (CControl.TVN_FIRST-4)
	static TVN_SETDISPINFOW := (CControl.TVN_FIRST-53)
	static TVIF_DI_SETITEM := 0x1000
	static TVN_ITEMEXPANDINGA := (CControl.TVN_FIRST-5)
	static TVN_ITEMEXPANDINGW := (CControl.TVN_FIRST-54)
	static TVN_ITEMEXPANDEDA := (CControl.TVN_FIRST-6)
	static TVN_ITEMEXPANDEDW := (CControl.TVN_FIRST-55)
	static TVN_BEGINDRAGA := (CControl.TVN_FIRST-7)
	static TVN_BEGINDRAGW := (CControl.TVN_FIRST-56)
	static TVN_BEGINRDRAGA := (CControl.TVN_FIRST-8)
	static TVN_BEGINRDRAGW := (CControl.TVN_FIRST-57)
	static TVN_DELETEITEMA := (CControl.TVN_FIRST-9)
	static TVN_DELETEITEMW := (CControl.TVN_FIRST-58)
	static TVN_BEGINLABELEDITA := (CControl.TVN_FIRST-10)
	static TVN_BEGINLABELEDITW := (CControl.TVN_FIRST-59)
	static TVN_ENDLABELEDITA := (CControl.TVN_FIRST-11)
	static TVN_ENDLABELEDITW := (CControl.TVN_FIRST-60)
	static TVN_KEYDOWN := (CControl.TVN_FIRST-12)
	static TVN_GETINFOTIPA := (CControl.TVN_FIRST-13)
	static TVN_GETINFOTIPW := (CControl.TVN_FIRST-14)
	static TVN_SINGLEEXPAND := (CControl.TVN_FIRST-15)
	static TVNRET_DEFAULT := 0
	static TVNRET_SKIPOLD := 1
	static TVNRET_SKIPNEW := 2
	static TVN_ITEMCHANGINGA := (CControl.TVN_FIRST-16)
	static TVN_ITEMCHANGINGW := (CControl.TVN_FIRST-17)
	static TVN_ITEMCHANGEDA := (CControl.TVN_FIRST-18)
	static TVN_ITEMCHANGEDW := (CControl.TVN_FIRST-19)
	static TVN_ASYNCDRAW := (CControl.TVN_FIRST-20)
	static TVCDRF_NOIMAGES := 0x00010000
	static TVM_INSERTITEM := A_IsUnicode ? CTreeViewCtrl.TVM_INSERTITEMW : CTreeViewCtrl.TVM_INSERTITEMA
	static TVM_GETITEM := A_IsUnicode ? CTreeViewCtrl.TVM_GETITEMW : CTreeViewCtrl.TVM_GETITEMA
	static TVM_SETITEM := A_IsUnicode ? CTreeViewCtrl.TVM_SETITEMW : CTreeViewCtrl.TVM_SETITEMA
	static TVM_EDITLABEL := A_IsUnicode ? CTreeViewCtrl.TVM_EDITLABELW : CTreeViewCtrl.TVM_EDITLABELA
	static TVM_GETISEARCHSTRING := A_IsUnicode ? CTreeViewCtrl.TVM_GETISEARCHSTRINGW : CTreeViewCtrl.TVM_GETISEARCHSTRINGA
	static TVN_SELCHANGING := A_IsUnicode ? CTreeViewCtrl.TVN_SELCHANGINGW : CTreeViewCtrl.TVN_SELCHANGINGA
	static TVN_SELCHANGED := A_IsUnicode ? CTreeViewCtrl.TVN_SELCHANGEDW : CTreeViewCtrl.TVN_SELCHANGEDA
	static TVN_GETDISPINFO := A_IsUnicode ? CTreeViewCtrl.TVN_GETDISPINFOW : CTreeViewCtrl.TVN_GETDISPINFOA
	static TVN_SETDISPINFO := A_IsUnicode ? CTreeViewCtrl.TVN_SETDISPINFOW : CTreeViewCtrl.TVN_SETDISPINFOA
	static TVN_ITEMEXPANDING := A_IsUnicode ? CTreeViewCtrl.TVN_ITEMEXPANDINGW : CTreeViewCtrl.TVN_ITEMEXPANDINGA
	static TVN_ITEMEXPANDED := A_IsUnicode ? CTreeViewCtrl.TVN_ITEMEXPANDEDW : CTreeViewCtrl.TVN_ITEMEXPANDEDA
	static TVN_BEGINDRAG := A_IsUnicode ? CTreeViewCtrl.TVN_BEGINDRAGW : CTreeViewCtrl.TVN_BEGINDRAGA
	static TVN_BEGINRDRAG := A_IsUnicode ? CTreeViewCtrl.TVN_BEGINRDRAGW : CTreeViewCtrl.TVN_BEGINRDRAGA
	static TVN_DELETEITEM := A_IsUnicode ? CTreeViewCtrl.TVN_DELETEITEMW : CTreeViewCtrl.TVN_DELETEITEMA
	static TVN_BEGINLABELEDIT := A_IsUnicode ? CTreeViewCtrl.TVN_BEGINLABELEDITW : CTreeViewCtrl.TVN_BEGINLABELEDITA
	static TVN_ENDLABELEDIT := A_IsUnicode ? CTreeViewCtrl.TVN_ENDLABELEDITW : CTreeViewCtrl.TVN_ENDLABELEDITA
	static TVN_GETINFOTIP := A_IsUnicode ? CTreeViewCtrl.TVN_GETINFOTIPW : CTreeViewCtrl.TVN_GETINFOTIPA
	static TVN_ITEMCHANGING := A_IsUnicode ? CTreeViewCtrl.TVN_ITEMCHANGINGW : CTreeViewCtrl.TVN_ITEMCHANGINGA
	static TVN_ITEMCHANGED := A_IsUnicode ? CTreeViewCtrl.TVN_ITEMCHANGEDW : CTreeViewCtrl.TVN_ITEMCHANGEDA
	
	__New(dwStyle, dwExStyle := 0)
	{
		Gui, Add, TreeView, HWNDhWnd %dwStyle% E%dwExStyle% hidden
		base.__New(hWnd)
	}
	CreateDragImage(hItem)
	{
		return this.SendMessage(this.TVM_CREATEDRAGIMAGE, 0, hItem)
	}
	EnsureVisible(hItem)
	{
		return this.SendMessage(this.TVM_ENSUREVISIBLE, 0, hItem)
	}
	GetItemRect(hItem, bOnlyText)
	{
		VarSetCapacity(rc, 16, 0)
		NumPut(hItem, rc, 0, "ptr")
		if (this.SendMessage(this.TVM_GETITEMRECT, bOnlyText, &rc))
		{
			return this.GetRECTObject(&rc)
		}
		return 0
	}
	GetDropHiLight()
	{
		return this.GetNextItem(0, this.TVGN_DROPHILITE)
	}
	GetNextItem(hItem, flags)
	{
		return this.SendMessage(this.TVM_GETNEXTITEM , flags, hItem)
	}
	GetNMTREEVIEWObject(ptr)
	{
		o := Object()
		o.hdr := this.GetNMHDRObject(ptr)
		o.action := NumGet(ptr + 0, A_PtrSize == 8 ? 24 : 12, "UInt")
		o.itemOld := this.GetTVITEMObject(ptr + (A_PtrSize == 8 ? 32 : 16))
		o.itemNew := this.GetTVITEMObject(ptr + (A_PtrSize == 8 ? 88 : 56)) 
		o.ptDrag := this.GetPOINTObject(ptr + (A_PtrSize == 8 ? 144 : 96)) 
		return o
	}
	GetItemText(hItem, cchTextMax := 256)
	{
		VarSetCapacity(pszText , (cchTextMax + 1) * (A_IsUnicode ? 2 : 1),0)
		mask := this.TVIF_HANDLE | this.TVIF_TEXT
		VarSetCapacity(TVITEM, A_PtrSize == 8 ? 56 : 40, 0)
		NumPut(mask, TVITEM, 0, "UInt")
		NumPut(hItem, TVITEM, A_PtrSize == 8 ? 8 : 4, "Ptr")
		;NumPut(state, TVITEM, A_PtrSize == 8 ? 16 : 8, "UInt")
		;NumPut(stateMask, TVITEM, A_PtrSize == 8 ? 20 : 12, "UInt")
		NumPut(&pszText, TVITEM, A_PtrSize == 8 ? 24 : 16, "Ptr")
		NumPut(cchTextMax, TVITEM, A_PtrSize == 8 ? 32 : 20, "Int")
		;NumPut(iImage, TVITEM, A_PtrSize == 8 ? 36 : 24, "Int")
		;NumPut(iSelectedImage, TVITEM, A_PtrSize == 8 ? 40 : 28, "Int")
		;NumPut(cChildren, TVITEM, A_PtrSize == 8 ? 44 : 32, "Int")
		;NumPut(lParam, TVITEM, A_PtrSize == 8 ? 48 : 36, "Ptr")
		if (this.SendMessage(this.TVM_GETITEM, 0, &TVITEM))
		{
			return StrGet(&pszText)		
		}
		return ""
	}
	GetTVITEMObject(ptr)
	{
		o := Object()
		o.mask := NumGet(ptr + 0, 0, "UInt")
		o.hItem := NumGet(ptr + 0, A_PtrSize == 8 ? 8 : 4, "Ptr")
		o.state := NumGet(ptr + 0, A_PtrSize == 8 ? 16 : 8, "UInt")
		o.stateMask := NumGet(ptr + 0, A_PtrSize == 8 ? 20 : 12, "UInt")
		o.pszText := NumGet(ptr + 0, A_PtrSize == 8 ? 24 : 16, "Ptr")
		o.cchTextMax := NumGet(ptr + 0, A_PtrSize == 8 ? 32 : 20, "Int")
		o.iImage := NumGet(ptr + 0, A_PtrSize == 8 ? 36 : 24, "Int")
		o.iSelectedImage := NumGet(ptr + 0, A_PtrSize == 8 ? 40 : 28, "Int")
		o.cChildren := NumGet(ptr + 0, A_PtrSize == 8 ? 44 : 32, "Int")
		o.lParam := NumGet(ptr + 0, A_PtrSize == 8 ? 48 : 36, "Ptr")
		return o
	}
	HitTest(x, y)
	{
		VarSetCapacity(TVHITTESTINFO, A_PtrSize == 8 ? 24 : 16, 0)
		NumPut(x, TVHITTESTINFO, 0, "Int")
		NumPut(y, TVHITTESTINFO, 4, "Int")
		if (this.SendMessage(this.TVM_HITTEST, 0, &TVHITTESTINFO))
		{
			o := Object()
			o.x := x
			o.y := y
			o.flags := NumGet(TVHITTESTINFO, 8, "UInt")
			o.hItem := NumGet(TVHITTESTINFO, A_PtrSize == 8 ? 16 : 12, "Ptr")
			return o
		}
		return false
	}
	InsertItem(szText, hParent, hInsertAfter, iImage, iSelectedImage)
	{
		mask := this.TVIF_TEXT | this.TVIF_IMAGE | this.TVIF_SELECTEDIMAGE
		cchTextMax := StrLen(szText)
		VarSetCapacity(TVINSERTSTRUCT, A_PtrSize == 8 ? 96 : 68, 0)
		NumPut(hParent, TVINSERTSTRUCT, 0, "Ptr")
		NumPut(hInsertAfter, TVINSERTSTRUCT, A_PtrSize == 8 ? 8 : 4, "Ptr")
		item_offset := A_PtrSize == 8 ? 16 : 8
		; TVITEM
		NumPut(mask, TVINSERTSTRUCT, item_offset + 0, "UInt")
		;NumPut(hItem, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 8 : 4), "Ptr")
		;NumPut(state, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 16 : 8), "UInt")
		;NumPut(stateMask, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 20 : 12), "UInt")
		NumPut(&szText, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 24 : 16), "Ptr")
		NumPut(cchTextMax, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 32 : 20), "Int")
		NumPut(iImage, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 36 : 24), "Int")
		NumPut(iSelectedImage, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 40 : 28), "Int")
		;NumPut(cChildren, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 44 : 32), "Int")
		;NumPut(lParam, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 48 : 36), "Ptr")
		; TVITEMEX
		;NumPut(iIntegral, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 56 : 40), "Int")
		;NumPut(uStateEx, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 60 : 44), "UInt")
		;NumPut(hwnd, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 64 : 48), "Ptr")
		;NumPut(iExpandedImage, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 72 : 52), "Int")
		;NumPut(iReserved, TVINSERTSTRUCT, item_offset + (A_PtrSize == 8 ? 76 : 56), "Int")
		
		return this.SendMessage(this.TVM_INSERTITEM, 0, &TVINSERTSTRUCT)
	}
	SelectDropTarget(hItem)
	{
		return this.SelectItem(hItem, this.TVGN_DROPHILITE)
	}
	SelectItem(hItem, flags)
	{
		return this.SendMessage(this.TVM_SELECTITEM, flags, hItem)
	}
	SetImageList(himl, type)
	{
		return this.SendMessage(this.TVM_SETIMAGELIST, type, himl)
	}
}
class CListViewCtrl extends CControl
{
	static WND_CLASS := "SysListView32"
	static LVS_ICON := 0x0000
	static LVS_REPORT := 0x0001
	static LVS_SMALLICON := 0x0002
	static LVS_LIST := 0x0003
	static LVS_TYPEMASK := 0x0003
	static LVS_SINGLESEL := 0x0004
	static LVS_SHOWSELALWAYS := 0x0008
	static LVS_SORTASCENDING := 0x0010
	static LVS_SORTDESCENDING := 0x0020
	static LVS_SHAREIMAGELISTS := 0x0040
	static LVS_NOLABELWRAP := 0x0080
	static LVS_AUTOARRANGE := 0x0100
	static LVS_EDITLABELS := 0x0200
	static LVS_OWNERDATA := 0x1000
	static LVS_NOSCROLL := 0x2000
	static LVS_TYPESTYLEMASK := 0xfc00
	static LVS_ALIGNTOP := 0x0000
	static LVS_ALIGNLEFT := 0x0800
	static LVS_ALIGNMASK := 0x0c00
	static LVS_OWNERDRAWFIXED := 0x0400
	static LVS_NOCOLUMNHEADER := 0x4000
	static LVS_NOSORTHEADER := 0x8000
	static LVM_SETUNICODEFORMAT := CControl.CCM_SETUNICODEFORMAT
	static LVM_GETUNICODEFORMAT := CControl.CCM_GETUNICODEFORMAT
	static LVM_GETBKCOLOR := (CControl.LVM_FIRST + 0)
	static LVM_SETBKCOLOR := (CControl.LVM_FIRST + 1)
	static LVM_GETIMAGELIST := (CControl.LVM_FIRST + 2)
	static LVSIL_NORMAL := 0
	static LVSIL_SMALL := 1
	static LVSIL_STATE := 2
	static LVSIL_GROUPHEADER := 3
	static LVM_SETIMAGELIST := (CControl.LVM_FIRST + 3)
	static LVM_GETITEMCOUNT := (CControl.LVM_FIRST + 4)
	static LVIF_TEXT := 0x00000001
	static LVIF_IMAGE := 0x00000002
	static LVIF_PARAM := 0x00000004
	static LVIF_STATE := 0x00000008
	static LVIF_INDENT := 0x00000010
	static LVIF_NORECOMPUTE := 0x00000800
	static LVIF_GROUPID := 0x00000100
	static LVIF_COLUMNS := 0x00000200
	static LVIF_COLFMT := 0x00010000
	static LVIS_FOCUSED := 0x0001
	static LVIS_SELECTED := 0x0002
	static LVIS_CUT := 0x0004
	static LVIS_DROPHILITED := 0x0008
	static LVIS_GLOW := 0x0010
	static LVIS_ACTIVATING := 0x0020
	static LVIS_OVERLAYMASK := 0x0F00
	static LVIS_STATEIMAGEMASK := 0xF000
	static I_INDENTCALLBACK := (-1)
	static I_GROUPIDCALLBACK := (-1)
	static I_GROUPIDNONE := (-2)
	static I_IMAGECALLBACK := (-1)
	static I_IMAGENONE := (-2)
	static I_COLUMNSCALLBACK := 0xFFFFFFFF
	static LVM_GETITEMA := (CControl.LVM_FIRST + 5)
	static LVM_GETITEMW := (CControl.LVM_FIRST + 75)
	static LVM_SETITEMA := (CControl.LVM_FIRST + 6)
	static LVM_SETITEMW := (CControl.LVM_FIRST + 76)
	static LVM_INSERTITEMA := (CControl.LVM_FIRST + 7)
	static LVM_INSERTITEMW := (CControl.LVM_FIRST + 77)
	static LVM_DELETEITEM := (CControl.LVM_FIRST + 8)
	static LVM_DELETEALLITEMS := (CControl.LVM_FIRST + 9)
	static LVM_GETCALLBACKMASK := (CControl.LVM_FIRST + 10)
	static LVM_SETCALLBACKMASK := (CControl.LVM_FIRST + 11)
	static LVM_GETITEM := A_IsUnicode ? CListViewCtrl.LVM_GETITEMW : CListViewCtrl.LVM_GETITEMA
	static LVM_SETITEM := A_IsUnicode ? CListViewCtrl.LVM_SETITEMW : CListViewCtrl.LVM_SETITEMA
	static LVM_INSERTITEM := A_IsUnicode ? CListViewCtrl.LVM_INSERTITEMW : CListViewCtrl.LVM_INSERTITEMA
	static LVNI_ALL := 0x0000
	static LVNI_FOCUSED := 0x0001
	static LVNI_SELECTED := 0x0002
	static LVNI_CUT := 0x0004
	static LVNI_DROPHILITED := 0x0008
	static LVNI_STATEMASK := (CListViewCtrl.LVNI_FOCUSED | CListViewCtrl.LVNI_SELECTED | CListViewCtrl.LVNI_CUT | CListViewCtrl.LVNI_DROPHILITED)
	static LVNI_VISIBLEORDER := 0x0010
	static LVNI_PREVIOUS := 0x0020
	static LVNI_VISIBLEONLY := 0x0040
	static LVNI_SAMEGROUPONLY := 0x0080
	static LVNI_ABOVE := 0x0100
	static LVNI_BELOW := 0x0200
	static LVNI_TOLEFT := 0x0400
	static LVNI_TORIGHT := 0x0800
	static LVNI_DIRECTIONMASK := (CListViewCtrl.LVNI_ABOVE | CListViewCtrl.LVNI_BELOW | CListViewCtrl.LVNI_TOLEFT | CListViewCtrl.LVNI_TORIGHT)
	static LVM_GETNEXTITEM := (CControl.LVM_FIRST + 12)
	static LVFI_PARAM := 0x0001
	static LVFI_STRING := 0x0002
	static LVFI_SUBSTRING := 0x0004
	static LVFI_PARTIAL := 0x0008
	static LVFI_WRAP := 0x0020
	static LVFI_NEARESTXY := 0x0040
	static LVM_FINDITEMA := (CControl.LVM_FIRST + 13)
	static LVM_FINDITEMW := (CControl.LVM_FIRST + 83)
	static LVM_FINDITEM := A_IsUnicode ? CListViewCtrl.LVM_FINDITEMW : CListViewCtrl.LVM_FINDITEMA
	static LVIR_BOUNDS := 0
	static LVIR_ICON := 1
	static LVIR_LABEL := 2
	static LVIR_SELECTBOUNDS := 3
	static LVM_GETITEMRECT := (CControl.LVM_FIRST + 14)
	static LVM_SETITEMPOSITION := (CControl.LVM_FIRST + 15)
	static LVM_GETITEMPOSITION := (CControl.LVM_FIRST + 16)
	static LVM_GETSTRINGWIDTHA := (CControl.LVM_FIRST + 17)
	static LVM_GETSTRINGWIDTHW := (CControl.LVM_FIRST + 87)
	static LVM_GETSTRINGWIDTH := A_IsUnicode ? CListViewCtrl.LVM_GETSTRINGWIDTHW : CListViewCtrl.LVM_GETSTRINGWIDTHA
	static LVHT_NOWHERE := 0x00000001
	static LVHT_ONITEMICON := 0x00000002
	static LVHT_ONITEMLABEL := 0x00000004
	static LVHT_ONITEMSTATEICON := 0x00000008
	static LVHT_ONITEM := (CListViewCtrl.LVHT_ONITEMICON | CListViewCtrl.LVHT_ONITEMLABEL | CListViewCtrl.LVHT_ONITEMSTATEICON)
	static LVHT_ABOVE := 0x00000008
	static LVHT_BELOW := 0x00000010
	static LVHT_TORIGHT := 0x00000020
	static LVHT_TOLEFT := 0x00000040
	static LVHT_EX_GROUP_HEADER := 0x10000000
	static LVHT_EX_GROUP_FOOTER := 0x20000000
	static LVHT_EX_GROUP_COLLAPSE := 0x40000000
	static LVHT_EX_GROUP_BACKGROUND := 0x80000000
	static LVHT_EX_GROUP_STATEICON := 0x01000000
	static LVHT_EX_GROUP_SUBSETLINK := 0x02000000
	static LVHT_EX_GROUP := (CListViewCtrl.LVHT_EX_GROUP_BACKGROUND | CListViewCtrl.LVHT_EX_GROUP_COLLAPSE | CListViewCtrl.LVHT_EX_GROUP_FOOTER | CListViewCtrl.LVHT_EX_GROUP_HEADER | CListViewCtrl.LVHT_EX_GROUP_STATEICON | CListViewCtrl.LVHT_EX_GROUP_SUBSETLINK)
	static LVHT_EX_ONCONTENTS := 0x04000000
	static LVHT_EX_FOOTER := 0x08000000
	static LVM_HITTEST := (CControl.LVM_FIRST + 18)
	static LVM_ENSUREVISIBLE := (CControl.LVM_FIRST + 19)
	static LVM_SCROLL := (CControl.LVM_FIRST + 20)
	static LVM_REDRAWITEMS := (CControl.LVM_FIRST + 21)
	static LVA_DEFAULT := 0x0000
	static LVA_ALIGNLEFT := 0x0001
	static LVA_ALIGNTOP := 0x0002
	static LVA_SNAPTOGRID := 0x0005
	static LVM_ARRANGE := (CControl.LVM_FIRST + 22)
	static LVM_EDITLABELA := (CControl.LVM_FIRST + 23)
	static LVM_EDITLABELW := (CControl.LVM_FIRST + 118)
	static LVM_EDITLABEL := A_IsUnicode ? CListViewCtrl.LVM_EDITLABELW : CListViewCtrl.LVM_EDITLABELA
	static LVM_GETEDITCONTROL := (CControl.LVM_FIRST + 24)
	static LVCF_FMT := 0x0001
	static LVCF_WIDTH := 0x0002
	static LVCF_TEXT := 0x0004
	static LVCF_SUBITEM := 0x0008
	static LVCF_IMAGE := 0x0010
	static LVCF_ORDER := 0x0020
	static LVCF_MINWIDTH := 0x0040
	static LVCF_DEFAULTWIDTH := 0x0080
	static LVCF_IDEALWIDTH := 0x0100
	static LVCFMT_LEFT := 0x0000
	static LVCFMT_RIGHT := 0x0001
	static LVCFMT_CENTER := 0x0002
	static LVCFMT_JUSTIFYMASK := 0x0003
	static LVCFMT_IMAGE := 0x0800
	static LVCFMT_BITMAP_ON_RIGHT := 0x1000
	static LVCFMT_COL_HAS_IMAGES := 0x8000
	static LVCFMT_FIXED_WIDTH := 0x00100
	static LVCFMT_NO_DPI_SCALE := 0x40000
	static LVCFMT_FIXED_RATIO := 0x80000
	static LVCFMT_LINE_BREAK := 0x100000
	static LVCFMT_FILL := 0x200000
	static LVCFMT_WRAP := 0x400000
	static LVCFMT_NO_TITLE := 0x800000
	static LVCFMT_TILE_PLACEMENTMASK := (CListViewCtrl.LVCFMT_LINE_BREAK | CListViewCtrl.LVCFMT_FILL)
	static LVCFMT_SPLITBUTTON := 0x1000000
	static LVM_GETCOLUMNA := (CControl.LVM_FIRST + 25)
	static LVM_GETCOLUMNW := (CControl.LVM_FIRST + 95)
	static LVM_SETCOLUMNA := (CControl.LVM_FIRST + 26)
	static LVM_SETCOLUMNW := (CControl.LVM_FIRST + 96)
	static LVM_INSERTCOLUMNA := (CControl.LVM_FIRST + 27)
	static LVM_INSERTCOLUMNW := (CControl.LVM_FIRST + 97)
	static LVM_DELETECOLUMN := (CControl.LVM_FIRST + 28)
	static LVM_GETCOLUMNWIDTH := (CControl.LVM_FIRST + 29)
	static LVM_SETCOLUMNWIDTH := (CControl.LVM_FIRST + 30)
	static LVM_GETHEADER := (CControl.LVM_FIRST + 31)
	static LVM_CREATEDRAGIMAGE := (CControl.LVM_FIRST + 33)
	static LVM_GETVIEWRECT := (CControl.LVM_FIRST + 34)
	static LVM_GETTEXTCOLOR := (CControl.LVM_FIRST + 35)
	static LVM_SETTEXTCOLOR := (CControl.LVM_FIRST + 36)
	static LVM_GETTEXTBKCOLOR := (CControl.LVM_FIRST + 37)
	static LVM_SETTEXTBKCOLOR := (CControl.LVM_FIRST + 38)
	static LVM_GETTOPINDEX := (CControl.LVM_FIRST + 39)
	static LVM_GETCOUNTPERPAGE := (CControl.LVM_FIRST + 40)
	static LVM_GETORIGIN := (CControl.LVM_FIRST + 41)
	static LVM_UPDATE := (CControl.LVM_FIRST + 42)
	static LVM_SETITEMSTATE := (CControl.LVM_FIRST + 43)
	static LVM_GETITEMSTATE := (CControl.LVM_FIRST + 44)
	static LVM_GETITEMTEXTA := (CControl.LVM_FIRST + 45)
	static LVM_GETITEMTEXTW := (CControl.LVM_FIRST + 115)
	static LVM_SETITEMTEXTA := (CControl.LVM_FIRST + 46)
	static LVM_SETITEMTEXTW := (CControl.LVM_FIRST + 116)
	static LVSICF_NOINVALIDATEALL := 0x00000001
	static LVSICF_NOSCROLL := 0x00000002
	static LVM_SETITEMCOUNT := (CControl.LVM_FIRST + 47)
	static LVM_SORTITEMS := (CControl.LVM_FIRST + 48)
	static LVM_SETITEMPOSITION32 := (CControl.LVM_FIRST + 49)
	static LVM_GETSELECTEDCOUNT := (CControl.LVM_FIRST + 50)
	static LVM_GETITEMSPACING := (CControl.LVM_FIRST + 51)
	static LVM_GETISEARCHSTRINGA := (CControl.LVM_FIRST + 52)
	static LVM_GETISEARCHSTRINGW := (CControl.LVM_FIRST + 117)
	static LVM_SETICONSPACING := (CControl.LVM_FIRST + 53)
	static LVM_SETEXTENDEDLISTVIEWSTYLE := (CControl.LVM_FIRST + 54)
	static LVM_GETEXTENDEDLISTVIEWSTYLE := (CControl.LVM_FIRST + 55)
	static LVM_GETCOLUMN := A_IsUnicode ? CListViewCtrl.LVM_GETCOLUMNW : CListViewCtrl.LVM_GETCOLUMNA
	static LVM_SETCOLUMN := A_IsUnicode ? CListViewCtrl.LVM_SETCOLUMNW : CListViewCtrl.LVM_SETCOLUMNA
	static LVM_INSERTCOLUMN := A_IsUnicode ? CListViewCtrl.LVM_INSERTCOLUMNW : CListViewCtrl.LVM_INSERTCOLUMNA
	static LVM_GETITEMTEXT := A_IsUnicode ? CListViewCtrl.LVM_GETITEMTEXTW : CListViewCtrl.LVM_GETITEMTEXTA
	static LVM_SETITEMTEXT := A_IsUnicode ? CListViewCtrl.LVM_SETITEMTEXTW : CListViewCtrl.LVM_SETITEMTEXTA
	static LVM_GETISEARCHSTRING := A_IsUnicode ? CListViewCtrl.LVM_GETISEARCHSTRINGW : CListViewCtrl.LVM_GETISEARCHSTRINGA
	static LVS_EX_GRIDLINES := 0x00000001
	static LVS_EX_SUBITEMIMAGES := 0x00000002
	static LVS_EX_CHECKBOXES := 0x00000004
	static LVS_EX_TRACKSELECT := 0x00000008
	static LVS_EX_HEADERDRAGDROP := 0x00000010
	static LVS_EX_FULLROWSELECT := 0x00000020
	static LVS_EX_ONECLICKACTIVATE := 0x00000040
	static LVS_EX_TWOCLICKACTIVATE := 0x00000080
	static LVS_EX_FLATSB := 0x00000100
	static LVS_EX_REGIONAL := 0x00000200
	static LVS_EX_INFOTIP := 0x00000400
	static LVS_EX_UNDERLINEHOT := 0x00000800
	static LVS_EX_UNDERLINECOLD := 0x00001000
	static LVS_EX_MULTIWORKAREAS := 0x00002000
	static LVS_EX_LABELTIP := 0x00004000
	static LVS_EX_BORDERSELECT := 0x00008000
	static LVS_EX_DOUBLEBUFFER := 0x00010000
	static LVS_EX_HIDELABELS := 0x00020000
	static LVS_EX_SINGLEROW := 0x00040000
	static LVS_EX_SNAPTOGRID := 0x00080000
	static LVS_EX_SIMPLESELECT := 0x00100000
	static LVS_EX_JUSTIFYCOLUMNS := 0x00200000
	static LVS_EX_TRANSPARENTBKGND := 0x00400000
	static LVS_EX_TRANSPARENTSHADOWTEXT := 0x00800000
	static LVS_EX_AUTOAUTOARRANGE := 0x01000000
	static LVS_EX_HEADERINALLVIEWS := 0x02000000
	static LVS_EX_AUTOCHECKSELECT := 0x08000000
	static LVS_EX_AUTOSIZECOLUMNS := 0x10000000
	static LVS_EX_COLUMNSNAPPOINTS := 0x40000000
	static LVS_EX_COLUMNOVERFLOW := 0x80000000
	static LVM_GETSUBITEMRECT := (CControl.LVM_FIRST + 56)
	static LVM_SUBITEMHITTEST := (CControl.LVM_FIRST + 57)
	static LVM_SETCOLUMNORDERARRAY := (CControl.LVM_FIRST + 58)
	static LVM_GETCOLUMNORDERARRAY := (CControl.LVM_FIRST + 59)
	static LVM_SETHOTITEM := (CControl.LVM_FIRST + 60)
	static LVM_GETHOTITEM := (CControl.LVM_FIRST + 61)
	static LVM_SETHOTCURSOR := (CControl.LVM_FIRST + 62)
	static LVM_GETHOTCURSOR := (CControl.LVM_FIRST + 63)
	static LVM_APPROXIMATEVIEWRECT := (CControl.LVM_FIRST + 64)
	static LV_MAX_WORKAREAS := 16
	static LVM_SETWORKAREAS := (CControl.LVM_FIRST + 65)
	static LVM_GETWORKAREAS := (CControl.LVM_FIRST + 70)
	static LVM_GETNUMBEROFWORKAREAS := (CControl.LVM_FIRST + 73)
	static LVM_GETSELECTIONMARK := (CControl.LVM_FIRST + 66)
	static LVM_SETSELECTIONMARK := (CControl.LVM_FIRST + 67)
	static LVM_SETHOVERTIME := (CControl.LVM_FIRST + 71)
	static LVM_GETHOVERTIME := (CControl.LVM_FIRST + 72)
	static LVM_SETTOOLTIPS := (CControl.LVM_FIRST + 74)
	static LVM_GETTOOLTIPS := (CControl.LVM_FIRST + 78)
	static LVM_SORTITEMSEX := (CControl.LVM_FIRST + 81)
	static LVBKIF_SOURCE_NONE := 0x00000000
	static LVBKIF_SOURCE_HBITMAP := 0x00000001
	static LVBKIF_SOURCE_URL := 0x00000002
	static LVBKIF_SOURCE_MASK := 0x00000003
	static LVBKIF_STYLE_NORMAL := 0x00000000
	static LVBKIF_STYLE_TILE := 0x00000010
	static LVBKIF_STYLE_MASK := 0x00000010
	static LVBKIF_FLAG_TILEOFFSET := 0x00000100
	static LVBKIF_TYPE_WATERMARK := 0x10000000
	static LVBKIF_FLAG_ALPHABLEND := 0x20000000
	static LVM_SETBKIMAGEA := (CControl.LVM_FIRST + 68)
	static LVM_SETBKIMAGEW := (CControl.LVM_FIRST + 138)
	static LVM_GETBKIMAGEA := (CControl.LVM_FIRST + 69)
	static LVM_GETBKIMAGEW := (CControl.LVM_FIRST + 139)
	static LVM_SETSELECTEDCOLUMN := (CControl.LVM_FIRST + 140)
	static LV_VIEW_ICON := 0x0000
	static LV_VIEW_DETAILS := 0x0001
	static LV_VIEW_SMALLICON := 0x0002
	static LV_VIEW_LIST := 0x0003
	static LV_VIEW_TILE := 0x0004
	static LV_VIEW_MAX := 0x0004
	static LVM_SETVIEW := (CControl.LVM_FIRST + 142)
	static LVM_GETVIEW := (CControl.LVM_FIRST + 143)
	static LVGF_NONE := 0x00000000
	static LVGF_HEADER := 0x00000001
	static LVGF_FOOTER := 0x00000002
	static LVGF_STATE := 0x00000004
	static LVGF_ALIGN := 0x00000008
	static LVGF_GROUPID := 0x00000010
	static LVGF_SUBTITLE := 0x00000100
	static LVGF_TASK := 0x00000200
	static LVGF_DESCRIPTIONTOP := 0x00000400
	static LVGF_DESCRIPTIONBOTTOM := 0x00000800
	static LVGF_TITLEIMAGE := 0x00001000
	static LVGF_EXTENDEDIMAGE := 0x00002000
	static LVGF_ITEMS := 0x00004000
	static LVGF_SUBSET := 0x00008000
	static LVGF_SUBSETITEMS := 0x00010000
	static LVGS_NORMAL := 0x00000000
	static LVGS_COLLAPSED := 0x00000001
	static LVGS_HIDDEN := 0x00000002
	static LVGS_NOHEADER := 0x00000004
	static LVGS_COLLAPSIBLE := 0x00000008
	static LVGS_FOCUSED := 0x00000010
	static LVGS_SELECTED := 0x00000020
	static LVGS_SUBSETED := 0x00000040
	static LVGS_SUBSETLINKFOCUSED := 0x00000080
	static LVGA_HEADER_LEFT := 0x00000001
	static LVGA_HEADER_CENTER := 0x00000002
	static LVGA_HEADER_RIGHT := 0x00000004
	static LVGA_FOOTER_LEFT := 0x00000008
	static LVGA_FOOTER_CENTER := 0x00000010
	static LVGA_FOOTER_RIGHT := 0x00000020
	static LVM_INSERTGROUP := (CControl.LVM_FIRST + 145)
	static LVM_SETGROUPINFO := (CControl.LVM_FIRST + 147)
	static LVM_GETGROUPINFO := (CControl.LVM_FIRST + 149)
	static LVM_REMOVEGROUP := (CControl.LVM_FIRST + 150)
	static LVM_MOVEGROUP := (CControl.LVM_FIRST + 151)
	static LVM_GETGROUPCOUNT := (CControl.LVM_FIRST + 152)
	static LVM_GETGROUPINFOBYINDEX := (CControl.LVM_FIRST + 153)
	static LVM_MOVEITEMTOGROUP := (CControl.LVM_FIRST + 154)
	static LVGGR_GROUP := 0
	static LVGGR_HEADER := 1
	static LVGGR_LABEL := 2
	static LVGGR_SUBSETLINK := 3
	static LVM_GETGROUPRECT := (CControl.LVM_FIRST + 98)
	static LVGMF_NONE := 0x00000000
	static LVGMF_BORDERSIZE := 0x00000001
	static LVGMF_BORDERCOLOR := 0x00000002
	static LVGMF_TEXTCOLOR := 0x00000004
	static LVM_SETGROUPMETRICS := (CControl.LVM_FIRST + 155)
	static LVM_GETGROUPMETRICS := (CControl.LVM_FIRST + 156)
	static LVM_ENABLEGROUPVIEW := (CControl.LVM_FIRST + 157)
	static LVM_SORTGROUPS := (CControl.LVM_FIRST + 158)
	static LVM_INSERTGROUPSORTED := (CControl.LVM_FIRST + 159)
	static LVM_REMOVEALLGROUPS := (CControl.LVM_FIRST + 160)
	static LVM_HASGROUP := (CControl.LVM_FIRST + 161)
	static LVM_GETGROUPSTATE := (CControl.LVM_FIRST + 92)
	static LVM_GETFOCUSEDGROUP := (CControl.LVM_FIRST + 93)
	static LVTVIF_AUTOSIZE := 0x00000000
	static LVTVIF_FIXEDWIDTH := 0x00000001
	static LVTVIF_FIXEDHEIGHT := 0x00000002
	static LVTVIF_FIXEDSIZE := 0x00000003
	static LVTVIF_EXTENDED := 0x00000004
	static LVTVIM_TILESIZE := 0x00000001
	static LVTVIM_COLUMNS := 0x00000002
	static LVTVIM_LABELMARGIN := 0x00000004
	static LVM_SETTILEVIEWINFO := (CControl.LVM_FIRST + 162)
	static LVM_GETTILEVIEWINFO := (CControl.LVM_FIRST + 163)
	static LVM_SETTILEINFO := (CControl.LVM_FIRST + 164)
	static LVM_GETTILEINFO := (CControl.LVM_FIRST + 165)
	static LVIM_AFTER := 0x00000001
	static LVM_SETINSERTMARK := (CControl.LVM_FIRST + 166)
	static LVM_GETINSERTMARK := (CControl.LVM_FIRST + 167)
	static LVM_INSERTMARKHITTEST := (CControl.LVM_FIRST + 168)
	static LVM_GETINSERTMARKRECT := (CControl.LVM_FIRST + 169)
	static LVM_SETINSERTMARKCOLOR := (CControl.LVM_FIRST + 170)
	static LVM_GETINSERTMARKCOLOR := (CControl.LVM_FIRST + 171)
	static LVM_SETINFOTIP := (CControl.LVM_FIRST + 173)
	static LVM_GETSELECTEDCOLUMN := (CControl.LVM_FIRST + 174)
	static LVM_ISGROUPVIEWENABLED := (CControl.LVM_FIRST + 175)
	static LVM_GETOUTLINECOLOR := (CControl.LVM_FIRST + 176)
	static LVM_SETOUTLINECOLOR := (CControl.LVM_FIRST + 177)
	static LVM_CANCELEDITLABEL := (CControl.LVM_FIRST + 179)
	static LVM_MAPINDEXTOID := (CControl.LVM_FIRST + 180)
	static LVM_MAPIDTOINDEX := (CControl.LVM_FIRST + 181)
	static LVM_ISITEMVISIBLE := (CControl.LVM_FIRST + 182)
	static LVM_GETEMPTYTEXT := (CControl.LVM_FIRST + 204)
	static LVM_GETFOOTERRECT := (CControl.LVM_FIRST + 205)
	static LVFF_ITEMCOUNT := 0x00000001
	static LVM_GETFOOTERINFO := (CControl.LVM_FIRST + 206)
	static LVM_GETFOOTERITEMRECT := (CControl.LVM_FIRST + 207)
	static LVFIF_TEXT := 0x00000001
	static LVFIF_STATE := 0x00000002
	static LVFIS_FOCUSED := 0x0001
	static LVM_GETFOOTERITEM := (CControl.LVM_FIRST + 208)
	static LVM_GETITEMINDEXRECT := (CControl.LVM_FIRST + 209)
	static LVM_SETITEMINDEXSTATE := (CControl.LVM_FIRST + 210)
	static LVM_GETNEXTITEMINDEX := (CControl.LVM_FIRST + 211)
	static LVM_SETBKIMAGE := A_IsUnicode ? CListViewCtrl.LVM_SETBKIMAGEW : CListViewCtrl.LVM_SETBKIMAGEA
	static LVM_GETBKIMAGE := A_IsUnicode ? CListViewCtrl.LVM_GETBKIMAGEW : CListViewCtrl.LVM_GETBKIMAGEA
	static LVKF_ALT := 0x0001
	static LVKF_CONTROL := 0x0002
	static LVKF_SHIFT := 0x0004
	static LVCDI_ITEM := 0x00000000
	static LVCDI_GROUP := 0x00000001
	static LVCDI_ITEMSLIST := 0x00000002
	static LVCDRF_NOSELECT := 0x00010000
	static LVCDRF_NOGROUPFRAME := 0x00020000
	static LVN_ITEMCHANGING := (CControl.LVN_FIRST-0)
	static LVN_ITEMCHANGED := (CControl.LVN_FIRST-1)
	static LVN_INSERTITEM := (CControl.LVN_FIRST-2)
	static LVN_DELETEITEM := (CControl.LVN_FIRST-3)
	static LVN_DELETEALLITEMS := (CControl.LVN_FIRST-4)
	static LVN_BEGINLABELEDITA := (CControl.LVN_FIRST-5)
	static LVN_BEGINLABELEDITW := (CControl.LVN_FIRST-75)
	static LVN_ENDLABELEDITA := (CControl.LVN_FIRST-6)
	static LVN_ENDLABELEDITW := (CControl.LVN_FIRST-76)
	static LVN_COLUMNCLICK := (CControl.LVN_FIRST-8)
	static LVN_BEGINDRAG := (CControl.LVN_FIRST-9)
	static LVN_BEGINRDRAG := (CControl.LVN_FIRST-11)
	static LVN_ODCACHEHINT := (CControl.LVN_FIRST-13)
	static LVN_ODFINDITEMA := (CControl.LVN_FIRST-52)
	static LVN_ODFINDITEMW := (CControl.LVN_FIRST-79)
	static LVN_ITEMACTIVATE := (CControl.LVN_FIRST-14)
	static LVN_ODSTATECHANGED := (CControl.LVN_FIRST-15)
	static LVN_HOTTRACK := (CControl.LVN_FIRST-21)
	static LVN_GETDISPINFOA := (CControl.LVN_FIRST-50)
	static LVN_GETDISPINFOW := (CControl.LVN_FIRST-77)
	static LVN_SETDISPINFOA := (CControl.LVN_FIRST-51)
	static LVN_SETDISPINFOW := (CControl.LVN_FIRST-78)
	static LVIF_DI_SETITEM := 0x1000
	static LVN_KEYDOWN := (CControl.LVN_FIRST-55)
	static LVN_MARQUEEBEGIN := (CControl.LVN_FIRST-56)
	static LVGIT_UNFOLDED := 0x0001
	static LVN_GETINFOTIPA := (CControl.LVN_FIRST-57)
	static LVN_GETINFOTIPW := (CControl.LVN_FIRST-58)
	static LVN_INCREMENTALSEARCHA := (CControl.LVN_FIRST-62)
	static LVN_INCREMENTALSEARCHW := (CControl.LVN_FIRST-63)
	static LVN_COLUMNDROPDOWN := (CControl.LVN_FIRST-64)
	static LVN_COLUMNOVERFLOWCLICK := (CControl.LVN_FIRST-66)
	static LVN_BEGINSCROLL := (CControl.LVN_FIRST-80)
	static LVN_ENDSCROLL := (CControl.LVN_FIRST-81)
	static LVN_LINKCLICK := (CControl.LVN_FIRST-84)
	static EMF_CENTERED := 0x00000001
	static LVN_GETEMPTYMARKUP := (CControl.LVN_FIRST-87)
	static LVN_ODFINDITEM := A_IsUnicode ? CListViewCtrl.LVN_ODFINDITEMW : CListViewCtrl.LVN_ODFINDITEMA
	static LVN_BEGINLABELEDIT := A_IsUnicode ? CListViewCtrl.LVN_BEGINLABELEDITW : CListViewCtrl.LVN_BEGINLABELEDITA
	static LVN_ENDLABELEDIT := A_IsUnicode ? CListViewCtrl.LVN_ENDLABELEDITW : CListViewCtrl.LVN_ENDLABELEDITA
	static LVN_GETDISPINFO := A_IsUnicode ? CListViewCtrl.LVN_GETDISPINFOW : CListViewCtrl.LVN_GETDISPINFOA
	static LVN_SETDISPINFO := A_IsUnicode ? CListViewCtrl.LVN_SETDISPINFOW : CListViewCtrl.LVN_SETDISPINFOA
	static LVN_GETINFOTIP := A_IsUnicode ? CListViewCtrl.LVN_GETINFOTIPW : CListViewCtrl.LVN_GETINFOTIPA
	static LVN_INCREMENTALSEARCH := A_IsUnicode ? CListViewCtrl.LVN_INCREMENTALSEARCHW : CListViewCtrl.LVN_INCREMENTALSEARCHA
	__New(dwStyle, dwExStyle := 0, hWnd := 0)
	{
		if (!hWnd)
		{
			Gui, Add, ListView, HWNDhWnd %dwStyle% E%dwExStyle% hidden
		}
		base.__New(hWnd)
	}
	DeleteAllItems()
	{
		return this.SendMessage(this.LVM_DELETEALLITEMS, 0, 0)
	}
	DeleteColumn(iSubItem)
	{
		return this.SendMessage(this.LVM_DELETECOLUMN, iSubItem, 0)
	}
	DeleteItem(iItem)
	{
		return this.SendMessage(this.LVM_DELETEITEM , iItem, 0)
	}
	EnsureVisible(iItem, fPartialOK := false)
	{
		return this.SendMessage(this.LVM_ENSUREVISIBLE, iItem, fPartialOK)
	}
	GetColumnCount()
	{
		static HDM_GETITEMCOUNT := (CControl.HDM_FIRST + 0)
		header := new CWindow(this.SendMessage(this.LVM_GETHEADER, 0, 0))
		return header.SendMessage(HDM_GETITEMCOUNT, 0, 0)
	}
	GetItemCount()
	{
		return this.SendMessage(this.LVM_GETITEMCOUNT, 0, 0)
	}
	GetItemText(iItem, iSubItem)
	{
		cchTextMax := 256
		mask := this.LVIF_TEXT
		VarSetCapacity(szText , cchTextMax * (A_IsUnicode ? 2 : 1),0)
		VarSetCapacity(LVITEM, A_PtrSize == 8 ? 88 : 60, 0)
		NumPut(mask, LVITEM, 0, "UInt")
		NumPut(iItem, LVITEM, 4, "Int")
		NumPut(iSubItem, LVITEM, 8, "Int")
		NumPut(&szText, LVITEM, A_PtrSize == 8 ? 24 : 20, "Ptr")
		NumPut(cchTextMax, LVITEM, A_PtrSize == 8 ? 32 : 24, "Int")
		length := this.SendMessage(this.LVM_GETITEMTEXT, iItem, &LVITEM)
		return StrGet(&szText, length)
	}
	GetLVITEMObject(ptr)
	{
		o := Object()
		o.mask := NumGet(ptr + 0, 0, "UInt")
		o.iItem := NumGet(ptr + 0, 4, "Int")
		o.iSubItem := NumGet(ptr + 0, 8, "Int")
		o.state := NumGet(ptr + 0, 12, "UInt")
		o.stateMask := NumGet(ptr + 0, 16, "UInt")
		o.pszText := NumGet(ptr + 0, A_PtrSize == 8 ? 24 : 20, "Ptr")
		o.cchTextMax := NumGet(ptr + 0, A_PtrSize == 8 ? 32 : 24, "Int")
		o.iImage := NumGet(ptr + 0, A_PtrSize == 8 ? 36 : 28, "Int")
		o.lParam := NumGet(ptr + 0, A_PtrSize == 8 ? 40 : 32, "Ptr")
		o.iIndent := NumGet(ptr + 0, A_PtrSize == 8 ? 48 : 36, "Int")
		o.iGroupId := NumGet(ptr + 0, A_PtrSize == 8 ? 52 : 40, "Int")
		o.cColumns := NumGet(ptr + 0, A_PtrSize == 8 ? 56 : 44, "UInt")
		o.puColumns := NumGet(ptr + 0, A_PtrSize == 8 ? 64 : 48, "Ptr")
		o.iGroup := NumGet(ptr + 0, A_PtrSize == 8 ? 80 : 56, "Int")
		return o
	}
	GetLVDISPINFOObject(ptr)
	{
		o := Object()
		o.hdr := this.GetNMHDRObject(ptr)
		o.item := this.GetLVITEMObject(ptr + (A_PtrSize == 8 ? 24 : 12))
		return o
	}
	GetNextItem(hItem, flags)
	{
		return this.SendMessage(this.LVM_GETNEXTITEM, hItem, flags)
	}
	GetSelectedCount()
	{
		return this.SendMessage(this.LVM_GETSELECTEDCOUNT, 0, 0)
	}
	InsertColumn(iSubItem, szText, cx, fmt, addMask := 0, iImage := 0, iOrder := 0, cxMin := 0, cxDefault :=0, cxIdeal := 0)
	{
		mask := this.LVCF_SUBITEM | this.LVCF_TEXT | this.LVCF_WIDTH | this.LVCF_FMT | addMask
		cchTextMax := StrLen(szText)
		VarSetCapacity(LVCOLUMN, A_PtrSize == 8 ? 56 : 44, 0)
		NumPut(mask, LVCOLUMN, 0, "UInt")
		NumPut(fmt, LVCOLUMN, 4, "Int")
		NumPut(cx, LVCOLUMN, 8, "Int")
		NumPut(&szText, LVCOLUMN, A_PtrSize == 8 ? 16 : 12, "Ptr")
		NumPut(cchTextMax, LVCOLUMN, A_PtrSize == 8 ? 24 : 16, "Int")
		NumPut(iSubItem, LVCOLUMN, A_PtrSize == 8 ? 28 : 20, "Int")
		NumPut(iImage, LVCOLUMN, A_PtrSize == 8 ? 32 : 24, "Int")
		NumPut(iOrder, LVCOLUMN, A_PtrSize == 8 ? 36 : 28, "Int")
		NumPut(cxMin, LVCOLUMN, A_PtrSize == 8 ? 40 : 32, "Int")
		NumPut(cxDefault, LVCOLUMN, A_PtrSize == 8 ? 44 : 36, "Int")
		NumPut(cxIdeal, LVCOLUMN, A_PtrSize == 8 ? 48 : 40, "Int")
		return this.SendMessage(this.LVM_INSERTCOLUMN, iSubItem, &LVCOLUMN)
	}
	InsertItem(iItem, iSubItem, szText, iImage := 0, state := 0, stateMask :=0)
	{
		mask := this.LVIF_TEXT | this.LVIF_STATE
		uMsg := this.LVM_INSERTITEM
		if (iSubItem)
		{
			uMsg := this.LVM_SETITEMTEXT
		}
		else
		{
			mask |= this.LVIF_IMAGE
		}
		cchTextMax := StrLen(szText)
		VarSetCapacity(LVITEM, A_PtrSize == 8 ? 88 : 60, 0)
		NumPut(mask, LVITEM, 0, "UInt")
		NumPut(iItem, LVITEM, 4, "Int")
		NumPut(iSubItem, LVITEM, 8, "Int")
		NumPut(state, LVITEM, 12, "UInt")
		NumPut(stateMask, LVITEM, 16, "UInt")
		NumPut(&szText, LVITEM, A_PtrSize == 8 ? 24 : 20, "Ptr")
		NumPut(cchTextMax, LVITEM, A_PtrSize == 8 ? 32 : 24, "Int")
		NumPut(iImage, LVITEM, A_PtrSize == 8 ? 36 : 28, "Int")
		;NumPut(lParam, LVITEM, A_PtrSize == 8 ? 40 : 32, "Ptr")
		;NumPut(iIndent, LVITEM, A_PtrSize == 8 ? 48 : 36, "Int")
		;NumPut(iGroupId, LVITEM, A_PtrSize == 8 ? 52 : 40, "Int")
		;NumPut(cColumns, LVITEM, A_PtrSize == 8 ? 56 : 44, "UInt")
		;NumPut(puColumns, LVITEM, A_PtrSize == 8 ? 64 : 48, "Ptr")
		;NumPut(iGroup, LVITEM, A_PtrSize == 8 ? 80 : 56, "Int")
		return this.SendMessage(uMsg, iItem, &LVITEM)
	}
	SetImageList(himl, type)
	{
		return this.SendMessage(this.LVM_SETIMAGELIST, type, himl)
	}
	SetItemCount(cItems)
	{
		return this.SendMessage(this.LVM_SETITEMCOUNT, cItems, 0)
	}
	SetExtendedListViewStyle(dwStyles, dwMask := 0)
	{
		return this.SendMessage(this.LVM_SETEXTENDEDLISTVIEWSTYLE, dwMask, dwStyles)
	}
}
class CListView extends CListViewCtrl
{
	__New(name, dwStyle := 0, dwExStyle := 0)
	{
		this.name := name
		base.__New(dwStyle, dwExStyle)
		DllCall("UxTheme.dll\SetWindowTheme", "Ptr", this.hWnd, "WStr", "Explorer", "Ptr", 0)		
		row_height := 20
		iml := new CImageList()
		iml.Create(1, row_height, 8, 1, 1)
		this.SetImageList(iml.himl, CListViewCtrl.LVSIL_SMALL)
	}
	Activate()
	{
		Gui, ListView, % this.hWnd
	}
	Add(options, cols*)
	{
		return this.Insert(this.GetItemCount() + 1, options, cols*)
	}
	DeleteAll()
	{
		return this.DeleteAllItems()
	}
	Delete(row)
	{
		return this.DeleteItem(row - 1)
	}
	Insert(row, options, cols*)
	{
		if (!options || Format("{:L}", options) = "select")
		{
			Loop, % cols.Length()
			{
				state := (A_Index = 1 && options) ? this.LVIS_SELECTED : 0
				this.InsertItem(row - 1, A_Index - 1, cols[A_Index], 0, state, state)
			}
		}
		else
		{
			this.Activate()
			LV_Insert(row, options, cols*)
		}
	}
	Modify(row, options, cols*)
	{
		this.Activate()
		LV_Modify(row, options, cols*)
	}
	ModifyCol(col, options, title := "")
	{
		this.Activate()
		if (title)
		{
			LV_ModifyCol(col, options, title)
		}
		else
		{
			LV_ModifyCol(col, options)
		}
	}
	InsertCol(col, options, title)
	{
		this.Activate()
		LV_InsertCol(col, options, title)
	}
	DeleteCol(col)
	{
		return this.DeleteColumn(col - 1)
	}
	GetCount(mode :="")
	{
		if (!mode)
		{
			return this.GetItemCount()
		}
		else if (RegExMatch(mode, "i)^s"))
		{
			return this.GetSelectedCount()
		}
		else if (RegExMatch(mode, "i)^col"))
		{
			return this.GetColumnCount()
		}
		return 0
	}
	GetItemText(row, col)
	{
		return base.GetItemText(row - 1, col - 1)
	}
	GetNext(StartingRowNumber := 0, RowType := "")
	{
		if (!RowType)
		{
			return this.GetNextItem(StartingRowNumber - 1, this.LVNI_SELECTED) + 1
		}
		else
		{
			this.Activate()
			return LV_GetNext(StartingRowNumber, RowType)
		}
	}
	GetSelectedRowText(col := 1)
	{
		if (this.GetSelectedCount() = 1)
		{
			return this.GetItemText(this.GetNext(0), col)
		}
		return ""
	}
	Sort(col)
	{
		this.ModifyCol(col, "Sort")
	}
	SortDesc(col)
	{
		this.ModifyCol(col, "SortDesc")
	}
	ScrollToEnd()
	{
		iItem := this.GetItemCount() - 1
		if (iItem >=0)
		{
			this.EnsureVisible(iItem)
		}
	}
}
class CExplorerView extends CControl
{
	__New(name)
	{
		static wb
		this.name := name
		Gui, Add, ActiveX, HWNDhWnd vwb, Shell.Explorer
		base.__New(hWnd)
		this.wb := wb
/*
		Gui, Add, ActiveX, hWndhWnd VScroll HScroll, about:<!DOCTYPE html><meta http-equiv='X-UA-Compatible' content='IE=Edge'>
		base.__New(hWnd)
		this.SetWebBrowser()
*/		
		this.SetWindowPos()
	}
	SetWebBrowser()
	{
		static IID_IWebBrowserApp := "{0002DF05-0000-0000-C000-000000000046}" 
		static IID_IHTMLWindow2 := "{332C4425-26CB-11D0-B483-00C04FD90119}"
		static msg := DllCall("RegisterWindowMessage", "str", "WM_HTML_GETOBJECT")
		; AtlAxWin>Shell Embedding>Shell DocObject View>Internet Explorer_Server
		wnd := new CWindow(this.GetWindow(CWU.GW_CHILD))
		wnd := new CWindow(wnd.GetWindow(CWU.GW_CHILD))
		Loop, 5
		{
			wnd_ie := new CWindow(wnd.GetWindow(CWU.GW_CHILD))
			if (wnd_ie.IsWindow())
			{
				VarSetCapacity(GUID, 16, 0)
				DllCall("ole32\CLSIDFromString", "wstr", &IID_IHTMLWindow2, "ptr",&GUID)
				lresult := wnd_ie.SendMessage(msg, 0, 0)
				pdoc := 0
				DllCall("oleacc\ObjectFromLresult", "ptr", lresult, "ptr", &GUID, "ptr", 0, "ptr*", pdoc)
				this.wb := ComObject(VT_DISPATCH := 9, ComObjQuery(pdoc, IID_IWebBrowserApp, IID_IWebBrowserApp), 1)
				ObjRelease(pdoc)
				return true
			}
			Sleep, 100
		}
		this.wb := ""
		return false
	}
	Navigate(url)
	{
		this.wb.Navigate(url)
		Loop, 100
		{
			if (this.wb.document.readyState = "complete")
			{
				return true
			}
			Sleep, 50
		}
		return false
	}
}

FormatMessage(ErrorCode)
{
	ErrorCode := A_LastError
	VarSetCapacity(msg, 2000)
	DllCall("FormatMessage"
      , "UInt", 0x1000      ; FORMAT_MESSAGE_FROM_SYSTEM
      , "UInt", 0
      , "UInt", ErrorCode
      , "UInt", 0x800 ;LANG_SYSTEM_DEFAULT (LANG_USER_DEFAULT=0x400)
      , "Str", msg
      , "UInt", 500
      , "UInt", 0)
	return msg ? msg : "Unknown code [" Format("{1:0X}", ErrorCode) "]"
}
LoadLibrary(filepath)
{
	if (!FileExist(filepath))
	{
		filepath := A_ScriptDir (A_PtrSize = 8 ? "\bin\x64\" : "\bin\x86\") filepath
	}
	
	if (!hModule := DllCall("LoadLibrary", "str", filepath))
	{
		msg := filepath "`r`n`r`n Error : " FormatMessage(A_LastError)
		MsgBox, 0x30, LoadLibrary, %msg%
		ExitApp
	}
	return hModule
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ModalDialog
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
class CAxIEEventHandler 
{
	OnKeyPress(doc) 
	{
		static keys := {1:"selectall", 3:"copy", 22:"paste", 24:"cut"}
		keyCode := doc.parentWindow.event.keyCode
		if keys.HasKey(keyCode)
			Doc.ExecCommand(keys[keyCode])
	}
}
class CModalDialog extends CWindow
{
	__New(html, w, h, title := "", center := true)
	{
		Gui, popup:+Owner1 +hWndhWnd -Resize -MaximizeBox -MinimizeBox
		Gui, popup:Add, ActiveX, hWndhWndAX x0 y0 w%w% h%h%, about:<!DOCTYPE html><meta http-equiv='X-UA-Compatible' content='IE=Edge'>
		base.__New(hWnd)
		
		Sleep, 100
		if (StrLen(html) < 256 && SubStr(html, -4) = ".html")
		{
			filepath := A_ScriptDir "/html/" html
			if (FileExist(filepath))
			{
				FileRead, html, %filepath%
			}
			else
			{
				MsgBox, %filepath% 파일이 존재하지 않습니다
				
			}
		}
		wb := WBGet("ahk_id " hWndAX)
		wb.document.open()
		wb.document.write(html)
		wb.document.close()		
		Loop, 100
		{
			if (wb.document.readyState = "complete")
			{
				break
			}
			Sleep, 50
		}
		; ComObjConnect 의 첫번째 파라미터는 static, global 변수만.
		static doc
		ComObjConnect(doc:= wb.document, new CAxIEEventHandler)		
		
		winAX := new CWindow(hWndAX)
		this.hWndIE := winAX.ControlGet("Internet Explorer_Server1")
		this.iaxie := new CIOleInPlaceActiveObject(wb)
		this.ie := wb.document.parentWindow
		this.ie._EventHandler := this._EventHandler.Bind(this)
		Gui, 1:+Disabled
		if (center)
		{
			parent := new CWindow(this.GetParent())
			pos := parent.GetPos()
			x := "x" (pos.x + pos.w / 2 - w / 2)
			y := "y" (pos.y + pos.h / 2 - h / 2)
			Gui, popup:Show, %x% %y% w%w% h%h%, %title%
		}
		else
		{
			Gui, popup:Show, w%w% h%h%, %title%
		}
		this.clicked_button := 0
		this.ID_CANCEL := 0
		this.ID_OK := 1
	}
	_EventHandler(e)
	{
		if (this["On" e.type])
		{
			this["On" e.type](e)
		}
	}
	OnClick(e)
	{
		if (e.srcElement.type = "button")
		{
			if (e.srcElement.value = "확 인")
			{
				if (!this.ApplyChange())
				{
					return
				}
				this.clicked_button := this.ID_OK
			}
			else
			{
				this.clicked_button := this.ID_CANCEL
			}
			this.PostMessage(CWU.WM_CLOSE, 0, 0)
		}
	}
	ApplyChnage()
	{
		return false
	}
	DoModal()
	{
		WinWaitClose, % this.title
		Gui, popup:Hide
		Gui, popup:Destroy
		Gui, 1:-Disabled
		Gui, 1:Show
		return this.clicked_button
	}
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Setting & Dialogs
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
class CAutosenderSetting
{
	__New()
	{
		this.width := 640
		this.height := 480
		this.chrome_id := 0
		this.fetcher := 0
		this.chrome_debugging_port := 9222
		this.cert_cooldown := 70
		this.cert_retry := 3
		this.receivers := Array()
	}
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
class CAutoSenderSettingDialog extends CModalDialog
{
	__New()
	{
		w := 320
		h := 200
		html =
(
<html lang="ko">
<head>
<meta charset="utf-8">
<style>
html, body {margin:0;padding:0;overflow: hidden;font-size:12px; background-color:#F0F0F0;width:100`%;height:100`%;}
table{width:100`%;height:100`%; border-spacing:0}
.setting td:nth-child(1) {font-size:12px; width: 120px; text-align: right}
.setting td:nth-child(2) {font-size:12px; padding-left:20px}
.setting input{ border-radius: 3px; border : 1px solid #ADADAD}
.setting input[type=checkbox] {}
.setting input[type=text] {width:60px; text-align:right;}
.setting input[type=button] {font-size:12px; ; background-color:#E1E1E1;width:80px;height:25px; }
td.bottom{align:right; padding:10px;}
.separator > td {border-top: 1px solid #ADADAD;}
</style>
<script>
var setting = null;
var proptype_checkbox = {"chrome_id" : true, "fetcher" : true, "fetcher_check_time" : true, "fetcher_filter" : true, "skip_send_logging" : true, "ex_charging_auto" : true, "ex_charging_target" : true};
function EventHandler(e)
{
	if (_EventHandler)
	{
		_EventHandler(e)
	}
}
function Init(data)
{
	if (data)
	{
		setting = JSON.parse(data);
	}
	for (prop in setting)
	{
		if (proptype_checkbox[prop])
		{
			var elms = document.getElementsByClassName(prop);
			for (var i=0; i < elms.length; i++)
			{
				elms[i].checked = (i == setting[prop]);
				elms[i].addEventListener("click", OnClickCheckbox)
			}
		}
		else
		{
			var elm = document.getElementById(prop);
			if (elm)
			{
				elm.value = setting[prop];
			}		
		}	
	}
	var elms = document.getElementsByClassName("button");
	for (var i=0; i < elms.length; i++)
	{
		elms[i].addEventListener("click", EventHandler);
	}
}
function Submit()
{
	for (prop in setting)
	{
		if (proptype_checkbox[prop])
		{
			var elms = document.getElementsByClassName(prop);
			for (var i=0; i < elms.length; i++)
			{
				if  (elms[i].checked)
				{
					setting[prop] = i;
					break;
				}
			}
			
		}
		else
		{
			var elm = document.getElementById(prop)
			if (elm)
			{
				setting[prop] = parseInt(elm.value);
			}
		}
	}
	return JSON.stringify(setting);
}
function OnClickCheckbox(e)
{
	var elms = document.getElementsByClassName(e.srcElement.className);
	for (var i=0; i < elms.length; i++)
	{
		if (elms[i] == e.srcElement)
		{
		}
		else
		{
			elms[i].checked = false;
		}
	}
}
</script>
</head>
<body>
<table class=setting>
  <tr>
    <td>브라우저</td>
	<td> 
	  <input type=checkbox class=chrome_id>크롬 <input type=checkbox class=chrome_id> 엣지 <input type=checkbox class=chrome_id> 웨일
	</td>
  </tr>
  <tr>
    <td>크롬디버거포트</td>
	<td> <input type=text maxlength=5 id=chrome_debugging_port> (default: 9222)</td>
  </tr>
  <tr class=separator>
    <td>문자수신 어플</td>
	<td>
	   <input type=checkbox class=fetcher> 휴대폰과 연결 <br>
	   <input type=checkbox class=fetcher> 구글메시지웹<input type=checkbox class=fetcher>Pushbullet웹
	</td>
  </tr>
  <tr class=separator>
    <td>인증요청 대기시간</td>
	<td> <input type=text maxlength=3 id=cert_cooldown> 초 (최소 60)</td>
  </tr>
  <tr>
    <td>수신실패시 재요청</td>
	<td> <input type=text maxlength=1 id=cert_retry> 회</td>
  </tr>
  <tr class=separator>
	<td colspan=2 class=bottom> 
	   <input type=button class=button value="취 소">&nbsp; 
	   <input type=button class=button value="확 인">
	</td>
  </tr>
</table>
</body>
</html>
)
		base.__New(html, w, h, "설정")
	}
	DoModal(setting)
	{
		this.setting := setting
		this.ie.Init(JSON.Dump(this.setting))
		return base.DoModal()
	}
	ApplyChange()
	{
		data := this.ie.Submit()
		setting := JSON.Load(data)
		err := ""
		if (setting.cert_cooldown < 60)
		{
			err .= "인증요청 대기시간은 SK텔레콤의 제한으로 최소 60초 이상 설정해야합니다"
		}
		else if (setting.chrome_debugging_port < 1024 || setting.chrome_debuggin_port > 65535)
		{
			err .= "디버거포트는 1024 ~ 65535 범위내에서만 설정가능합니다."
		}
		if (err)
		{
			MsgBox, 0x10, 입력오류, %err%
			return false
		}
		this.setting := setting
		return true
	}	
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
class CReceiverInfoDialog extends CModalDialog
{
	__New(n := "", p := "", a := 0)
	{
		this.name := n
		this.phone := p
		this.amount :=a
		
		html = 
(
<html lang="ko">
<head>
<meta charset="utf-8">
<style>
body { font-size:12px; margin:0px; overflow:hidden;background-color:#F0F0F0}
table{width:300px;height:200px;}
.setting td:nth-child(1) {font-size:12px; width: 100px; text-align: center}
.setting td:nth-child(2) {font-size:12px; padding-left:20px}
.setting input[type=text] { border : 1px solid black; width:120px;}
.setting input[type=button] {font-size:12px; border : 1px solid #ADADAD; background-color:#E1E1E1;width:80px;height:25px; }
td.bottom{text-align:center;}
td.title{text-align:left; color:red;}
</style>
<script>
function EventHandler(e)
{
	if (_EventHandler)
	{
		_EventHandler(e)
	}
}
function Init(name, phone, amount)
{
	document.getElementById("name").value = name;
	document.getElementById("phone").value = phone;
	document.getElementById("amount").value = amount;
	var elms = document.getElementsByClassName("button");
	for (var i=0; i < elms.length; i++)
	{
		elms[i].addEventListener("click", EventHandler);
	}
}
function Submit()
{
	return {
	"name" : document.getElementById("name").value, 
	"phone" : document.getElementById("phone").value,
	"amount" : document.getElementById("amount").value,
	};
}
</script>
</head>
<body>
<table class=setting>
  <tr>
    <td>수신자 실명</td>
	<td><input type=text maxlength=20 id=name></td>
  </tr>
  <tr>
    <td>수신자 번호</td>
	<td> <input type=text maxlength=14 id=phone></td>
  </tr>
  <tr>
    <td>선물 용량</td>
	<td> <input type=text maxlength=4 id=amount> GB</td>
  </tr>
  <tr>
	<td colspan=2 class=bottom> 
	   	   <input type=button class=button value="취 소">&nbsp;
			<input type=button class=button value="확 인">
	</td>
  </tr>
</table>
</body>
</html>
)
		base.__New(html, 300, 200, "수신자 추가")
	}
	__Delete()
	{
	}
	DoModal()
	{
		this.ie.Init(this.name, this.phone, this.amount)
		return base.DoModal()
	}
	ApplyChange()
	{
		ret := this.ie.Submit()
		phone := RegExReplace(ret.phone, "[^\d]")
		if (StrLen(phone) != 11 && StrLen(phone) != 10)
		{
			MsgBox, 0x10, 입력오류, 전화번호 길이가 맞지않습니다.
			return false
		}
		name := ret.name
/* 외국인 사용자 위해 실명길이 체크 안함		
		name := RegExReplace(ret.name, "\s")
		if (StrLen(name) >= 5)
		{
			MsgBox, 0x10, 입력오류, 실명은 4자이하로 입력하세요.
			return false
		}
*/		
		amount := RegExReplace(ret.amount, "[^\d\.]")
		if (StrLen(amount) >= 5 || RegExMatch(amount, "\.\d{2,}"))
		{
			MsgBox, 0x10, 입력오류, 선물 용량 소수점이하 한자리까지 99기가 이하입니다.
			return false
		}
		this.phone := SubStr(phone, 1, 3) "-" SubStr(phone, 4, 4) "-" SubStr(phone, 8, 4)
		this.name := name
		this.amount := amount
		return true
	}
}	






;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; COM support & helper classes
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
vt(p,n)
{
	return NumGet(NumGet(p+0,"ptr")+n*A_PtrSize,"ptr")
}
_HRESULT(result, caller := "")
{
	static err := { 0x8000FFFF : "Catastrophic failure error."
		, 0x80004001 : "Not implemented error."
		, 0x8007000E : "Out of memory error."
		, 0x80070057 : "One or more arguments are not valid error."
		, 0x80004002 : "Interface not supported error."
		, 0x80004003 : "Pointer not valid error."
		, 0x80070006 : "Handle not valid error."
		, 0x80004004 : "Operation aborted error."
		, 0x80004005 : "Unspecified error."
		, 0x80070005 : "General access denied error."
		, 0x800401E5 : "The object identified by this moniker could not be found."
		, 0x80040201 : "UIA_E_ELEMENTNOTAVAILABLE"
		, 0x80040200 : "UIA_E_ELEMENTNOTENABLED"
		, 0x80131509 : "UIA_E_INVALIDOPERATION"
		, 0x80040202 : "UIA_E_NOCLICKABLEPOINT"
		, 0x80040204 : "UIA_E_NOTSUPPORTED"
		, 0x80040203 : "UIA_E_PROXYASSEMBLYNOTLOADED" }
	if (IsObject(result))
	{
		return false
	}
	result += 0
	result &= 0xFFFFFFFF
	if (result < 0x8000000)
	{
		return result
	}
	return result
}
Variant(ByRef var, type := 0, val := 0) 
{
	sizeof_varaint := 8+ 2 * A_PtrSize
	VarSetCapacity(var, sizeof_varaint, 0)
	NumPut(type, &var, 0, "ushort")
	if (type = 8)
	{
		NumPut(DllCall("oleaut32\SysAllocString", "WStr", val), &var, 8, "ptr")
	}
	else if (type = 3)
	{
		NumPut(val, &var, 8, "int")
	}
}	
class CCOMInterface
{
	__New(pi)
	{
		this.pi := pi
	}
	__Delete()
	{
		if (this.pi)
		{
			this.Release()
			this.pi := ""
		}
	}
	Ptr
	{
		get
		{
			return this.pi
		}
	}
	QueryInterface(strIid)
	{
		VarSetCapacity(iid, 16)
		DllCall("ole32\CLSIDFromString", "wstr", strIid, "ptr", &iid)
		_HRESULT(DllCall(vt(this.pi,0),"ptr",this.pi, "ptr", &iid, "ptr*", &pi := 0), A_ThisFunc)
		return pi
	}
	AddRef()
	{
		return _HRESULT(DllCall(vt(this.pi,1),"ptr",this.pi), A_ThisFunc)
	}
	Release()
	{
		return _HRESULT(DllCall(vt(this.pi,2),"ptr",this.pi), A_ThisFunc)
	}
	
}
class CUnknown
{
    __New(methods*)
	{
		methods.InsertAt(1, "QueryInterface", "AddRef", "Release")
		this.methods := Array()
		Loop, % methods.MaxIndex()
		{
			this.AddInterfaceMethod(methods[A_Index])
		}
		this.vtbl := new CHeap(this.methods.MaxIndex() * A_PtrSize)
		Loop, % this.methods.MaxIndex()
		{
			NumPut(this.methods[A_Index].cb, this.vtbl.p, (A_Index - 1) * A_PtrSize)
		}
		this.pi := new CHeap(A_PtrSize)
        NumPut(this.vtbl.p, this.pi.p)
		this.ref_counter := 0
    }
    __Delete() 
	{
    }
	Ptr
	{
		get
		{
			return this.pi.p
		}
	}
	AddInterfaceMethod(name)
	{
		if (!IsFunc(this[name]))
		{
			return false
		}
		cbo := Object()
		cbo.name := name
		cbo.fn := this[name].bind(this)
		cbo.params := this[name].MinParams - 1
		cbo.cb := RegisterCallback(CUnknown.CallBack, "" , cbo.params, &cbo)
		this.methods.Push(cbo)
		return true
	}
	FreeInterfaceMethod()
	{
		Loop, % this.methods.MaxIndex()
		{
			DllCall("GlobalFree", "Ptr", this.methods[A_Index].cb, "Ptr")
		}
		this.methods := Array()
	}
	CallBack(p*)
	{
		static p_type := A_PtrSize == 4 ? "uint" : "int64"
		cbo := object(A_EventInfo)
		if (cbo.params >= 1)
		{
			args := [this]
			Loop, % (cbo.params - 1)
			{
				args.push(NumGet(p + 0 , A_PtrSize * (A_Index - 1), p_type))
			}
			return cbo.fn.call(args*)
		}
		return cbo.fn.call()
	}
    QueryInterface(riid, ppvObject) 
	{
    }
    AddRef(interface) 
	{
		this.ref_counter++
		;LogMessage(A_ThisFunc ":" this.ref_counter)
		return this.ref_counter
    }
    Release(interface) 
	{
		this.ref_counter--
		;LogMessage(A_ThisFunc ":" this.ref_counter)
		return this.ref_counter
    }
}
TaskStrGet(ptr) 
{
	if (A_IsUnicode)
	{
		s := StrGet(ptr)
	}
	else
	{
		ECMAScript.WideCharToMultiByte(ptr, s := "", 0, true)
	}
	DllCall("ole32\CoTaskMemFree", "ptr", ptr)
	return s
}
class CIOleInPlaceActiveObject extends CCOMInterface
{
	__New(obj)
	{
		this.iid := "{00000117-0000-0000-C000-000000000046}"
		base.__New(ComObjQuery(obj, this.iid))
	}
	TranslateAccelerator(pMsg)
	{
		_HRESULT(DllCall(vt(this.pi, 5),"ptr",this.pi,"ptr",pMsg), A_ThisFunc)
	}
}
class CHtmlUtil
{
	__New()
	{
		Gui 1:+LastFoundExist
		if !WinExist()
		{
			Gui, New
			Gui, 1:Show
		}
		Gui, Add, ActiveX, hWndhWnd hidden, about:<!DOCTYPE html><meta http-equiv='X-UA-Compatible' content='IE=Edge'>
		this.wb := WBGet("ahk_id " hWnd)
		this.wb.document.open()
		this.wb.document.write("")
		this.wb.document.close()
	}
	Escape(str)
	{
		return this.wb.document.parentWindow.escape(str)
	}
	Unescape(str)
	{
		return this.wb.document.parentWindow.unescape(str)
	}
	ReplaceHtmlSpeicalChars(str)
	{
		this.SetBody(str)
		return this.wb.document.body.innerText
	}
	SetBody(html)
	{
		this.wb.document.body.innerHTML := html
	}
	GetTable(html, index)
	{
		this.SetBody(html)
		elms := this.wb.document.body.getElementsByTagName("table")
		if (elms && elms.length > index)
		{
			return elms[index]
		}
		return ""
	}
	GetTableData(html)
	{
		this.SetBody(html)
		elms := this.wb.document.body.getElementsByTagName("table")
		data := Array()
		if (elms)
		{
			Loop, % elms.length
			{
				table := elms[A_Index - 1]
				rows := Array()
				Loop, % table.rows.length
				{
					cells := Array()
					row := A_Index - 1
					Loop, % table.rows[row].cells.length
					{
						cells.Push(table.rows[row].cells[A_Index - 1].innerText)
					}
					rows.Push(cells)
				}
				data.Push(rows)
			}
		}
		return data
	}
}


class CFormData
{
	__New()
	{
		this.data := Object()
	}
	Set(key, val)
	{
		this.data[key] := val
	}
	GetData()
	{
		data := ""
		for key, val in this.data
		{
			data .= key "=" ECMAScript.encodeURI(val) "&"
		}
		return Trim(data, "&")
	}
}
class CHTTP
{
	__New()
	{
		this.http := ComObjCreate("WinHttp.WinHttpRequest.5.1")
		this.Win7WinHttpTLS12Support()
	}
	__Delete()
	{
		if (this.http)
		{
			ObjRelease(this.http)
		}
	}
	Win7WinHttpTLS12Support()
	{
		; WIN_7 에서 WinHttp.WinHttpRequest.5.1로 httprequest 할 경우 TLS 1.2 보안채널 사용하게 하기위한 코드 백업. 
		; TLS 1.2 미만은 최신 서버, 브라우저들에서 더이상 사용되지 않는데 win 7 winhttp에선 tls 1.2가 default 로 사용되지 않는다.
		if (A_OSVersion != "WIN_7")
			return
		RegRead, DefaultSecureProtocols, HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings\WinHttp, DefaultSecureProtocols
		if !(DefaultSecureProtocols & 0x800)
		{
			if (A_IsAdmin)
			{
				commandLine := DllCall("GetCommandLine", "str")
				isRestarted := !!RegExMatch(commandLine, " /restart(?!\S)")
				if (A_Is64bitOS)
				{
					RegWrite, REG_DWORD, HKEY_LOCAL_MACHINE\SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Internet Settings\WinHttp, DefaultSecureProtocols, 0xA00
				}
				RegWrite, REG_DWORD, HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings\WinHttp, DefaultSecureProtocols, 0xA00
				if (isRestarted)
				{
					MsgBox, 레지스트리값을 설정완료후 프로그램을 종료합니다. 일반권한으로 자동입력기를 다시 실행해주세요.
					ExitApp
				}
			}
			else 
			{
				msg := "윈도우 7 사용자는 WinHttp를 사용하려면 기본 보안 프로토콜을 TLS 1.2로 설정해야합니다. `n`n"
				msg .= "윈도우 8 이상부터는 TLS 1.2 이상이 기본이고 인터넷접속시 보안이 더 강화되는 설정입니다.`n"
				msg .= "최근 웹서버들은 보안에 취약한 프로토콜을 사용하면 접속자체를 거부해서 업데이트 기능을 사용할 수 없습니다. `n"
				msg .= "관련 정보는 아래 사이트에서 확인할수 있습니다`n`n"
				msg .= "https://docs.microsoft.com/ko-kr/mem/configmgr/core/plan-design/security/enable-tls-1-2-client`n`n"
				msg .= "지금 관리자권한으로 자동 설정하시겠습니까?"
				MsgBox, 4, WinHttp 보안프로토콜 설정, %msg%
				IfMsgBox, Yes
				{
					MsgBox, 관리자권한으로 다시 실행합니다. 프로그램이 재시작한 후 WinHttp사용(도움말->업데이트 확인)을 수작업으로 다시 해주셔야 설정이 완료됩니다.
					if A_IsCompiled
						Run *RunAs "%A_ScriptFullPath%" /restart
					else
						Run *RunAs "%A_AhkPath%" /restart "%A_ScriptFullPath%"
					ExitApp
				}
				else
				{
					MsgBox, 직접 HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings\WinHttp아래에 DWORD키 DefaultSecureProtocols의 값을 16진수 A00 (10진수 2560) 으로 생성 혹은 설정해주세요.
				}
			}
		}
	}
	Get(url)
	{
		this.http.setTimeouts(1000, 1000, 1000, 1000)
		try
		{
			this.http.open("GET", url, False)
			if (A_OSVersion = "WIN_7")
			{
				; Msxml2.XMLHTTP 에서는 max-age=0 가 안 먹힘.
				this.http.setRequestHeader("Pragma", "no-cache")
				this.http.setRequestHeader("Cache-Control", "no-cache, no-store")
				this.http.setRequestHeader("If-Modified-Since", "Sat, 1 Jan 2000 00:00:00 GMT")
			}
			else
			{
				this.http.setRequestHeader("Cache-Control", "max-age=0")
			}
			this.http.Send()
			return this.http.responseText
		}
		catch e
		{
			msg := e.Message
			MsgBox, %msg%
		}
		return ""
	}
	Post(url, form_urlencoded, file := "")
	{
		this.http.open("POST", url, false)
		this.http.setRequestHeader("Content-Type", "application/x-www-form-urlencoded")
		this.http.send(form_urlencoded)
		if (file && this.http.getResponseHeader("Content-Type") = "application/octet-stream")
		{
			ado := ComObjCreate("ADODB.Stream")
			ado.Type := 1 ; adTypeBinary
			ado.Open()
			ado.Write(this.http.responseBody)
			ado.SaveToFile(file, 2)
			ado.Close()			
			ObjRelease(ado)
			return file
		}
		return this.http.responseText
	}	
}
;WebSocket Protocol Component API
; https://docs.microsoft.com/en-us/windows/win32/websock/web-socket-protocol-component-api-portal
; example code 
; https://docs.microsoft.com/en-us/samples/microsoft/windows-classic-samples/web-socket-protocol/
class CWinHttp
{
	static WINHTTP_OPTION_UPGRADE_TO_WEB_SOCKET := 114
	static WINHTTP_WEB_SOCKET_SEND_OPERATION := 0
	static WINHTTP_WEB_SOCKET_RECEIVE_OPERATION := 1
	static WINHTTP_WEB_SOCKET_CLOSE_OPERATION := 2
	static WINHTTP_WEB_SOCKET_SHUTDOWN_OPERATION := 3
	static WINHTTP_WEB_SOCKET_BINARY_MESSAGE_BUFFER_TYPE := 0
	static WINHTTP_WEB_SOCKET_BINARY_FRAGMENT_BUFFER_TYPE := 1
	static WINHTTP_WEB_SOCKET_UTF8_MESSAGE_BUFFER_TYPE := 2
	static WINHTTP_WEB_SOCKET_UTF8_FRAGMENT_BUFFER_TYPE := 3
	static WINHTTP_WEB_SOCKET_CLOSE_BUFFER_TYPE := 4
	static WINHTTP_WEB_SOCKET_SUCCESS_CLOSE_STATUS := 1000
	static WINHTTP_WEB_SOCKET_ENDPOINT_TERMINATED_CLOSE_STATUS := 1001
	static WINHTTP_WEB_SOCKET_PROTOCOL_ERROR_CLOSE_STATUS := 1002
	static WINHTTP_WEB_SOCKET_INVALID_DATA_TYPE_CLOSE_STATUS := 1003
	static WINHTTP_WEB_SOCKET_EMPTY_CLOSE_STATUS := 1005
	static WINHTTP_WEB_SOCKET_ABORTED_CLOSE_STATUS := 1006
	static WINHTTP_WEB_SOCKET_INVALID_PAYLOAD_CLOSE_STATUS := 1007
	static WINHTTP_WEB_SOCKET_POLICY_VIOLATION_CLOSE_STATUS := 1008
	static WINHTTP_WEB_SOCKET_MESSAGE_TOO_BIG_CLOSE_STATUS := 1009
	static WINHTTP_WEB_SOCKET_UNSUPPORTED_EXTENSIONS_CLOSE_STATUS := 1010
	static WINHTTP_WEB_SOCKET_SERVER_ERROR_CLOSE_STATUS := 1011
	static WINHTTP_WEB_SOCKET_SECURE_HANDSHAKE_ERROR_CLOSE_STATUS := 1015
	__New()
	{
		this.hSession := 0
		this.hConnection := 0
		this.hWebSocket := 0
		this.hRequest := 0
		this.strError := ""
	}
	__Delete()
	{
		this.CloseConnection()
	}
	Open(szAgent)
	{
		this.hSession := DllCall("winhttp.dll\WinHttpOpen"
                       ,"WStr", szAgent
                       ,"Uint", 0
                       ,"Ptr" , 0
                       ,"Ptr", 0
                       ,"Uint", 0)
		if (!this.hSession)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(A_LastError)
			return false
		}
		return true
	}
	Connect(szHost, wPort)
	{
		this.hConnection := DllCall("winhttp.dll\WinHttpConnect"
						, "Ptr", this.hSession
						, "WStr", szHost
						, "UShort", wPort
						, "Uint", 0)
		if (!this.hConnection)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(A_LastError)
			return false
		}
		return true
	}
	OpenRequest(szPath)
	{
		this.hRequest := DllCall("winhttp.dll\WinHttpOpenRequest"
						, "Ptr", this.hConnection
						, "WStr", "GET"
						, "WStr", szPath
						, "Ptr", 0
						, "Ptr", 0
						, "Ptr", 0
						,"Uint", 0)
		if (!this.hRequest)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(A_LastError)
			return false
		}
		return true
	}
	SetOption(dwOption)
	{
		bSuccess := DllCall("winhttp.dll\WinHttpSetOption"
						, "Ptr", this.hRequest
						, "UInt", dwOption
						, "Ptr", 0
						, "UInt", 0)
		if (!bSuccess)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(A_LastError)
			return false
		}	
		return true
	}
	SendRequest()
	{
		bSuccess := DllCall("winhttp.dll\WinHttpSendRequest"
						, "Ptr", this.hRequest
						, "Ptr", 0
						, "UInt", 0
						, "Ptr", 0
						, "UInt", 0
						, "UInt", 0
						, "Ptr", 0)
						
		if (!bSuccess)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(A_LastError)
			return false
		}
		return true
	}
	ReceiveResponse()
	{
		bSuccess := DllCall("winhttp.dll\WinHttpReceiveResponse"
						, "Ptr", this.hRequest
						, "Ptr", 0)
		if (!bSuccess)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(A_LastError)
			return false
		}
		return true
	}
	WebSocketCompleteUpgrade()
	{
		this.hWebsocket := DllCall("winhttp.dll\WinHttpWebSocketCompleteUpgrade"
						, "Ptr", this.hRequest
						, "Ptr", 0)
		if (!this.hWebsocket)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(A_LastError)
			return false
		}
		this.CloseHandle(this.hRequest)
		this.hRequest := 0

		return true
	}
	WebSocketSend(msg)
	{
		dwMsgLength := StrLen(msg)
		dwError := DllCall("winhttp.dll\WinHttpWebSocketSend"
						, "Ptr", this.hWebsocket
						, "Int", this.WINHTTP_WEB_SOCKET_UTF8_MESSAGE_BUFFER_TYPE
						, "AStr", msg
						, "Uint", dwMsgLength)
		if (dwError)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(dwError)
			return false
		}
		return true
	}	
	WebSocketRecv()
	{
		data := ""
		while(1)
		{
			VarSetCapacity(buffer, 1024, 0)
			dwError := DllCall("winhttp.dll\WinHttpWebSocketReceive"
							, "Ptr", this.hWebsocket
							, "Ptr", &buffer
							, "Uint", 1024
							, "UInt*", dwBytesRead := 0
							, "Int*", iBufferType := 0)
			if (dwError)
			{
				this.strError := A_ThisFunc " Error : " FormatMessage(dwError)
				return  false
			}
			data .= StrGet(&buffer, dwBytesRead, "UTF-8")
			if (iBufferType != this.WINHTTP_WEB_SOCKET_UTF8_FRAGMENT_BUFFER_TYPE)
			{
				break
			}
		}
		return data
	}
	WebSocketClose()
	{
		dwError := DllCall("winhttp.dll\WinHttpWebSocketClose"
							, "Ptr", this.hWebsocket
							, "UShort", this.WINHTTP_WEB_SOCKET_SUCCESS_CLOSE_STATUS
							, "Ptr", 0
							, "UInt", 0)
		if (dwError)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(dwError)
			return false
		}	
		return true
	}
	WebSocketShutdown()
	{
		dwError := DllCall("winhttp.dll\WinHttpWebSocketShutdown"
							, "Ptr", this.hWebsocket
							, "UShort", this.WINHTTP_WEB_SOCKET_SUCCESS_CLOSE_STATUS
							, "Ptr", 0
							, "UInt", 0)
		if (dwError)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(dwError)
			return false
		}	
		return true
	}
	WebSocketQueryCloseStatus()
	{
		VarSetCapacity(buffer, 1024, 0)
		dwError := DllCall("winhttp.dll\WinHttpWebSocketQueryCloseStatus"
						, "Ptr", this.hWebsocket
						, "UShort*", status := 0
						, "Ptr", &buffer,
						, "UInt", 1024
						, "UInt*", dwBytesWritten := 0)
		if (dwError)
		{
			this.strError := A_ThisFunc " Error : " FormatMessage(dwError)
			return false
		}
		reason := StrGet(&buffer, dwBytesWritten, "UTF-8")
		ret := Object()
		ret.status := status
		ret.reason := reason
		return ret
	}
	CloseHandle(h)
	{
		if (h)
		{
			DllCall("winhttp.dll\WinHttpCloseHandle", "ptr", h)
		}	
	}
	CloseConnection()
	{
		if (this.hWebSocket)
		{
			this.WebSocketShutdown()
			this.WebSocketClose()
			this.CloseHandle(this.hWebSocket)
		}
		this.CloseHandle(this.hRequest)
		this.CloseHandle(this.hConnection)
		this.CloseHandle(this.hSession)
		this.hRequest := this.hWebSocket := this.hConnection := this.hSession := 0
		this.strError := "연결이 끊어졌습니다"
	}
	OnConnectionClosed()
	{
	}
}

class CWebSocketWinHttp extends CWinHttp
{
	__New()
	{
		base.__New()
		this.bConnected := false
		this.strError := ""
	}
	Connect(url)
	{
		if (this.bConnected)
		{
			this.Disconnect()
		}
		if (RegExMatch(url,  "O)^ws://([^:]+):(\d+)(/devtools/page/.+)$", matches))
		{
			szHost := matches[1]
			wPort := matches[2]
			szPath := matches[3]
			if (!this.Open("DevTools"))
			{
				return false
			}
			if (!base.Connect(szHost, wPort))
			{
				return false
			}
			if (!this.OpenRequest(szPath))
			{
				return false
			}
			if (!this.SetOption(CWinHttp.WINHTTP_OPTION_UPGRADE_TO_WEB_SOCKET))
			{
				return false
			}
			if (!this.SendRequest())
			{
				return false
			}
			if (!this.ReceiveResponse())
			{
				return false
			}
			if (!this.WebSocketCompleteUpgrade())
			{
				return false
			}
			this.bConnected := true
			this.OnConnected()
			return true
		}
		this.strError := "올바른 디버거 주소가 아닙니다"
		return false
	}
	Disconnect()
	{
		if (this.IsConnected())
		{
			base.CloseConnection()
			this.bConnected := false
			this.OnDisconnected()
		}
	}
	GetError()
	{
		return this.strError
	}
	IsConnected()
	{
		return this.bConnected
	}
	Recv()
	{
		return this.WebSocketRecv()
	}
	Send(data)
	{
		if (this.IsConnected())
		{
			return this.WebSocketSend(data)	
		}
		return false
	}
	OnConnected()
	{
	}
	OnDisconnected(bExit := false)
	{
	}
	OnConnectionClosed()
	{
		ret := this.WebSocketQueryCloseStatus()
		base.CloseConnection()
		this.bConnected := false
		this.OnDisconnected(true)
	}	
}
class CWebSocketIE11 extends CWindow
{
	__New()
	{
		static id := 0
		static html := "<script> var ws; function Connect(url) { ws = new WebSocket(url); ws.onopen = ws.onclose = ws.onerror = ws.onmessage = function(e){ if (_AHK) _AHK(e); };} </script>"
		guiname := "WebSocket" id++
		this.guiname := guiname
		Gui 1:+LastFoundExist
		if (!WinExist())
		{
			Gui, 1:New
		}
		Gui, %guiname%:+Owner1 +hWndhWnd
		Gui, %guiname%:Add, ActiveX, hWndhWndAX, about:<!DOCTYPE html><meta http-equiv='X-UA-Compatible' content='IE=edge'>
		Sleep, 100
		base.__New(hWnd)
		this.strError := ""
		wb := WBGet("ahk_id " hWndAX)
		wb.document.open()
		wb.document.write(html)
		wb.document.close()
		this.ie := wb.document.parentWindow
		this.ie._AHK := this.OnWebSocket.Bind(this)
		this.bConnected := false
		this.bError := false
		this.data := Array()
		this.conn := Array()
	}
	__Delete()
	{
	}
	OnWebSocket(e)
	{
		if (this["On" e.type])
		{
			this["On" e.type](e)
		}
	}
	Close(code := 1000, reason := "")
	{
		if (this.ie)
		{
			this.ie.ws.close(code, reason)
		}
	}
	Connect(url)
	{
		this.conn := Array()
		this.ie.Connect(url)
		Loop, 20
		{
			if (this.conn.Length())
			{
				this.bConnected := true
				this.OnConnected()
				break
			}
			Sleep, 50
		}
	}
	Destroy()
	{
		guiname := this.guiname
		Gui, %guiname%:Destroy
	}		
	Disconnect()
	{
		if (this.conn.Length())
		{
			this.bConnected := false
			this.Close()
			Loop, 20
			{
				if (!this.conn.Length())
				{
					break
				}
				Sleep, 50
			}
		}
	}
	GetError()
	{
		return this.strError
	}
	IsConnected()
	{
		return this.bConnected
	}
	Recv()
	{
		this.data := Array()
		Loop, 50
		{
			if (this.data.Length())
				break
			Sleep, 50
		}
		return this.data.Pop()
	}
	Send(data)
	{
		if (this.IsConnected())
		{
			try
			{
				this.ie.ws.send(data)
				return true
			}
			catch e
			{
				return false
			}
		}
		return false
	}
	OnClose(e)
	{
		this.conn.Pop()
		this.strError := e.code " : " e.reason
		this.ie._AHK := ""
		this.ie := ""
		this.Destroy()
		this.bConnected := false
		this.OnDisconnected(this.bError)
	}
	OnConnected()
	{
	}
	OnConnectionClosed()
	{
	}
	OnDisconnected(bExit := false)
	{
	}
	OnError(e)
	{
		if (this.bConnected)
		{
			this.bError := true
		}
	}
	OnMessage(e)
	{
		this.data.Push(e.data)
	}
	OnOpen(e)
	{
		this.conn.Push(1)
	}
}

class CDevToolsClient extends CWebSocketWinHttp
{
	__New(webSocketDebuggerUrl, bUseIEWebSocket := false)
	{
		if (bUseIEWebSocket || A_OSVersion = "WIN_7")
		{
			this.base.base := new CWebSocketIE11()
		}
		base.__New()
		static id := 1
		this.url := webSocketDebuggerUrl
		this.fnPing := this.Ping.Bind(this)
		this.url := webSocketDebuggerUrl
		this.target_id := RegExReplace(this.url, ".+/")
		this.client_id := id++
		this.boolean_false := {boolean : false}
		this.boolean_true := {boolean : true}
		this.Connect(this.url)
	}
	ActivateTarget()
	{
		params := {}
		params.targetId := this.target_id
		return this.GetResponse(this.Call("Target.activateTarget", params))
	}
	Call(method, params:="")
	{
		if (!this.bConnected)
		{
			this.OnConnectionClosed()
			;this.OnDisconnected()
			return ""
		}
		data :={}
		data.id := this.client_id
		data.method := method
		if (params)
		{
			data.params := params
		}
		datastr := StrReplace(StrReplace(JSON.Dump(data), "{""boolean"":0}", "false"), "{""boolean"":1}", "true")
		if (this.Send(datastr))
		{
			recv := this.Recv()
			if (recv)
			{
				response := JSON.Load(recv)
				if (response)
				{
					if (response.id == this.client_id)
					{
						return response.result
					}
					else if (response.method = "Inspector.detached")
					{
						this.OnConnectionClosed()
						this.strError := response.method " : " response.params.reason
						return  ""
					}
					this.strError := JSON.Dump(response)
				}
			}
		}
		this.strError := "서버와의 통신실패"
		return ""
	}
	Evaluate(expression)
	{
		params := {}
		params.expression := expression
		params.objectGroup := "console"
		params.includeCommandLineAPI := this.boolean_true
		params.silent := this.boolean_false
		params.returnByValue := this.boolean_false
		params.userGesture := this.boolean_true
		params.awaitPromise := this.boolean_false
		return this.GetResponse(this.Call("Runtime.evaluate", params), "result")
	}
	GetResponse(response, prop := "result")
	{
		if (response)
		{
			if (response.exceptionDetails)
			{
				return ""
			}
			if (prop && response[prop])
			{
				return response[prop]
			}
			return response
		}
		return ""	
	}
	GetVersion(prop := "product")
	{
		return this.GetResponse(this.Ping(), prop)
	}	
	GetWindowForTarget()
	{
		params := {}
		params.targetId := this.target_id
		return this.GetResponse(this.Call("Browser.getWindowForTarget", params))
	}
	HandleJavaScriptDialog(accept := true)
	{
		params := {}
		params.accept := accept ? this.boolean_true : this.boolean_false
		return this.GetResponse(this.Call("Page.handleJavaScriptDialog", params), "result")
	}
	Navigate(url)
	{
		params := {}
		params.url := url
		return this.GetResponse(this.Call("Page.navigate", params), "frameId")
	}
	SetWindowBounds(windowId, bounds)
	{
		params := {}
		params.windowId := windowId
		params.bounds := bounds
		return this.GetResponse(this.Call("Browser.setWindowBounds", params))
	}
	Ping()
	{
		return this.Call("Browser.getVersion")
	}
	OnConnected()
	{
		ping := this.fnPing
		SetTimer, %ping%, 15000
	}
	OnDisconnected(bExit := false)
	{
		ping := this.fnPing
		if (ping)
		{
			SetTimer, %ping%, Delete
			this.fnPing := ""
			this := ""
		}
		if (_app)
		{
			_app.PostMessage(_app.WM_WEBSOCKET_DISCONNECTED, bExit, this.client_id)
		}
		if (bExit)
		{
			Exit, 1
		}
	}
}

class CCUIAutomation extends CCOMInterface
{
	static CoreWindow_Pos := 1
	static TreeScope_Element := 1
	static TreeScope_Children := 2
	static TreeScope_Descendants := 3
	static TreeScope_Parent := 4
	static TreeScope_Ancestors := 5
	static TreeScope_Subtree := 6
	static UIA_InvokePatternId := 10000
	static UIA_ValuePatternId	:=	10002
	static UIA_ScrollPatternId := 10004
	static UIA_TextPatternId := 10014
	static UIA_ScrollItemPatternId	:=	10017
	static UIA_LegacyIAccessiblePatternId	:=	10018
	static UIA_ControlTypePropertyId	:=	30003
	static UIA_LocalizedControlTypePropertyId	:=	30004
	static UIA_NamePropertyId	:=	30005	
	static UIA_ClassNamePropertyId	:=	30012
	static UIA_AutomationIdPropertyId	:=	30011
	static UIA_ButtonControlTypeId	:=	50000
	static UIA_CheckBoxControlTypeId	:=	50002
	static UIA_EditControlTypeId	:=	50004
	static UIA_ImageControlTypeId	:=	50006
	static UIA_ListItemControlTypeId	:=	50007
	static UIA_ListControlTypeId	:=	50008
	static UIA_GroupControlTypeId := 50026
	static UIA_DocumentControlTypeId := 50030
	static UIA_PaneControlTypeId	:=	50033
	__New()
	{
		this.clsid := "{ff48dba4-60ef-4201-aa87-54103eef594e}"
		this.iid := "{30cbe57d-d9d0-452a-ab13-7ac5ac4825ee}"
		base.__New(ComObjCreate(this.clsid, this.iid))
	}
	GetRootElement()
	{
		_HRESULT(DllCall(vt(this.pi,5),"ptr",this.pi,"ptr*", element := 0), A_ThisFunc)
		return element
	}
	
	ElementFromHandle(hWnd)
	{
		_HRESULT(DllCall(vt(this.pi, 6),"ptr",this.pi,"ptr",hWnd,"ptr*",element := 0), A_ThisFunc)
		return element
	}
	CreateTrueCondition()
	{
		_HRESULT(DllCall(vt(this.pi, 21),"ptr",this.pi,"ptr*",newCondition := 0), A_ThisFunc)
		return newCondition
	}
	CreatePropertyCondition(propertyId, value, type := 8)
	{
		Variant(var := 0, type, value)
		if (A_PtrSize = 4)
		{
			_HRESULT(DllCall(vt(this.pi, 23)
				, "ptr", this.pi, "int", propertyId
				, "int64", NumGet(&var, 0, "int64")
				, "int64", NumGet(&var, 8, "int64")
				, "ptr*",newCondition := 0), A_ThisFunc)
		}
		else
		{
			_HRESULT(DllCall(vt(this.pi, 23)
				, "ptr", this.pi, "int", propertyId
				, "ptr", &var
				, "ptr*",newCondition := 0), A_ThisFunc)
		}
		return newCondition
	}
	CreateAndCondition(condition1, condition2)
	{
		_HRESULT(DllCall(vt(this.pi, 25), "ptr", this.pi, "ptr", condition1, "ptr", condition2, "ptr*",newCondition := 0), A_ThisFunc)
		return newCondition
	}
}
class CIUIAutomationElement extends CCOMInterface
{
	Click()
	{
		rc := this.CurrentBoundingRectangle()
		x := rc.left + Floor((rc.right-rc.left) / 2)
		y := rc.top + Floor((rc.bottom-rc.top) / 2)
		mode := A_CoordModeMouse
		CoordMode, Mouse, Screen
		MouseClick, left, x, y
		CoordMode, Mouse, %mode%
	}
	SetFocus()
	{
		return _HRESULT(DllCall(vt(this.pi,3),"ptr",this.pi), A_ThisFunc)
	}
	FindAll(scope, condition)
	{
		_HRESULT(DllCall(vt(this.pi, 6),"ptr",this.pi,"int",scope,"ptr",condition,"ptr*",found := 0), A_ThisFunc)
		return found
	}
	GetCurrentPattern(patternId)
	{
		_HRESULT(DllCall(vt(this.pi,16),"ptr",this.pi,"int",patternId,"ptr*",patternObject := 0), A_ThisFunc)
		return patternObject
	}
	CurrentControlType()
	{
		_HRESULT(DllCall(vt(this.pi,21),"ptr",this.pi,"uint*", type := 0), A_ThisFunc)
		return type
	}
	CurrentName()
	{
		_HRESULT(DllCall(vt(this.pi, 23),"ptr",this.pi,"ptr*",retVal := 0), A_ThisFunc)
		return StrGet(retVal,"utf-16")
	}
	CurrentHasKeyboardFocus()
	{
		_HRESULT(DllCall(vt(this.pi,26),"ptr",this.pi,"int*",retVal := 0), A_ThisFunc)
		return retVal
	}
	CurrentAutomationId()
	{
		_HRESULT(DllCall(vt(this.pi, 29),"ptr",this.pi,"ptr*",retVal := 0), A_ThisFunc)
		return StrGet(retVal,"utf-16")
	}
	ClassName()
	{
		_HRESULT(DllCall(vt(this.pi, 30),"ptr",this.pi,"ptr*",retVal := 0), A_ThisFunc)
		return StrGet(retVal,"utf-16")
	}
	CurrentNativeWindowHandle()
	{
		_HRESULT(DllCall(vt(this.pi, 36),"ptr",this.pi,"ptr*",retVal := 0), A_ThisFunc)
		return retVal
	}
	CurrentBoundingRectangle()
	{
		VarSetCapacity(rect, 16, 0)
		_HRESULT(DllCall(vt(this.pi, 43),"ptr",this.pi,"ptr", &rect), A_ThisFunc)
		return CWindow.GetRECTObject(&rect)
	}
	GetChildren(condition)
	{
		return this.FindAll(CCUIAutomation.TreeScope_Children, condition)
	}
	GetChild(index, condition)
	{
		children := new CIUIAutomationElementArray(this.GetChildren(condition))
		if (children.GetLength() > index)
		{
			return children.GetElement(index)
		}
		return 0
	}
}
class CIUIAutomationElementArray extends CCOMInterface
{
	GetLength()
	{
		_HRESULT(DllCall(vt(this.pi, 3),"ptr",this.pi,"int*",length := 0), A_ThisFunc)
		return length
	}
	GetElement(index)
	{
		_HRESULT(DllCall(vt(this.pi, 4),"ptr",this.pi,"int",index,"ptr*",element := 0), A_ThisFunc)
		return element
	}
}
class CIUIAutomationScrollPattern extends CCOMInterface
{
	Scroll(horizontalAmount,verticalAmount)
	{
		return _HRESULT(DllCall(vt(this.pi,3),"ptr",this.pi,"int",horizontalAmount,"int",verticalAmount), A_ThisFunc)
	}
	SetScrollPercent(horizontalPercent,verticalPercent)
	{
		return _HRESULT(DllCall(vt(this.pi,4),"ptr",this.pi,"double",horizontalPercent,"double",verticalPercent), A_ThisFunc)
	}
	CurrentVerticalScrollPercent()
	{
		_HRESULT(DllCall(vt(this.pi,6),"ptr",this.pi,"double*",retVal := 0), A_ThisFunc)
		return retVal
	}
	CurrentVerticalViewSize()
	{
		_HRESULT(DllCall(vt(this.pi,8),"ptr",this.pi,"double*",retVal := 0), A_ThisFunc)
		return retVal
	}
	CurrentVerticallyScrollable()
	{
		_HRESULT(DllCall(vt(this.pi,10),"ptr",this.pi,"int*",retVal := 0), A_ThisFunc)
		return retVal
	}
	ScrollVerticalTop()
	{
		if (this.CurrentVerticallyScrollable())
		{
			this.SetScrollPercent(-1, 0)
		}
	}
	ScrollVerticalBottom()
	{
		if (this.CurrentVerticallyScrollable())
		{
			this.SetScrollPercent(-1, 100)
		}
	}
	ScrollVerticalNext(vm := 1)
	{
		if (this.CurrentVerticallyScrollable())
		{
			current := this.CurrentVerticalScrollPercent()
			viewsize := this.CurrentVerticalViewSize()
			if ((current + viewsize) >= 100)
			{
				return false
			}
			next := current + viewsize * vm
			if (next > 100)
			{
				next := 100
			}
			this.SetScrollPercent(-1, next)
			return true
		}
		return false
	}
}
class CIUIAutomationInvokePattern extends CCOMInterface
{
	Invoke()
	{
		return _HRESULT(DllCall(vt(this.pi,3),"ptr",this.pi), A_ThisFunc)
	}
}
class CIUIAutomationScrollItemPattern extends CCOMInterface
{
	ScrollIntoView()
	{
		return _HRESULT(DllCall(vt(this.pi,3),"ptr",this.pi), A_ThisFunc)
	}
}
class CIUIAutomationValuePattern extends CCOMInterface
{
	SetValue(value, type := 8)
	{
		Variant(var, type, value)
		
		Variant(var := 0, type, value)
		if (A_PtrSize = 4)
		{
			return _HRESULT(DllCall(vt(this.pi,3),"ptr" ,this.pi
				, "int64", NumGet(&var, 0, "int64")
				, "int64", NumGet(&var, 8, "int64")), A_ThisFunc)
		}
		else
		{
			return _HRESULT(DllCall(vt(this.pi,3),"ptr" ,this.pi
				, "ptr", &var ), A_ThisFunc)
		}
		
	}
	CurrentValue()
	{
		_HRESULT(DllCall(vt(this.pi, 4),"ptr",this.pi,"ptr*",retVal := 0), A_ThisFunc)
		return StrGet(retVal,"utf-16")
	}
}
class CIUIAutomationLegacyIAccessiblePattern extends CCOMInterface
{
	Select(flagsSelect)
	{
		return _HRESULT(DllCall(vt(this.pi,3),"ptr",this.pi, "int", flagsSelect), A_ThisFunc)
	}
	DoDefaultAction()
	{
		return _HRESULT(DllCall(vt(this.pi,4),"ptr",this.pi), A_ThisFunc)
	}
	CurrentName()
	{
		_HRESULT(DllCall(vt(this.pi, 7),"ptr",this.pi,"ptr*",retVal := 0), A_ThisFunc)
		return StrGet(retVal,"utf-16")
	}
	CurrentDescription()
	{
		_HRESULT(DllCall(vt(this.pi, 9),"ptr",this.pi,"ptr*",retVal := 0), A_ThisFunc)
		return StrGet(retVal,"utf-16")
	}
}
class CIUIAutomationTextPattern extends CCOMInterface
{
;3 IUIAutomationTextPattern::RangeFromPoint
;4 IUIAutomationTextPattern::RangeFromChild
;5 IUIAutomationTextPattern::GetSelection
;6 IUIAutomationTextPattern::GetVisibleRanges
;7 IUIAutomationTextPattern::get_DocumentRange
	get_DocumentRange()
	{
		_HRESULT(DllCall(vt(this.pi, 7),"ptr",this.pi,"ptr*",retVal := 0), A_ThisFunc)
		return retVal
	}
;8 IUIAutomationTextPattern::get_SupportedTextSelection
}
class CIUIAutomationTextRange extends CCOMInterface
{
;3 IUIAutomationTextRange::Clone
;4 IUIAutomationTextRange::Compare
;5 IUIAutomationTextRange::CompareEndpoints
;6 IUIAutomationTextRange::ExpandToEnclosingUnit
;7 IUIAutomationTextRange::FindAttribute
;8 IUIAutomationTextRange::FindText
;9 IUIAutomationTextRange::GetAttributeValue
;10 IUIAutomationTextRange::GetBoundingRectangles
;11 IUIAutomationTextRange::GetEnclosingElement
;12 IUIAutomationTextRange::GetText
	GetText(maxLength := -1)
	{
		_HRESULT(DllCall(vt(this.pi, 12),"ptr",this.pi, "int", maxLength, "ptr*",retVal := 0), A_ThisFunc)
		return StrGet(retVal,"utf-16")
	}
;13 IUIAutomationTextRange::Move
;14 IUIAutomationTextRange::MoveEndpointByUnit
;15 IUIAutomationTextRange::MoveEndpointByRange
;16 IUIAutomationTextRange::Select
;17 IUIAutomationTextRange::AddToSelection
;18 IUIAutomationTextRange::RemoveFromSelection
;19 IUIAutomationTextRange::ScrollIntoView
;20 IUIAutomationTextRange::GetChildren
}

class CUIAApplication extends CWindow
{
	__New(hWnd)
	{
		base.__New(hWnd)
		this.uia := new CCUIAutomation()
		this.cond_true := this.uia.CreateTrueCondition()
		this.app := ""
	}
	FindElement(type, name)
	{
		cond_name := this.uia.CreatePropertyCondition(CCUIAutomation.UIA_NamePropertyId, name)
		cond_controltype := this.uia.CreatePropertyCondition(CCUIAutomation.UIA_ControlTypePropertyId, type, 3)
		cond := this.uia.CreateAndCondition(cond_name, cond_controltype)
		results := new CIUIAutomationElementArray(this.app.FindAll(CCUIAutomation.TreeScope_Subtree, cond))
		if (results.GetLength() != 1)
		{
			return 0
		}
		return results.GetElement(0)
	}
	GetAutomationElement(id)
	{
		cond := this.uia.CreatePropertyCondition(CCUIAutomation.UIA_AutomationIdPropertyId, id)
		results := new CIUIAutomationElementArray(this.app.FindAll(CCUIAutomation.TreeScope_Subtree, cond))
		if (results.GetLength() != 1)
		{
			return 0
		}
		return results.GetElement(0)
	}
	WaitAutomationElement(id, iWait := 0, iTimeOut := 5000, iSleep := 100)
	{
		if (iWait)
		{
			Sleep, %iWait%
		}
		iLoop := Floor(iTimeOut / iSleep)
		Loop, %iLoop%
		{
			if (pi := this.GetAutomationElement(id))
			{
				return pi
			}
			Sleep, %iSleep%
		}
		return 0
	}		
	ShowChildren()
	{
		children := new CIUIAutomationElementArray(this.app.FindAll(CCUIAutomation.TreeScope_Children, this.cond_true))
		Loop, % children.GetLength()
		{
			child := new CIUIAutomationElement(children.GetElement(A_Index -1))
		}
	}
	GetChild(ByRef parent, current_name, class_name := "")
	{
		children := new CIUIAutomationElementArray(parent.FindAll(CCUIAutomation.TreeScope_Children, this.cond_true))
		Loop, % children.GetLength()
		{
			child := new CIUIAutomationElement(children.GetElement(A_Index -1))
			if (current_name = child.CurrentName())
			{
				if (!class_name || class_name = child.ClassName())
				{
					return child
				}
			}
		}
	}
	Init()
	{
		if (!this.uia.pi)
		{
			this.err := "오류 : UI Automation 프레임워크를 사용할수 없습니다"
			return false
		}
		if (!this.hWnd)
		{
			if (this.WinWait(this.title, 5))
			{
				this.hWnd := this.Exist()
			}
			if (!this.hWnd)
			{
				this.err := "오류 : " this.title " 윈도우가 존재하지 않습니다"
				return false
			}
		}
		Sleep, 500
		this.app := new CIUIAutomationElement(this.uia.ElementFromHandle(this.hWnd))
		if (!this.app.pi)
		{
			this.err := "해당 윈도우 핸들로 Element를 찾을 수 없습니다"
			return false
		}
		;this.ShowChildren()
		return true
	}
	SendPassword(pass)
	{
		shift := Array("!", "@", "#", "$", "%", "^", "&", "*", "(", ")","_","+","{","}","|",":","""","<",">","?")
		normal := Array("1","2","3","4","5","6","7","8","9","0","-","=","[","]","\",";","'",",",".","/")
		special := Object()
		Loop, % shift.MaxIndex()
		{
			special[shift[A_Index]] := "{Shift Down}" normal[A_Index] "{Shift Up}"
		}
		str := ""
		Loop, % StrLen(pass)
		{
			ch := SubStr(pass, A_Index, 1)
			str .= special[ch] ? special[ch] : ch
		}
		iDelay := A_KeyDelay 
		iMode := A_SendMode
		SendMode, Event
		SetKeyDelay, 200
		Send, %str%
		SendMode, %iMode%
		SetKeyDelay, %iDelay%
		return true
	}
}


; ECMAScript escape/unescape, URI encode/decode implemenation for autohotkey v1.x
; Written by DDART(https://auto.ddart.net) 2023.04.15
class ECMAScript
{
	static ascii_word_characters := "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_"
	decode(str, preserve_escape_set)
	{
		if (!A_IsUnicode)
		{
			VarSetCapacity(wString, 2 * 2, 0)
		}
		decoded := ""
		len := StrLen(str)
		k := 1
		while(k <= len)
		{
			c := SubStr(str, k, 1)
			if (c = "%")
			{
				if (k + 2 > len)
				{
					return A_LineFile ":" A_LineNumber ":" A_Thisfunc " " "Invalid encoded string near strpos " k
				}
				escaped := SubStr(str, k, 3)
				hexdigits := SubStr(str, k + 1, 2)
				if (!ECMAScript.IsHexDigitString(hexdigits))
				{
					return A_LineFile ":" A_LineNumber ":" A_Thisfunc " " "Invalid encoded string near strpos " k
				}
				k += 2
				utf8byte := "0x" hexdigits
				utf8byte += 0
				n := 0
				Loop 8
				{
					if (utf8byte & 1 << (8-A_Index))
					{
						n++
					}
					else
					{
						break
					}
				}
				if (n == 0)
				{
					asciiChar := Chr(utf8byte)
					c := InStr(preserve_escape_set, asciiChar) ? escaped : asciiChar
				}
				else if (n == 1 || n > 4)
				{
					return A_LineFile ":" A_LineNumber ":" A_Thisfunc " " "Invalid encoded string near strpos " k
				}
				else
				{
					nc := (utf8byte << (n + 1) & 0xFF) >> (n + 1)
					j := 1
					while (j < n)
					{
						k++
						if (k + 2 > len || SubStr(str, k, 1) != "%")
						{
							return A_LineFile ":" A_LineNumber ":" A_Thisfunc " " "Invalid encoded string near strpos " k
						}
						hexdigits := SubStr(str, k + 1, 2)
						if (!ECMAScript.IsHexDigitString(hexdigits))
						{
							return A_LineFile ":" A_LineNumber ":" A_Thisfunc " " "Invalid encoded string near strpos " k
						}
						k += 2
						utf8byte := "0x" hexdigits
						utf8byte += 0
						if (utf8byte >> 6 != 2)
						{
							return A_LineFile ":" A_LineNumber ":" A_Thisfunc " " "Invalid encoded string near strpos " k
						}
						nc := nc << 6 | (utf8byte << 2 & 0xFF) >> 2
						j++
					}
					if (A_IsUnicode || nc <= 0x7F)
					{
						c := Chr(nc)
					}
					else
					{
						NumPut(nc, &wString, 0, "UShort")
						ECMAScript.WideCharToMultiByte(wString, sString)
						c := sString
					}
				}
				
			}
			decoded .= c
			k++
		}
		return decoded
	}
	decodeURI(encoded)
	{
		static  preserve_escape_set := ";/?:@&=+$,#"
		return ECMAScript.decode(encoded, preserve_escape_set)
	}
	decodeURIComponent(encoded)
	{
		return ECMAScript.decode(encoded, "")
	}
	encode(str, extra_unescaped)
	{
		static utf8_leadingbits_first := Array()
		static utf8_leadingbits_remains := 0x80
		static utf8_range := Array()
		static always_unescaped := ECMAScript.ascii_word_characters . "-.!~*'()"
		if (!utf8_range.Length())
		{
			Loop, 5
			{
				lb := 2 ; bin 10
				bits := 8 - (A_Index + 2) + 6 * A_Index
				utf8_range.Push((1 << bits) - 1)
				Loop, %A_Index%
				{	
					lb |= 1 << (1 + A_Index)	
				}
				lb <<= 6 - A_Index
				utf8_leadingbits_first.Push(lb)
			}
		}
		
		unescaped_set := always_unescaped . extra_unescaped
		encoded := ""
		len := A_IsUnicode ? StrLen(str) : ECMAScript.MultiByteToWideChar(str, wString)
		
		k := 1
		while(k <= len)
		{
			nc := A_IsUnicode ? Ord(SubStr(str, k, 1)) : NumGet(&wString, (k - 1) * 2, "UShort")
			k++
			if (nc <= 0x7F)
			{
				c := Chr(nc)
				encoded .= InStr(unescaped_set, c) ? c : "%" Format("{1:02X}", nc)
			}
			else
			{
				Loop, % utf8_range.Length()
				{
					n := A_Index
					if (nc <= utf8_range[n])
					{
						encoded .= "%" Format("{1:02X}", utf8_leadingbits_first[n] | nc >> 6 * n)
						Loop, %n%
						{
							bits_after := 6 * (n - A_Index)
							encoded .= "%" Format("{1:02X}", utf8_leadingbits_remains | (nc & 0x3F << bits_after)  >> bits_after)
						}
						break
					}
				}
			}
		}
		return encoded
	}
	encodeURI(uri)
	{
		static extra_unescaped  := ";/?:@&=+$,#"
		return ECMAScript.encode(uri, extra_unescaped)
	}
	encodeURIComponent(uri_component)
	{
		return ECMAScript.encode(uri_component, "")
	}
	escape(str)
	{
		static unescaped_set := ECMAScript.ascii_word_characters . "@*+-./"
		escaped := ""
		len := A_IsUnicode ? StrLen(str) : ECMAScript.MultiByteToWideChar(str, wString)
		k := 1
		while(k <= len)
		{
			nc := A_IsUnicode ? Ord(SubStr(str, k, 1)) : NumGet(&wString, (k - 1) * 2, "UShort")
			k++
			if (nc <= 0xFF)
			{
				c := Chr(nc)
				escaped .= InStr(unescaped_set, c) ? c : "%" Format("{1:02X}", nc)
			}
			else
			{
				escaped .= "%u" Format("{1:04X}", nc)
			}
		}
		return escaped
	}
	unescape(str, esc_char := "%")
	{
		if (!A_IsUnicode)
		{
			VarSetCapacity(wString, 2 * 2, 0)
		}
		unescaped := ""
		len := StrLen(str)
		k := 1
		while(k <= len)
		{
			c := SubStr(str, k, 1)
			if (c = esc_char)
			{
				hexdigits := ""
				optional_advance := 0
				if (k + 5 <= len && SubStr(str, k + 1, 1) = "u")
				{
					hexdigits := SubStr(str, k + 2, 4)
					optional_advance := 5
				}
				else if (k + 2 <= len)
				{
					hexdigits := SubStr(str, k + 1, 2)
					optional_advance := 2
				}
				if (ECMAScript.IsHexDigitString(hexdigits))
				{
					word := "0x" hexdigits
					word += 0
					if (A_IsUnicode || word <= 0xFF)
					{
						c := Chr(word)
					}
					else
					{
						NumPut(word, &wString, 0, "UShort")
						ECMAScript.WideCharToMultiByte(wString, sString)
						c := sString
					}
					k += optional_advance
				}
			}
			unescaped .= c
			k++
		}
		return unescaped
	}
	IsHexDigitString(str)
	{
		static hexdigits := "0123456789abcdefABCDEF"
		Loop, parse, str
		{
			if (!InStr(hexdigits, A_LoopField))
			{
				return false
			}
		}
		return true
	}
	MultiByteToWideChar(ByRef sString, ByRef wString, CP := 0)
	{
		nSize := DllCall("MultiByteToWideChar"
			, "Uint", CP
			, "Uint", 0
			, "Uint", &sString
			, "int", -1
			, "Uint", 0
			, "int", 0)
		VarSetCapacity(wString, nSize * 2)
		DllCall("MultiByteToWideChar"
			, "Uint", CP
			, "Uint", 0
			, "Uint", &sString
			, "int", -1
			, "Uint", &wString
			, "int", nSize)
		return nSize ? nSize - 1 : 0
	}	
	WideCharToMultiByte(ByRef wString, ByRef sString, CP := 0, bStrPtr := false)
	{
		pwStr := bStrPtr ? wString : &wString
		nSize := DllCall("WideCharToMultiByte"
			, "Uint", CP
			, "Uint", 0
			, "Uint", pwStr
			, "int", -1
			, "Uint", 0
			, "int", 0
			, "Uint", 0
			, "Uint", 0)
		VarSetCapacity(sString, nSize)
		DllCall("WideCharToMultiByte"
			, "Uint", CP
			, "Uint", 0
			, "Uint", pwStr
			, "int", -1
			, "str", sString
			, "int", nSize
			, "Uint", 0
			, "Uint", 0)
		return nSize ? nSize - 1 : 0
	}
}


class CChrome extends CWindow
{
	static ID_CHROME := 0
	static ID_EDGE := 1
	static ID_WHALE := 2
	static DEBUGGING_PORT := 9222
	__New(id := 0, port := 9222, url_base := "")
	{
		if (port = 0)
		{
			base.__New("ahk_class Chrome_WidgetWin_1")
		}
		this.bLaunched := false
		this.SetBaseUrl(url_base)
		this.SetBrowser(id, port)
	}
	__Delete()
	{
		this.InitDevtoolsClient()
	}
	ClickByClassName(name, idx := 0)
	{
		result := this.wb.Evaluate("var ret = 0; var o=document.getElementsByClassName('" name "'); if (o.length > " idx "){o[" idx "].click(); ret=1;} ret;")
		return result.value
	}
	ClickById(id)
	{
		result := this.wb.Evaluate("document.getElementById('" id "').click()")
		if (result)
		{
			return true
		}
		return false
	}
	ClickPopupButton(name)
	{
		uia := new CCUIAutomation()
		app := new CIUIAutomationElement(uia.ElementFromHandle(this.hWnd))
		if (!app.pi)
		{
			return false
		}
		cond_name := uia.CreatePropertyCondition(CCUIAutomation.UIA_NamePropertyId, name)
		cond_controltype := uia.CreatePropertyCondition(CCUIAutomation.UIA_ControlTypePropertyId, CCUIAutomation.UIA_ButtonControlTypeId, 3)
		cond := uia.CreateAndCondition(cond_name, cond_controltype)
		results := new CIUIAutomationElementArray(app.FindAll(CCUIAutomation.TreeScope_Subtree, cond))
		if (results.GetLength() = 1)
		{
			button := new CIUIAutomationElement(results.GetElement(0))
			button.Click()
			return true
		}
		return false
	}
	Close()
	{
		this.InitDevtoolsClient()
		base.Close()
	}
	CopyFrameHtml(nFrame)
	{
		if (this.IsFrameComplete(nFrame))
		{
			result := this.wb.Evaluate("escape(frames[" (nFrame - 1) "].document.body.parentElement.outerHTML)")
			clipboard := this.Unescape(result.value)
		}
	}
	CopyHtml(file := "")
	{
		if (this.IsComplete())
		{
			location := "document.location.href"
			result := this.wb.Evaluate(location)
			html := "====================================================`r`n"
			html .= location " = " result.value "`r`n"
			html .= "====================================================`r`n"
			result := this.wb.Evaluate("escape(document.body.parentElement.outerHTML)")
			html .= this.Unescape(result.value) "`r`n"
			clipboard := html
			if (file)
			{
				FileDelete, %file%
				FileAppend, %html%, %file%, UTF-8
			}
		}
	}
	DebugHtml(secs := 5)
	{
		Loop, %secs%
		{
			this.CopyHtml("debughtml" A_Index ".txt")
			Sleep, 1000
		}
	}
	Escape(value)
	{
		return ECMAScript.escape(value)
	}
	Exist()
	{
		if (this.port)
		{
			for proc in ComObjGet("winmgmts:").ExecQuery("Select * from Win32_Process where name='" this.exename ".exe'")
			{
				if (InStr(proc.CommandLine, "--remote-debugging-port=" this.port))
				{
					this.title := "ahk_pid " proc.ProcessId
					return base.Exist()
				}
			}
		}
		else
		{
			return base.Exist()
		}
		return false
	}
	GetAbsoluteUrl(url)
	{
		if (InStr(url, "/") = 1)
		{
			return this.url_base url
		}
		return url
	}
	GetBoundingClientRectByClassName(name, idx := 0)
	{
		result := this.wb.Evaluate("JSON.stringify(document.getElementsByClassName('" name "')[" idx "].getBoundingClientRect())")
		if (result.value)
		{
			return JSON.Load(result.value)
		}
		return ""
	}
	GetBoundingClientRectByTagName(name, idx := 0)
	{
		result := this.wb.Evaluate("JSON.stringify(document.getElementsByTagName('" name "')[" idx "].getBoundingClientRect())")
		if (result.value)
		{
			return JSON.Load(result.value)
		}
		return ""
	}
	GetBoundingClientRectById(id)
	{
		result := this.wb.Evaluate("JSON.stringify(document.getElementById('" id "').getBoundingClientRect())")
		if (result.value)
		{
			return JSON.Load(result.value)
		}
		return ""
	}
	GetBrowserInfo()
	{
		if (this.wb)
		{
			result := this.wb.GetVersion()
			return result.userAgent
		}
		return ""
	}
	GetDomain(url)
	{
		if (RegExMatch(url, "O)\/\/([\w\.]+)", matches))
		{
			return StrReplace(matches[1], "www.")
		}
		return ""
	}
	GetLocation()
	{
		result := this.wb.Evaluate("document.location.href")
		return result.value
	}
	GetRenderWidgetHostBoundingRect()
	{
		this.Activate()
		uia := new CCUIAutomation()
		app := new CIUIAutomationElement(uia.ElementFromHandle(this.hWnd))
		if (!app.pi)
		{
			return false
		}
		if (this.GetProcessName() = "msedge.exe")
		{
			cond_doc := uia.CreatePropertyCondition(CCUIAutomation.UIA_ControlTypePropertyId, CCUIAutomation.UIA_DocumentControlTypeId, 3)
			Loop, 10
			{
				results := new CIUIAutomationElementArray(app.FindAll(CCUIAutomation.TreeScope_Subtree, cond_doc))
				if (results.GetLength())
				{
					return new CIUIAutomationElement(results.GetElement(0)).CurrentBoundingRectangle()
				}
				Sleep, 500
			}
			return false
		}
		cond_classname := uia.CreatePropertyCondition(CCUIAutomation.UIA_ClassNamePropertyId, "Chrome_RenderWidgetHostHWND")
		results := new CIUIAutomationElementArray(app.FindAll(CCUIAutomation.TreeScope_Children, cond_classname))
		
		if (results.GetLength() >= 1)
		{
			render := new CIUIAutomationElement(results.GetElement(0))
			return render.CurrentBoundingRectangle()
		}
		return false
	}
	GetWebSocketUrl(value, byUrl := false)
	{
		http := new CHTTP()
		response := http.Get("http://127.0.0.1:" this.port "/json/list")
		list := response ? JSON.Load(response) : 0
		if (!list)
		{
			return -1
		}
		Loop, % list.MaxIndex()
		{
			if (byUrl)
			{
				;LogMessage(A_Index " : " value "<=>" list[A_Index].url)
				if (InStr(list[A_Index].url, value))
				{
					return list[A_Index].webSocketDebuggerUrl
				}
			}
			else 
			{
				title := RegExReplace(list[A_Index].title, "<.*?>")
				if (value = title)
				{
					return list[A_Index].webSocketDebuggerUrl
				}
			}
		}
		return ""
	}
	GotoUrl(url, bReload := true)
	{
		if (!this.wb)
		{
			return false
		}
		url := this.GetAbsoluteUrl(url)
		if (!bReload)
		{
			result := this.wb.Evaluate("document.location.href")
			if (InStr(result.value, url))
			{
				return this.WaitComplete()
			}
		}
		result := this.wb.Navigate(url)
		if (result)
		{
			if (this.WaitComplete())
			{
				return true
			}
		}
		return false
	}
	Init(url := "")
	{
		this.err := ""
		if (!url)
		{
			url := this.url_base
		}
		if (!url)
		{
			throw A_ThisFunc " 오류 : 이동할 주소값이 없습니다"
		}
		if (hWnd := this.Exist())
		{
			base.__New(hWnd)
		}
		else
		{
			if (!this.Launch())
			{
				this.err .= "개발자모드로 실행된 " this.korname "창이 존재하지 않습니다"
				return false
			}
		}
		ws_url := this.GetWebSocketUrl(url, true)
		if (ws_url = -1)
		{
			this.err := this.korname " DevTool의 /json/list 엔드포인트에 연결할수 없습니다."
			return false
		}
		if (!ws_url)
		{
			ws_url := this.GetWebSocketUrl(this.GetDomain(url), true)
		}
		if (this.wb)
		{
			if (this.wb.url = ws_url && this.wb.bConnected)
			{
				return true
			}
		}
		this.InitDevtoolsClient()
		this.bExistingTarget := false
		if (ws_url)
		{
			this.wb := new CDevToolsClient(ws_url)
			this.bExistingTarget := true
			this.wb.ActivateTarget()
			return true
		}
		if (InStr(this.GetTitle(), "새 탭") || InStr(this.GetTitle(), "New Tab"))
		{
			ws_url := this.GetWebSocketUrl(this.newtab_url, true)
			if (ws_url)
			{
				this.wb := new CDevToolsClient(ws_url)
			}
			else
			{
				this.err := "url " this.newtab_url "를 사용해 새 탭에 대한 디버깅주소를 못찾았습니다" 
				return false
			}
		}
		else if (!this.OpenNewTab())
		{
			this.err .= "새 탭을 열지못함"
			return false
		}
		if (this.wb.Navigate(url))
		{
			this.WaitComplete()
			return true
		}
		this.err .= url " 로 navigate 실패"
		return false
	}
	InitDevtoolsClient()
	{
		if (this.wb)
		{
			this.wb.Disconnect()
			this.wb := ""
		}
	}
	IsComplete()
	{
		result := this.wb.Evaluate("document.readyState")
		return (result && result.value = "complete")
	}
	IsFrameComplete(nFrame)
	{
		result := this.wb.Evaluate("frames[" (nFrame - 1) "].document.readyState")
		return (result && result.value = "complete")
	}
	IsProcessing()
	{
		return false
	}
	IsVerticalScrollable()
	{
		result := this.wb.Evaluate("var ret =0; if (document.body.scrollHeight > document.body.clientHeight){ret = 1;} ret;")
		return result.value
	}
	Launch()
	{
		if (!this.port)
		{
			this.err := "디버깅 포트가 지정되지 않았습니다"
			return false
		}
		exe := this.exename
		lnk := this.lnkname
		RegRead, path, HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\%exe%.exe
		if (!path)
		{
			FileGetShortcut, %A_StartMenuCommon%\Programs\%lnk%.lnk, path
			if (!path)
			{
				FileGetShortcut, %A_StartMenu%\Programs\%lnk%.lnk, path
			}
		}
		if (FileExist(path))
		{
			this.path := path
		}
		else
		{
			MsgBox, , %lnk% Error, "설치된 경로를 찾을수 없습니다"
			return false
		}
		profile := InStr(A_ScriptName, "autocoder") ? "profile_autocoder" : "profile"
		target := this.path (this.port ? " --remote-allow-origins=* --remote-debugging-port=" this.port " --user-data-dir=""" A_ScriptDir "\" exe profile """" : "")
		this.launch_cmd := target
		Run, %target%, , , pid
		if (pid)
		{
			this.bLaunched := true
			this.title := "ahk_pid " pid
			Loop, 20
			{
				Sleep, 250
				if (this.Exist())
				{
					this.Reset()
					Sleep, 1000
					return true
				}
				else
				{
				}
			}
		}
		this.err := FormatMessage(A_LastError)
		return false
	}
	Navigate(url)
	{
		if (!this.port)
		{
			; Non debug mode
			if (!this.Exist())
			{
				return false
			}
			this.Activate()
			if (this.GetTitle() != ("새 탭 - " this.titlesuffix))
			{
				Send, ^t
			}
			Send, %url%
			Send, {Enter}
			return true
		}
		if (!this.wb)
		{
			if (this.Init(url))
			{
				this.ResetBrowserWindow()
				this.WaitComplete()
				return true
			}
			else
			{
				MsgBox, % this.err
				return false
			}
		}
		result := this.wb.Navigate(url)
		if (result && this.WaitComplete(200))
		{
			return true
		}
		return false
	}
	MouseClickByBoundingClientRect(rc, x := 0, y := 0)
	{
		if (A_ScreenDPI != 96)
		{
			m := A_ScreenDPI / 96
			rc.left *= m
			rc.top *= m
			rc.right *= m
			rc.bottom *= m
		}
		rc_render := this.GetRenderWidgetHostBoundingRect()
		if (rc_render)
		{
			x := rc.left + rc_render.left + (x = 0 ? Floor((rc.right - rc.left)/2) : x)
			y := rc.top + rc_render.top +(y = 0 ? Floor((rc.bottom - rc.top)/2) : y)
			this.Activate()
			MouseClick, left, x, y
		}
	}
	MouseClickById(id, x :=0, y := 0)
	{
		this.MouseClickByBoundingClientRect(this.GetBoundingClientRectById(id), x, y)
	}
	MouseClickByClassName(name, index := 0, x :=0, y := 0)
	{
		this.MouseClickByBoundingClientRect(this.GetBoundingClientRectByClassName(name, index), x, y)
	}	
	MouseClickByTagName(name, index := 0, x :=0, y := 0)
	{
		this.MouseClickByBoundingClientRect(this.GetBoundingClientRectByTagName(name, index), x, y)
	}
	OpenNewTab()
	{
		this.InitDevtoolsClient()
		this.Activate()
		Send, ^t
		Sleep, 200
		ws_url := this.WaitWebSocketUrl(this.newtab_url, true)
		if (ws_url && ws_url != -1)
		{
			this.wb := new CDevToolsClient(ws_url)
			return true
		}
		return false
	}	
	ResetBrowserWindow(target := 0)
	{
		if (!target)
		{
			target := this.wb.GetWindowForTarget()
		}
		if (A_ScreenDPI != 96)
		{
			m := A_ScreenDPI / 96
			for key, val in target.bounds
			{
				target.bounds[key] := Floor(target.bounds[key] * m)
			}
		}

		uia := new CCUIAutomation()
		root := new CIUIAutomationElement(uia.GetRootElement())
		cond_classname := uia.CreatePropertyCondition(CCUIAutomation.UIA_ClassNamePropertyId, "Chrome_WidgetWin_1")
		children := new CIUIAutomationElementArray(root.FindAll(CCUIAutomation.TreeScope_Children, cond_classname))
		hWnds := Array()
		Loop, % children.GetLength()
		{
			child := new CIUIAutomationElement(children.GetElement(A_Index -1))
			bounds := child.CurrentBoundingRectangle()
			bounds.width := bounds.right - bounds.left
			bounds.height := bounds.bottom - bounds.top

			if (bounds.left = target.bounds.left && bounds.top = target.bounds.top && bounds.width = target.bounds.width && bounds.height = target.bounds.height)
			{
				SetFormat, Integer, hex
				hWnd := child.CurrentNativeWindowHandle()
				SetFormat, Integer, d
				hWnds.Push(hWnd)
			}
		}
		if (!hWnds.Count())
		{
			return true
		}
		if (hWnds.Count() = 1)
		{
			base.__New(hWnds[1])
			return true
		}
		
		target.bounds.left += 1
		target.bounds.top += 1
		result := this.wb.SetWindowBounds(result.windowId, target.bounds)
		return this.ResetBrowserWindow()
	}	
	ScrollIntoViewByTagName(name, index := 0, option := true)
	{
		this.wb.Evaluate("document.getElementsByTagName('" name "')[" index "].scrollIntoView(true)")
	}
	SetBrowser(id, port)
	{
		this.id := id
		this.port := port
		if (id = CChrome.ID_CHROME)
		{
			this.exename := "chrome"
			this.newtab_url := "chrome://newtab/"
			this.lnkname := "Chrome"
			this.korname := "크롬"
			this.titlesuffix := "Chrome"
		}
		else if (id = CChrome.ID_EDGE)
		{
			this.exename := "msedge"
			this.newtab_url := "edge://newtab/"
			this.lnkname := "Microsoft Edge"
			this.korname := "엣지"
			this.titlesuffix := "개인 - Microsoft​ Edge"
		}
		else if (id = CChrome.ID_WHALE)
		{
			this.exename := "whale"
			this.newtab_url := "chrome://newtab/"
			this.lnkname := "네이버 웨일"
			this.korname := "웨일"
			this.titlesuffix := "Whale"
		}
	}
	SetBaseUrl(url)
	{
		if (url && InStr(url, "http") != 1)
		{
			throw A_ThisFunc " Error : 주소가 http로 시작하지 않습니다"
		}
		this.url_base := url
	}
	Unescape(value)
	{
		return ECMAScript.unescape(value)
	}
	WaitComplete(iWait := 0, iTimeOut := 5000, iSleep := 100)
	{
		if (iWait)
		{
			Sleep, %iWait%
		}
		iLoop := Floor(iTimeOut / iSleep)
		Loop, %iLoop%
		{
			if (!this.wb.bConnected)
			{
				return false
			}
			if (this.IsComplete())
			{
				return true
			}
			Sleep, %iSleep%
		}
		return false
	}
	WaitFrameComplete(iFrame, iWait := 0, iTimeOut := 5000, iSleep := 100)
	{
		if (!this.WaitComplete(iWait, iTimeOut, iSleep))
		{
			return false
		}
		iLoop := Floor(iTimeOut / iSleep)
		frame_prefix := "frames[" iFrame "]."
		Loop, %iLoop%
		{
			result := this.wb.Evaluate(frame_prefix "document.readyState")
			if (result && result.value = "complete")
			{
				return true
			}
			Sleep, %iSleep%
		}
		return false
	}
	WaitLocation(url, iWait := 0, iTimeOut := 5000, iSleep := 100)
	{
		this.WaitComplete(iWait)
		iLoop := Floor(iTimeOut / iSleep)
		Loop, %iLoop%
		{
			result := this.wb.Evaluate("document.location.href")
			if (InStr(result.value, url))
			{
				return true
			}
			Sleep, %iSleep%
		}
		return false
	}		
	WaitProcessing(iWait := 0, iTimeOut := 5000, iSleep := 200)
	{
		if (iWait)
		{
			Sleep, %iWait%
		}
		this.WaitComplete(0, iTimeOut, iSleep)
		iLoop := Floor(iTimeOut / iSleep)
		Loop, %iLoop%
		{
			if (this.IsProcessing())
			{
				Sleep, %iSleep%
			}
			else
			{
				return true
			}
		}
		return false
	}		
	WaitTitle(title)
	{
		Loop, 10
		{
			Sleep, 500
			if (InStr(this.GetTitle(), title))
			{
				this.WaitComplete()
				return true
			}
		}
		return false
	}
	WaitWebSocketUrl(title, byUrl := false, iTimeOut := 5000, iSleep := 200)
	{
		iLoop := Floor(iTimeOut / iSleep)
		Loop, %iLoop%
		{
			url := this.GetWebSocketUrl(title, byUrl)
			if (url)
			{
				return url
			}
			Sleep, %iSleep%
		}
		return ""
	}
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; DataSender
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
class CSTKDataSender
{
	__New()
	{
		this.win := _app.GetChrome("https://www.tworld.co.kr/web/myt-data/giftdata")
		;base.__New("티월드", "", "https://www.tworld.co.kr/", send_path, send_path)
		this.sms := ""
		this.cool_down := 0
	}
	AutoSend(ByRef lv)
	{
		if (!lv.GetCount())
		{
			this.err := "보낼 번호가 없습니다"
			return false
		}
		if (lv.GetCount("S"))
		{
			row := 0
			while(row := lv.GetNext(row))
			{
				if (!this.AutoSendLVRow(lv, row))
				{
					return false
				}
			}
		}
		else
		{
			Loop, % lv.GetCount()
			{
				if (!this.AutoSendLVRow(lv, A_Index))
				{
					return false
				}
			}
		}
		return true
	}
	AutoSendLVRow(ByRef lv, row)
	{
		_app.ExitIfStop()
		name := lv.GetItemText(row, 1)
		phone := lv.GetItemText(row, 2)
		sent := lv.GetItemText(row, 4)
		left := lv.GetItemText(row, 5)
		m1 := SubStr(phone, 1, 3)
		m2 := SubStr(phone, 5, 4)
		m3 := SubStr(phone, 10, 4)
		_app.AppendLog(phone " " name " 님에게 총 " left " GB 데이터 선물을 시작합니다")
		while (left > 0)
		{
			amount := 1
			if (left < amount)
			{
				amount := left
			}
			if (this.SendData(name, m1, m2, m3, amount))
			{
				sent += amount
				left -= amount
				left := Format("{1:.1f}", left)
				lv.Modify(row, "Col4", Format("{1:.1f}", sent))
				lv.Modify(row, "Col5", left)
				_app.AppendLog(amount " GB 선물 성공 - " left "GB 남았습니다")
			}
			else
			{
				return false
			}
		}
		_app.AppendLog(phone " " name " 님에게 데이터 선물을 완료했습니다")
		return true
	}
	GetCertNumber()
	{
		bFound := false
		timeout := 60
		Loop, %timeout%
		{
			_app.ExitIfStop()
			Sleep, 1000
			this.sms.GetLastMessage(msg, prefix)
			if (msg && msg != this.lastmsg)
			{
				bFound := true
				break
			}
			else
			{
				_app.AppendLog(A_Index "/" timeout " : 새로운 문자 수신대기합니다.")
			}
		}
		if (!bFound)
		{
			this.err := "새로운 인증번호를 찾을수 없습니다"
			return false
		}
		_app.AppendLog("발견된 마지막 문자 : " msg)
		if (RegExMatch(msg, "O)인증.*번호[^\d]+(\d{6})", matches))
		;if (RegExMatch(msg, "O)T world 인증 번호 \[(\d{6})\]", matches))
		{
			certno := matches[1]
			_app.AppendLog("새로 발견한 인증번호 :" certno)
			return certno
		}
		this.err := "마지막 메세지가 인증번호 관련 메세지가 아닙니다"
		return 0
	}
	GetError()
	{
		return this.err
	}
	GetM1Index(m1)
	{
		avail := [ "010", "011", "016", "017", "018" ]
		Loop, % avail.MaxIndex()
		{
			if (m1 = avail[A_Index])
			{
				return A_Index -1
			}
		}
		return -1
	}
	GotoCharging()
	{
		return this.win.Navigate(this.win.url_base)
	}
	HandleCertPopup()
	{
		;---------------------------- 마지막 메세지 저장  --------------------------
		this.SaveLastMessage()
		;---------------------------- 본인인증 방식선택창 --------------------------
/*		
		url := this.win.WaitWebSocketUrl("본인인증 | T world")
		if (!url)
		{
			this.err := "본인인증 방식선택창을 찾을수 없습니다"
			return false
		}
		popup_cert_select := new CDevToolsClient(url)
		result := popup_cert_select.Evaluate("goSMS()")
*/		
		;---------------------------- 인증창 --------------------------
		WinWait, SMS 인증 | T world, , 10
		Sleep, 1000
		if (WinExist("SMS 인증 | T world"))
		{
			WinActivate, SMS 인증 | T world
			Send, {Enter}
		}
		url := this.win.WaitWebSocketUrl("SMS 인증 | T world")
		if (!url)
		{
			this.err := "SMS 인증 창을 찾을수 없습니다"
			return false
		}
		popup_cert_sms := new CDevToolsClient(url)
		certno := this.GetCertNumber()
		if (!certno)
		{
			this.err := "인증번호를 가져오지 못했습니다"
			;popup_cert_select.Close()
			popup_cert_sms.Close()
			return false
		}
		popup_cert_sms.Evaluate("document.getElementById('certNum').value='" certno "'")
		popup_cert_sms.Evaluate("document.getElementById('chkconfirm').click()")
		Loop, 10
		{
			;url := this.win.GetWebSocketUrl("SMS 인증 | T world")
			url := this.win.GetWebSocketUrl("본인인증 | T world")
			if (url)
			{
				Sleep, 500
			}
			else
			{
				break
			}
		}
		return true
	}
	Init(fetcher)
	{
		if (fetcher = 0)
		{
			this.sms := new CMSYourPhoneMessageReader()
		}
		else if (fetcher = 1)
		{
			this.sms := new CGoogleMessageReader()
		}
		else if (fetcher = 2)
		{
			this.sms := new CPushbulletMessageReader()
		}
		else
		{
			this.AppendLog("지원하지 않는 문자수신 방식입니다")
			return
		}
		if (!this.sms.Init())
		{
			this.err := this.sms.err
			return false
		}
		if (!this.sms.GetLastMessage(msg, prefix))
		{
			this.err := "문자어플에서 SK텔레콤을 선택하세요"
			return false
		}
		this.err := msg
		return this.InitWebBrowser()
	}
	InitWebBrowser()
	{
		if (this.win.Init())
		{
			this.win.Move(0, 0)
			this.err := "로그인 해주세요"
			if (this.IsLogined())
			{
				this.win.Activate()
				return true
			}
			return false
		}
		return false
	}
	IsLogined()
	{
		result := this.win.wb.Evaluate("escape(document.getElementsByClassName('login-option__button')[0].innerText)")
		if (InStr(ECMAScript.unescape(result.value), "로그아웃"))
		{
			return true
		}
		return false
	}
	SaveLastMessage()
	{
		this.sms.GetLastMessage(msg, prefix)
		this.lastmsg := msg
	}
	SendData(name, m1, m2, m3, amount)
	{
		this.WaitCoolDown()
		if (!this.GotoCharging())
		{
			return false
		}
		result := this.win.wb.Evaluate("document.getElementById('user_phone2')")
		if (result.subtype != "node")
		{
			this.err := "티월드 선물하기 입력창의 구성요소를 찾을수 없습니다."
			return false
		}
		
		m1_index := this.GetM1Index(m1)
		if (m1_index > 0)
		{
			this.win.wb.Evaluate("document.getElementsByClassName('con')[0].children[" m1_index "].firstChild.click()")
		}
		this.win.wb.Evaluate("document.getElementById('user_phone2').value='" m2 "'")
		this.win.wb.Evaluate("document.getElementById('user_phone3').value='" m3 "'")
		this.win.ClickById("giftDataSnd")
		this.win.WaitComplete(1000)
		;---------------------------- 1단계 선택 --------------------------
		if (!this.win.WaitLocation("/web/myt-data/giftdata/selectData"))
		{
			this.err := "1단계 데이터량 선택 화면을 찾지 못했습니다"
			return false
		}
		
		result := this.win.wb.Evaluate("escape(document.getElementsByClassName('remain_data')[0].innerText)")
		if (RegExMatch(this.win.Unescape(result.value), "O)선물 가능 데이터[^\d]+([\d\.]+)GB", matches))
		{
			_app.AppendLog("선물 가능 데이터 : " matches[1] " GB")
			if (amount > matches[1])
			{
				this.err := "데이터 잔여량이 부족합니다"
				return false
			}
		}
		result := this.win.wb.Evaluate("escape(document.getElementsByClassName('phoneNum')[1].innerText)")
		if (!RegExMatch(this.win.Unescape(result.value), "O)([^\s]+)님의", matches))
		{
			this.err := "받을 사람 이름 정보를 찾을수 없습니다"
			return false
		}
		_app.AppendLog(matches[1])
		re := StrReplace(matches[1], "*", ".*")
		if (!RegExMatch(name, re))
		{
			this.err := "받을 사람 이름 정보가 다릅니다"
			return false
		}
		select_id := "data" Round(amount * 1000)
		
		this.win.wb.Evaluate("document.getElementById('" select_id "').click()")
		this.win.wb.Evaluate("document.getElementsByClassName('btnM btnRed')[0].click()")
		Sleep, 2000
		this.win.Activate()
		Send, {Enter}
		this.win.WaitComplete(2000)
		
		;---------------------------- 2단계 SMS 인증 --------------------------
		iRetry := _app.setting.cert_retry
		Loop, %iRetry%
		{
			bSelect := false
			Loop, 3
			{
				if (this.win.ClickById("smsCert"))
				{
					bSelect := true
					break
				}
				Sleep, 1000
			}
			if (!bSelect)
			{
				this.err := "본인인증 인증방식 선택을 할수가 없습니다"
				return false
			}
			if (this.HandleCertPopup())
			{
				break
			}
			else if (A_Index = iRetry)
			{
				return false
			}
			else
			{
				this.sms.Refresh()
				_app.AppendLog("인증번호 수신에 실패 : " this.err)
				_app.AppendLog("새로고침 후 인증번호를 재요청 합니다 (" (iRetry-A_Index) "회 남음)")
			}
		}
		;---------------------------- 2단계 완료 --------------------------
		this.cool_down := A_TickCount + _app.setting.cert_cooldown * 1000
		if (this.win.WaitLocation("/web/myt-data/giftdata/completed"))
		{
			Loop, 10
			{
				result := this.win.wb.Evaluate("document.getElementsByClassName('chgComp').length")
				if (result.value == 1)
				{
					break
				}
				Sleep, 300
			}
			result := this.win.wb.Evaluate("escape(document.getElementsByClassName('chgComp')[0].innerText)")
			_app.AppendLog(this.win.Unescape(result.value))
			return true
		}
		this.err := "선물하기 완료 화면을 찾지 못했습니다"
		return false
	}
	WaitCoolDown()
	{
		if (this.cool_down)
		{
			while((left := this.cool_down - A_TickCount) > 0)
			{
				_app.ExitIfStop()
				_app.AppendLog("데이터선물 재전송 가능시간까지 " Round(left/1000) " 초 남았습니다")
				if (left > 2000)
				{
					Sleep, 2000
				}
				else
				{
					Sleep, %left%
				}
			}
		}
	}
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Phone Message Readers
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
class CMessageReader
{
	__New(n)
	{
		this.name := n
		this.messages := []
		this.page := ""
		this.last := ""
		this.skip_pincheck := false
		this.nTruncatedMessage := 0
	}
	Init()
	{
		return true
	}
	SetCurrentPage(name)
	{
		this.page := "[ " name " ]"
	}
	Fetch()
	{
		return false
	}
	SkipPinCheck(bCheck)
	{
		this.skip_pincheck := bCheck
	}
	AddMessage(msg, prefix := "")
	{
		bFound := this.skip_pincheck 
		pos := 1
		while (!bFound && pos := RegExMatch(msg, "O)([\d\-\_ ]{12,})", matches, pos))
		{
			if (StrLen(RegExReplace(matches[1], "[^\d]")) >= 12)
			{
				bFound := true
				break
			}
			pos += matches.Len(0)
		}
		if (bFound)
		{
			o := Object()
			o.prefix := prefix ? prefix : this.last
			o.time := this.GetTime(o.prefix)
			o.msg := msg
			this.last := o.prefix
			this.messages.Push(o)
		}
	}
	Filter(time, regex)
	{
		arr := Array()
		Loop, % this.messages.MaxIndex()
		{
			if (time = 0 || time <= this.messages[A_Index].time)
			{
				if (!regex || RegExMatch(this.messages[A_Index].msg, regex))
				{
					arr.Push(this.messages[A_Index])
				}
			}
			else
			{
			}
		}
		this.messages := arr
	}
	GetTime(prefix)
	{
		return A_Now
	}
	GetLastMessage(ByRef msg, ByRef prefix)
	{
	}
	Refresh()
	{
	}
}
class CMSYourPhoneMessageReader extends CMessageReader
{
	__New()
	{
		name := this.GetResourceAppName()
		base.__New(name)
		this.uia := new CCUIAutomation()
		this.cond_true := this.uia.CreateTrueCondition()	
		this.elm_messages := ""
	}
	GetResourceAppName()
	{
		; package full name이 버전업이 되면 바뀌고 app name도 바뀔수 있다
		ps_cmd := "powershell.exe -command ""get-appxpackage | where-object name -eq 'Microsoft.YourPhone' | select-object -property packagefullname | format-list | clip"""
		RunWait, %ps_cmd%, , hide
		if (RegExMatch(clipboard, "O)PackageFullName : ([^\s]+)", matches))
		{
			full_package_name := matches[1]
			resource_name_path := "@{" full_package_name "?ms-resource://Microsoft.YourPhone/Resources/AppName}"
			VarSetCapacity(buffer, 1024, 0)
			ret := DllCall("Shlwapi.dll\SHLoadIndirectString", "WStr", resource_name_path, "ptr", &buffer, "UInt", 1024, "ptr", 0, "int", "int")
			if (ret = 0)
			{
				name := StrGet(&buffer, "UTF-16")
				return name
			}
		}
		return "휴대폰과 연결"
	}
	GetChild(ByRef parent, current_name, class_name := "")
	{
		children := new CIUIAutomationElementArray(parent.FindAll(CCUIAutomation.TreeScope_Children, this.cond_true))
		Loop, % children.GetLength()
		{
			child := new CIUIAutomationElement(children.GetElement(A_Index -1))
			if (current_name = child.CurrentName())
			{
				if (!class_name || class_name = child.ClassName())
				{
					return child
				}
			}
		}
	}
	AddChildInfo(pi, depth)
	{
		ObjAddRef(pi)
		parent := new CIUIAutomationElement(pi)
		name := parent.CurrentName()
		classname := parent.ClassName()
		if (classname = "ListViewItem"  || classname = "HyperLink" || classname = "TextBlock")
		{
			return
		}
		children := new CIUIAutomationElementArray(parent.FindAll(CCUIAutomation.TreeScope_Children, this.cond_true))
		if (children.GetLength())
		{
			this.tree .= StrReplace( Format( "{:0" depth "}", "" ), 0, "|") "-" "[" name "][" classname "]`r`n"
		}
		Loop, % children.GetLength()
		{
			child := new CIUIAutomationElement(children.GetElement(A_Index -1))
			this.AddChildInfo(child.pi, depth + 1)
		}		
	}
	GetTree()
	{
		if (!this.uia.pi)
		{
			this.err := "오류 : UI Automation 프레임워크를 사용할수 없습니다"
			return false
		}
		root := new CIUIAutomationElement(this.uia.GetRootElement())
		
		phone_app := this.GetChild(root, this.name, "WinUIDesktopWin32WindowClass")
		if (!phone_app)
		{
			phone_app := this.GetChild(root, this.name, "ApplicationFrameWindow")
			if (!phone_app)
			{
				this.err := this.name "이 실행중이지 않습니다"
				return false
			}
		}
		this.tree := "`r`n"
		this.tree .= this.name " TREE 시작`r`n"
		this.tree .= "========================`r`n"
		this.AddChildInfo(phone_app.pi, 1)
		this.tree .= "========================`r`n"
		this.tree .= this.name " TREE 끝`r`n"
		return this.tree
	}
	Init()
	{
		if (!this.uia.pi)
		{
			this.err := "오류 : UI Automation 프레임워크를 사용할수 없습니다"
			return false
		}
		root := new CIUIAutomationElement(this.uia.GetRootElement())
		
		
/*		
		DesktopChildSiteBridge, UIA_PaneControlTypeId
		cond_app_name := this.uia.CreatePropertyCondition(CCUIAutomation.UIA_NamePropertyId, "DesktopChildSiteBridge")
		cond_app_controltype := this.uia.CreatePropertyCondition(CCUIAutomation.UIA_ControlTypePropertyId, CCUIAutomation.UIA_PaneControlTypeId, 3)
		cond_app := this.uia.CreateAndCondition(cond_app_name, cond_app_controltype)
*/		
		cond_winui := this.uia.CreatePropertyCondition(CCUIAutomation.UIA_ClassNamePropertyId, "WinUIDesktopWin32WindowClass")
		results := new CIUIAutomationElementArray(root.FindAll(CCUIAutomation.TreeScope_Children, cond_winui))
		Loop, % results.GetLength()
		{
			if (this.FindMessageElement(results.GetElement(A_Index - 1)))
			{
				return true
			}
		}
/*
		하위호환을 위한 예전방식
*/
		phone_app := this.GetChild(root, this.name, "ApplicationFrameWindow")
		if (phone_app)
		{
			if (this.FindMessageElement(phone_app.pi))
			{
				return true
			}
		}
		this.err := this.name "이 실행중이지 않거나 찾을 수 없거나 메시지가 열려있지 않습니다"
		return false
	}
	FindMessageElement(pi)
	{
		app := new CIUIAutomationElement(pi)
		cond_messages_name := this.uia.CreatePropertyCondition(CCUIAutomation.UIA_NamePropertyId, "메시지")
		cond_messages_controltype := this.uia.CreatePropertyCondition(CCUIAutomation.UIA_ControlTypePropertyId, CCUIAutomation.UIA_ListControlTypeId, 3)
		cond_messages := this.uia.CreateAndCondition(cond_messages_name, cond_messages_controltype)
		results := new CIUIAutomationElementArray(app.FindAll(CCUIAutomation.TreeScope_Subtree, cond_messages))
		if (results.GetLength() = 0)
		{
			this.err := this.name "의 메시지 Elements를 찾을수 없습니다"
			return false
		}
		else if (results.GetLength() > 1)
		{
			this.err := this.name "의 메세지 Elements가 2개 이상 존재합니다`r`n"
			Loop, % results.GetLength()
			{
				child := new CIUIAutomationElement(results.GetElement(A_Index -1))
				this.err .= A_Index " : " child.CurrentName() " : " child.ClassName() "`r`n"
			}
			return false
		}
		this.elm_messages := new CIUIAutomationElement(results.GetElement(0))
		this.elm_messages_scroll := new CIUIAutomationScrollPattern(this.elm_messages.GetCurrentPattern(CCUIAutomation.UIA_ScrollPatternId))
		return true
	}
	Test(bLast := false, nWait := 200)
	{
	}
	GetMessages(bLast := false, nWait := 200)
	{
		if (bLast)
		{
			this.elm_messages_scroll.ScrollVerticalBottom()
		}
		else
		{
			this.elm_messages_scroll.ScrollVerticalTop()
		}
		
		this.msglist := Array()
		cond_automation_id_messagebody := this.uia.CreatePropertyCondition(CCUIAutomation.UIA_AutomationIdPropertyId, "MessageBody")
		while(1)
		{
			Sleep, %nWait%
			children := new CIUIAutomationElementArray(this.elm_messages.FindAll(CCUIAutomation.TreeScope_Children, this.cond_true))
			Loop, % children.GetLength()
			{
				elm_msg := new CIUIAutomationElement(children.GetElement(A_Index -1))
				msg := elm_msg.CurrentName()
				if (StrLen(msg) >=2000 && SubStr(msg, -4) != "의 메시지")
				{
					; 문자메시지 + 휴대폰과 연결앱이 덧붙인 정보가 2047를 넘어갈수 없다.
					; UIAutomation 의 property 는 4KB 의 길이제한이 있어 Name에서 읽으면 문자메시지내용뿐만 아니라 휴대폰과의 연결앱이 덧붙이는 수신시간정보까지 잘린다
					; 4KB초과하면 메시지 리스트뷰의 하위 message body의 텍스트 컨트롤을 textrange 인터페이스를 이용해 읽으면 길이제한없이 가져올수 있다.
					results := new CIUIAutomationElementArray(elm_msg.FindAll(CCUIAutomation.TreeScope_Subtree, cond_automation_id_messagebody))
					if (results.GetLength() = 1)
					{
						elm_msg_body := new CIUIAutomationElement(results.GetElement(0))
						text_pattern := new CIUIAutomationTextPattern(elm_msg_body.GetCurrentPattern(CCUIAutomation.UIA_TextPatternId))
						document := new CIUIAutomationTextRange(text_pattern.get_DocumentRange())
						msg := document.GetText(-1)
					}
				}
				bFound := false
				Loop, % this.msglist.MaxIndex()
				{
					if (this.msglist[A_Index] = msg)
					{
						bFound := true
						break
					}
				}
				if (!bFound)
				{
					this.msglist.Push(msg)
				}
			}
			if (!this.elm_messages_scroll.ScrollVerticalNext(4))
			{
				break
			}
		}
		return this.msglist.MaxIndex()
	}
	GetLastMessage(ByRef msg, ByRef prefix)
	{
		this.SkipPinCheck(true)
		this.GetMessages(true, 100)
		this.ParseMessages()
		if (this.messages.MaxIndex())
		{
			o := this.messages[this.messages.MaxIndex()]
			;LogMessage(o)
			msg := o.msg
			prefix := o.prefix
			return true
		}
		msg := ""
		prefix := ""
		this.err := "메세지가 하나도 없습니다"
		return false
	}	
	Copy()
	{
		this.GetMessages(false, 500)
		msg := ""
		Loop, % this.msglist.MaxIndex()
		{
			msg .= StrReplace(this.msglist[A_Index], "`n", "`r`n") "`r`n"
		}
		clipboard := msg
		return true
	}
	ParseMessages()
	{
		static wdays := {"일요일" : 1 , "월요일" : 2, "화요일" : 3, "수요일" : 4, "목요일" : 5, "금요일" : 6, "토요일" : 7, "오늘" : 8, "어제" : 9}
		Loop, % this.msglist.MaxIndex()
		{
			msg := StrReplace(StrReplace(this.msglist[A_Index], ", 메시지 내용을 탐색하려면 Tab 키를 누르세요."), "의 메시지의 메시지", "의 메시지")
			;LogMessage(msg)
			if (RegExMatch(msg, "O)\. +([^\.]+)\.?의 메시지", matches))
			{
				prefix := StrReplace(matches[1], chr(8206))
				msg := StrReplace(msg, matches[0])
				if (RegExMatch(msg, "O)^([^\.]+)\.", matches))
				{
					msg := StrReplace(msg, matches[0])
					prefix := matches[1] " : " prefix
					this.SetCurrentPage(matches[1])
				}
				this.AddMessage(msg, prefix)
			}
			else if (SubStr(msg, 1, 8) = "대화 구분 기호")
			{
				msg := StrReplace(StrReplace(SubStr(msg, 10), chr(8206)), " ")
				wd := wdays[msg]
				day_start := ""
				if (wd)
				{
					if (wd >= 8)
					{
						diff_day := 8 - wd
					}
					else
					{
						diff_day := wd - A_WDay
					}
					if (diff_day > 0)
					{
						diff_day -= 7
					}
					day_start := A_Year A_MM A_DD "000000"
					day_start += diff_day, days
				}
				else if (RegExMatch(msg, "O)(\d+)월(\d+),(\d+)", matches))
				{
					y := matches[3]
					m := matches[1] < 10 ? "0" matches[1] : matches[1]
					d := matches[1] < 10 ? "0" matches[2] : matches[2]
					
					day_start := y m d "000000"
					
				}
				if (day_start)
				{
					FormatTime, last_prefix, %day_start%, yyyy-MM-dd tt h:mm
					this.last := last_prefix
				}
			}
			else
			{
				if (StrLen(msg) >= 2000)
				{
					alarm_msg = "***********************************************************************************`n"
					alarm_msg .= "*`n"
					alarm_msg .= "* [상품권자동입력기 알림]`n"
					alarm_msg .= "*`n"
					alarm_msg .= "* 이 메시지는 휴대폰과 연결앱으로부터 수신시간정보를 가져올수 없어서 (4KB초과, 길이 약 2000여 자 초과시) `n"
					alarm_msg .= "*`n"
					alarm_msg .= "* 수신시간 정보는 직전 마지막 메세지의 수신시간 혹은 마지막 대화 구분 기호의 날짜를 기준으로 가져왔습니다`n" 
					alarm_msg .= "*`n"
					alarm_msg .= "***********************************************************************************`n"
					this.nTruncatedMessage++
					this.AddMessage(alarm_msg . msg)
				}
				else
				{

				}
			}
		}
	}
	Fetch()
	{
		;nWait := (this.elm_messages_scroll.CurrentVerticalViewSize() < 10) ? 1000 : 500
		this.GetMessages(false, 1000)
		;LogMessage(this.msglist)
		this.ParseMessages()
		return true
	}
	GetTime(prefix)
	{
		;[+82 1588-4763 : ‎2020‎-‎03‎-‎04 ‎오후‎ ‎9‎:‎50]
		if (RegExMatch(prefix, "O)([\d\-]+) +(오전|오후) +([^ ]+)$", matches))
		{
			time := RegExReplace(matches[3], "[^\d]")
			if (Instr(matches[2], "오후"))
			{
				time += 1200
			}
			time := RegExReplace(matches[1], "[^\d]") . Format("{1:04d}", time) . "00"
			return time
		}
		else
		{
		}
		return A_Now
	}
	Refresh()
	{
		win := new CWindow(this.name)
		if (win.hWnd)
		{
			win.Activate()
			Send, {F5}
		}
	}
}
class CChromeMessageReader extends CMessageReader
{
	__New(name, url_base)
	{
		base.__New(name)
		this.win := _app.GetChrome(url_base)
	}
	Init()
	{
		if (this.win.Init())
		{
			this.win.Activate()
			this.win.Move(0, 0)
			this.err := ""
			return true
		}
		this.err := this.win.err
		return false
	}
}
class CGoogleMessageReader extends CChromeMessageReader
{
	__New()
	{
		base.__New("구글 메시지", "https://messages.google.com/web")
	}
	Init()
	{
		if (!base.Init())
		{
			return false
		}
		result := this.win.wb.Evaluate("escape(document.getElementsByClassName('title-container')[0].innerText)")
		if (!result.value)
		{
			this.err := this.name "에서 문자를 선택해 열어주세요"
			return false
		}
		this.SetCurrentPage(this.win.Unescape(result.value))
		return true
	}
	Fetch()
	{
		result := this.win.wb.Evaluate("document.getElementsByTagName('mws-tombstone-message-wrapper')[0].parentElement.children.length")
		if (result.value < 1)
		{
			this.err := "구성요소를 찾을수 없습니다. 문자가 열려있는지 확인해주세요"
			return false
		}
		prefix := ""
		Loop, % result.value
		{
			root_i := A_Index - 1
			result := this.win.wb.Evaluate("document.getElementsByTagName('mws-tombstone-message-wrapper')[0].parentElement.children[" root_i "].tagName")
			if (result.value = "MWS-TOMBSTONE-MESSAGE-WRAPPER")
			{
				result := this.win.wb.Evaluate("escape(document.getElementsByTagName('mws-tombstone-message-wrapper')[0].parentElement.children[" root_i "].innerText)")
				prefix := this.win.Unescape(result.value)
			}
			else if (result.value = "MWS-MESSAGE-WRAPPER")
			{
				result := this.win.wb.Evaluate("document.getElementsByTagName('mws-tombstone-message-wrapper')[0].parentElement.children[" root_i "].getElementsByClassName('text-msg-content').length")
				if (result.value)
				{
					Loop, % result.value
					{
						result := this.win.wb.Evaluate("escape(document.getElementsByTagName('mws-tombstone-message-wrapper')[0].parentElement.children[" root_i "].getElementsByClassName('text-msg-content')[" A_Index - 1 "].innerText)")
						this.AddMessage(this.win.Unescape(result.value), prefix)
					}
				}
			}
		}
		return true
	}	
	GetTime(prefix)
	{
		static wdays := {"일요일" : 1 , "월요일" : 2, "화요일" : 3, "수요일" : 4, "목요일" : 5, "금요일" : 6, "토요일" : 7}
		if (RegExMatch(prefix, "O)(\d+년)?\s*(\d+월)?\s*(\d+일)?\s*([^ ]+요일)?.+(오전|오후) ([^ ]+)",matches))
		{
			year := matches[1] ? RegExReplace(matches[1], "[^\d]") : A_Year
			month := matches[2] ? Format("{1:02d}", RegExReplace(matches[2], "[^\d]")) : A_MM
			day := matches[3] ? Format("{1:02d}", RegExReplace(matches[3], "[^\d]")) : A_DD
			wday := matches[4] ? wdays[matches[4]] : A_Wday
			time := RegExReplace(matches[6], "[^\d]")
			if (matches[5] = "오후")
			{
				time += 1200
			}
			date := year month day Format("{1:04d}", time) "00"
			if (!matches[3])
			{
				day_before := wday - A_Wday
				if (day_before > 0)
				{
					day_before -= 7
				}
				date += %day_before%, D
			}
			return date
		}
		return A_Now
	}	
	GetLastMessage(ByRef msg, ByRef prefix)
	{
		msg := ""
		prefix := ""
		result := this.win.wb.Evaluate("document.getElementsByTagName('mws-relative-timestamp').length")
		if (result.value > 0)
		{
			result := this.win.wb.Evaluate("escape(document.getElementsByTagName('mws-relative-timestamp')[" (result.value - 1) "].innerText)")
			prefix := this.win.Unescape(result.value)
		}
		result := this.win.wb.Evaluate("document.getElementsByClassName('text-msg').length")
		if (result.value > 0)
		{
			result := this.win.wb.Evaluate("escape(document.getElementsByClassName('text-msg')[" (result.value - 1) "].innerText)")
			msg := this.win.Unescape(result.value)
			return true
		}
		return false
	}
}
class CPushbulletMessageReader extends CChromeMessageReader
{
	__New()
	{
		base.__New("Pushbullet", "https://www.pushbullet.com")
	}
	Init()
	{
		if (!base.Init())
		{
			return false
		}
		result := this.win.wb.Evaluate("document.getElementById('innerlist').children.length")
		if (result.value < 1)
		{
			this.err := this.name "에서 문자를 선택해 열어주세요"
			return false
		}
		return true
	}
	Fetch()
	{
		result := this.win.wb.Evaluate("document.getElementById('innerlist').children.length")
		prefix := ""
		if (result.value > 0)
		{
			length := result.value
			Loop, %length%
			{
				result := this.win.wb.Evaluate("document.getElementById('innerlist').children[" (A_Index - 1) "].children.length")
				; 맨마지막엔 메세지 수신시간이 아니라 마지막수신이 얼마전이었는지 표시됨
				if (A_Index != length && result.value = 2)
				{
					result := this.win.wb.Evaluate("escape(document.getElementById('innerlist').children[" (A_Index - 1) "].children[0].innerText)")
					prefix := this.win.Unescape(result.value)
					result := this.win.wb.Evaluate("escape(document.getElementById('innerlist').children[" (A_Index - 1) "].children[1].innerText)")
					msg := this.win.Unescape(result.value)
				}
				else
				{
					result := this.win.wb.Evaluate("escape(document.getElementById('innerlist').children[" (A_Index - 1) "].innerText)")
					msg := this.win.Unescape(result.value)
				}
				this.AddMessage(msg, prefix)
			}
		}
		return true
	}
	GetTime(prefix)
	{
		static wdays := {"Sunday" : 1 , "Monday" : 2, "Tuesday" : 3, "Wednesday" : 4, "Thursday" : 5, "Friday" : 6, "Saturday" : 7, "Today" : A_Wday, "Yesterday" : A_Wday -1 }
		static today := A_Year A_MM A_DD
		;  02/07/2020 
		; [Last Monday at 9:21 PM]
		if (RegExMatch(prefix, "O)(\d{2})/(\d{2})/(\d{4})", matches))
		{
			return matches[3] matches[1] matches[2] "000000"
		}
		else if (RegExMatch(prefix, "O)([^ ]+) at ([^ ]+) ([AP]M)", matches))
		{
			time := RegExReplace(matches[2], "[^\d]")
			if (matches[3] = "PM")
			{
				time += 1200
			}
			date := today Format("{1:04d}", time)
			wday := wdays[matches[1]]
			day_before := wday - A_Wday
			if (day_before > 0)
			{
				day_before -= 7
			}
			date += %day_before%, D
			return date
		}
		return A_Now
	}
	GetLastMessage(ByRef msg, ByRef prefix)
	{
		result := this.win.wb.Evaluate("document.getElementById('innerlist').children.length")
		if (result.value > 0)
		{
			index := result.value - 1
			result := this.win.wb.Evaluate("escape(document.getElementById('innerlist').children[" index "].children[0].innerText)")
			prefix := this.win.Unescape(result.value)
			result := this.win.wb.Evaluate("escape(document.getElementById('innerlist').children[" index "].children[1].innerText)")
			msg := this.win.Unescape(result.value)
			return true
		}
		return false
	}
}
class CGMarketMessageReader extends CChromeMessageReader
{
	__New()
	{
		base.__New("G마켓 선물함", "https://gift.gmarket.co.kr/m/mygift/rcv")
	}
	Init()
	{
		if (!base.Init())
		{
			this.err := this.win.err
			return false
		}
		if (!this.IsLogged())
		{
			this.err := "로그인 하세요."
			return false
		}
		if (!this.IsGiftBox())
		{
			this.err := "받은 선물함으로 이동해 가져오기할 선물을 마지막 단계(핀번호가 보이는 곳)까지 열어주세요"
			return false
		}
		return true
	}
	IsLogged()
	{
		return !InStr(this.win.GetLocation(), "login")
	}
	IsGiftBox()
	{
		result := this.win.wb.Evaluate("document.getElementsByClassName('box__component box__component-e-coupon').length")
		return result.value
	}
	Fetch()
	{
		result := this.win.wb.Evaluate("document.getElementsByClassName('swiper-slide').length")
		Loop, % result.value
		{
			result := this.win.wb.Evaluate("document.getElementsByClassName('swiper-slide')[" A_Index - 1 "].innerText")
			msg := RegExReplace(StrReplace(StrReplace(result.value, "(스마일캐시 직접충전 가능)"), "복사"), "(\d) (\d)", "$1-$2")
			this.AddMessage(msg, A_Now)
		}
		return true
	}
	GetTime(prefix)
	{
		return prefix
	}	
}
class CGifticonMessageReader extends CChromeMessageReader
{
	__New()
	{
		base.__New("기프티콘 선물함", "https://www.gifticon.com")
	}
	Init()
	{
		if (!base.Init())
		{
			this.err := this.win.err
			return false
		}
		if (!this.IsLogged())
		{
			this.win.Navigate("/util/util_login.do")
			this.err := "로그인 해주세요"
			return false
		}
		if (!this.GotoGiftBox())
		{
			return false
		}
		return true
	}
	IsLogged()
	{
		result := this.win.wb.Evaluate("document.getElementsByClassName('topsearch')[0].innerText") 
		return InStr(result.value, "로그아웃")
	}
	GotoGiftBox()
	{
		this.err := ""
		charging_url := "/mygifticon/present.do?kind=recive"
		result := this.win.wb.Evaluate("document.location.href")
		if (!InStr(result.value, charging_url) && !this.win.GotoUrl(charging_url))
		{
			this.err := "받은 선물함이동에 실패했습니다"
			return false
		}
		this.win.Activate()
		Sleep, 1000
		return true
	}
	Fetch()
	{
		script =
(
	var t = document.getElementsByClassName('tablestyle2')[0].children[0]; 
	var codes = ''; 
	for (var i= 1; i < t.rows.length; i++)
	{ 
		if (t.rows[i].cells.length < 7)
		{
			continue;
		} 
		t.rows[i].cells[4].children[0].click();
		codes += t.rows[i].cells[1].innerText + '\t';
		codes += t.rows[i].cells[7].innerText + '\t';
		codes += document.getElementById('reProdAcepInfo').innerText + ' 기프티콘\t';
		codes += document.getElementById('reProdCpnoInfo').innerText + '\t';
		codes += '\r\n';
	} 
	reProdInfoPopupClose();
	escape(codes);
)
		result := this.win.wb.Evaluate(script) 
		messages := this.win.Unescape(result.value)
		Loop, parse, messages, `n, `r
		{
			msginfo := StrSplit(A_LoopField, "`t")
			if (msginfo[2] = "교환대기")
			{
				msg := "`r`n" msginfo[3] "`t" msginfo[4] "`r`n"
				this.AddMessage(msg, StrReplace(msginfo[1], ".") "000000000")
			}
		}
		if (!this.messages.MaxIndex())
		{
			this.err := "`r`n`r`n교환대기중인 기프티콘이 없습니다 `r`n`r`n"
			this.err .= "날짜`t`t상태`t상품명`t`t`t번호`r`n" messages "`r`n"
			return false
		}
		return true
	}
	GetTime(prefix)
	{
		return prefix
	}
}
class CKakaoTalkMessageReader extends CMessageReader
{
	__New()
	{
		base.__New("카카오톡")
		this.win := ""
	}
	Init()
	{
		WinGet, windows, List
		Loop, %windows%
		{
			win := new CWindow(windows%A_Index%)
			if (win.GetClass() = "#32770" && win.GetProcessName() = "KakaoTalk.exe")
			{
				win.Activate()
				this.SetCurrentPage(win.GetTitle()) 
				this.win := new CWindow(win.ControlGet("EVA_VH_ListControl_Dblclk1"))
				if (this.win.Exist())
				{
					return true
				}
				this.err .= "카카오톡 대화창이 아닙니다"
				return false
			}
		}
		this.err := this.name " 대화창을 찾을수 없습니다"
		return false
	}
	Fetch()
	{
		; 전체선택
		this.win.SendMessage(0x07E9, 0x65, 0)
		; 복사
		this.win.SendMessage(0x07E9, 0x64, 0)
		txt := clipboard
		
		if (!RegExMatch(txt, "\r\n"))
		{
			txt := StrReplace(txt, "`n", "`r`n")
		}
		arr_daily := Array()
		arr_daily_date := Array()
		pos := 1
		while (pos := RegExMatch(txt, "Om)^(\d{4}년.*요일)$", matches, pos))
		{
			arr_daily.Push(pos)
			arr_daily_date.Push(matches[1])
			pos += matches.Len(0)
		}
		Loop, % arr_daily.MaxIndex()
		{
			daily_date := arr_daily_date[A_Index]
			pos_start := arr_daily[A_Index] + StrLen(daily_date) + 2
			if (A_Index = arr_daily.MaxIndex())
			{
				daily_msg := SubStr(txt, pos_start)
			}
			else
			{
				daily_msg := Substr(txt, pos_start, arr_daily[A_Index +1] - pos_start)
			}
			arr_msg := Array()
			arr_msg_prefix := Array()
			arr_msg_datetime := Array()
			pos := 1
			while (pos := RegExMatch(daily_msg, "Om)^\[[^\]]+\]\s+\[(오. [\d\:꞉]+)\] ", matches, pos))
			{
				arr_msg.Push(pos)
				arr_msg_prefix.Push(matches[0])
				arr_msg_datetime.Push(matches[1])
				pos += matches.Len(0)
			}
			Loop, % arr_msg.MaxIndex()
			{
				msg_date := daily_date " " arr_msg_datetime[A_Index]
				pos_start := arr_msg[A_Index] + StrLen(arr_msg_prefix[A_Index])
				if (A_Index = arr_msg.MaxIndex())
				{
					msg := SubStr(daily_msg, pos_start)
				}
				else
				{
					msg := Substr(daily_msg, pos_start, arr_msg[A_Index +1] - pos_start - 2)
				}
				this.AddMessage(msg, msg_date)
			}
		}
		return true
	}
	GetTime(prefix)
	{
		;[2020년 3월 4일 수요일 오전 10:28]
		if (RegExMatch(prefix, "O)(\d{4})년 (\d+)월 (\d+)일(.+)", matches))
		{
			time := RegExReplace(matches[4], "[^\d]")
			if (InStr(matches[4], "오후"))
			{
				time += 1200
			}
			return matches[1] . Format("{1:02d}", matches[2]) . Format("{1:02d}", matches[3]) .  Format("{1:04d}", time) . "00"
		}
		return A_Now
	}
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; [WBGet function for AHK v1.1]
; WBGet function - AutoHotkey Community
; https://autohotkey.com/boards/viewtopic.php?f=6&t=39869
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
WBGet(WinTitle="ahk_class IEFrame", Svr#=1) {               ;// based on ComObjQuery docs
   static msg := DllCall("RegisterWindowMessage", "str", "WM_HTML_GETOBJECT")
        , IID := "{0002DF05-0000-0000-C000-000000000046}"   ;// IID_IWebBrowserApp
;//     , IID := "{332C4427-26CB-11D0-B483-00C04FD90119}"   ;// IID_IHTMLWindow2
   SendMessage msg, 0, 0, Internet Explorer_Server%Svr#%, %WinTitle%
   if (ErrorLevel != "FAIL") {
      lResult:=ErrorLevel, VarSetCapacity(GUID,16,0)
      if DllCall("ole32\CLSIDFromString", "wstr","{332C4425-26CB-11D0-B483-00C04FD90119}", "ptr",&GUID) >= 0 {
         DllCall("oleacc\ObjectFromLresult", "ptr",lResult, "ptr",&GUID, "ptr",0, "ptr*",pdoc :=0)
         return ComObj(9,ComObjQuery(pdoc,IID,IID),1), ObjRelease(pdoc)
      }
   }
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; JSON lib for AutoHotkey v2.1.3 [updated 04/18/2016 (MM/DD/YYYY)]
; 배포사이트 : https://github.com/cocobelgica/AutoHotkey-JSON
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
class JSON
{
class Load extends JSON.Functor
{
Call(self, ByRef text, reviver:="")
{
this.rev := IsObject(reviver) ? reviver : false
this.keys := this.rev ? {} : false
static quot := Chr(34), bashq := "\" . quot
, json_value := quot . "{[01234567890-tfn"
, json_value_or_array_closing := quot . "{[]01234567890-tfn"
, object_key_or_object_closing := quot . "}"
key := ""
is_key := false
root := {}
stack := [root]
next := json_value
pos := 0
while ((ch := SubStr(text, ++pos, 1)) != "") {
if InStr(" `t`r`n", ch)
continue
if !InStr(next, ch, 1)
this.ParseError(next, text, pos)
holder := stack[1]
is_array := holder.IsArray
if InStr(",:", ch) {
next := (is_key := !is_array && ch == ",") ? quot : json_value
} else if InStr("}]", ch) {
ObjRemoveAt(stack, 1)
next := stack[1]==root ? "" : stack[1].IsArray ? ",]" : ",}"
} else {
if InStr("{[", ch) {
static json_array := Func("Array").IsBuiltIn || ![].IsArray ? {IsArray: true} : 0
(ch == "{")
? ( is_key := true
, value := {}
, next := object_key_or_object_closing )
: ( value := json_array ? new json_array : []
, next := json_value_or_array_closing )
ObjInsertAt(stack, 1, value)
if (this.keys)
this.keys[value] := []
} else {
if (ch == quot) {
i := pos
while (i := InStr(text, quot,, i+1)) {
value := StrReplace(SubStr(text, pos+1, i-pos-1), "\\", "\u005c")
static tail := A_AhkVersion<"2" ? 0 : -1
if (SubStr(value, tail) != "\")
break
}if (!i)
this.ParseError("'", text, pos)
  value := StrReplace(value,  "\/",  "/")
, value := StrReplace(value, bashq, quot)
, value := StrReplace(value,  "\b", "`b")
, value := StrReplace(value,  "\f", "`f")
, value := StrReplace(value,  "\n", "`n")
, value := StrReplace(value,  "\r", "`r")
, value := StrReplace(value,  "", "`t")
pos := i
i := 0
while (i := InStr(value, "\",, i+1)) {
if !(SubStr(value, i+1, 1) == "u")
this.ParseError("\", text, pos - StrLen(SubStr(value, i+1)))
uffff := Abs("0x" . SubStr(value, i+2, 4))
if (A_IsUnicode || uffff < 0x100)
value := SubStr(value, 1, i-1) . Chr(uffff) . SubStr(value, i+6)
}if (is_key) {
key := value, next := ":"
continue
}} else {
value := SubStr(text, pos, i := RegExMatch(text, "[\]\},\s]|$",, pos)-pos)
static number := "number", integer :="integer"
if value is %number%
{
if value is %integer%
value += 0
}else if (value == "true" || value == "false")
value := %value% + 0
else if (value == "null")
value := ""
else
this.ParseError(next, text, pos, i)
pos += i-1
}next := holder==root ? "" : is_array ? ",]" : ",}"
}is_array? key := ObjPush(holder, value) : holder[key] := value
if (this.keys && this.keys.HasKey(holder))
this.keys[holder].Push(key)
}}
return this.rev ? this.Walk(root, "") : root[""]
}ParseError(expect, ByRef text, pos, len:=1)
{
static quot := Chr(34), qurly := quot . "}"
line := StrSplit(SubStr(text, 1, pos), "`n", "`r").Length()
col := pos - InStr(text, "`n",, -(StrLen(text)-pos+1))
msg := Format("{1}`n`nLine:`t{2}`nCol:`t{3}`nChar:`t{4}"
,     (expect == "")     ? "Extra data"
: (expect == "'")    ? "Unterminated string starting at"
: (expect == "\")    ? "Invalid \escape"
: (expect == ":")    ? "Expecting ':' delimiter"
: (expect == quot)   ? "Expecting object key enclosed in double quotes"
: (expect == qurly)  ? "Expecting object key enclosed in double quotes or object closing '}'"
: (expect == ",}")   ? "Expecting ',' delimiter or object closing '}'"
: (expect == ",]")   ? "Expecting ',' delimiter or array closing ']'"
: InStr(expect, "]") ? "Expecting JSON value or array closing ']'"
:                      "Expecting JSON value(string, number, true, false, null, object or array)"
, line, col, pos)
static offset := A_AhkVersion<"2" ? -3 : -4
throw Exception(msg, offset, SubStr(text, pos, len))
}Walk(holder, key)
{
value := holder[key]
if IsObject(value) {
for i, k in this.keys[value] {
v := this.Walk(value, k)
if (v != JSON.Undefined)
value[k] := v
else
ObjDelete(value, k)
}}
return this.rev.Call(holder, key, value)
}}
class Dump extends JSON.Functor
{
Call(self, value, replacer:="", space:="")
{
this.rep := IsObject(replacer) ? replacer : ""
this.gap := ""
if (space) {
static integer := "integer"
if space is %integer%
Loop, % ((n := Abs(space))>10 ? 10 : n)
this.gap .= " "
else
this.gap := SubStr(space, 1, 10)
this.indent := "`n"
}return this.Str({"": value}, "")
}Str(holder, key)
{
value := holder[key]
if (this.rep)
value := this.rep.Call(holder, key, ObjHasKey(holder, key) ? value : JSON.Undefined)
if IsObject(value) {
static type := A_AhkVersion<"2" ? "" : Func("Type")
if (type ? type.Call(value) == "Object" : ObjGetCapacity(value) != "") {
if (this.gap) {
stepback := this.indent
this.indent .= this.gap
}is_array := value.IsArray
if (!is_array) {
for i in value
is_array := i == A_Index
until !is_array
}str := ""
if (is_array) {
Loop, % value.Length() {
if (this.gap)
str .= this.indent
v := this.Str(value, A_Index)
str .= (v != "") ? v . "," : "null,"
}} else {
colon := this.gap ? ": " : ":"
for k in value {
v := this.Str(value, k)
if (v != "") {
if (this.gap)
str .= this.indent
str .= this.Quote(k) . colon . v . ","
}}
}if (str != "") {
str := RTrim(str, ",")
if (this.gap)
str .= stepback
}if (this.gap)
this.indent := stepback
return is_array ? "[" . str . "]" : "{" . str . "}"
}} else
return ObjGetCapacity([value], 1)=="" ? value : this.Quote(value)
}Quote(string)
{
static quot := Chr(34), bashq := "\" . quot
if (string != "") {
  string := StrReplace(string,  "\",  "\\")
, string := StrReplace(string, quot, bashq)
, string := StrReplace(string, "`b",  "\b")
, string := StrReplace(string, "`f",  "\f")
, string := StrReplace(string, "`n",  "\n")
, string := StrReplace(string, "`r",  "\r")
, string := StrReplace(string, "`t",  "")
static rx_escapable := A_AhkVersion<"2" ? "O)[^\x20-\x7e]" : "[^\x20-\x7e]"
while RegExMatch(string, rx_escapable, m)
string := StrReplace(string, m.Value, Format("\u{1:04x}", Ord(m.Value)))
}return quot . string . quot
}}
Undefined[]
{
get {
static empty := {}, vt_empty := ComObject(0, &empty, 1)
return vt_empty
}}
class Functor
{
__Call(method, ByRef arg, args*)
{
if IsObject(method)
return (new this).Call(method, arg, args*)
else if (method == "")
return (new this).Call(arg, args*)
}}
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CmdRun(command, bIncludeStdErr := true)
{
	cmd := "cmd /c """ command """" (bIncludeStdErr ? " 2>&1" : "") " | clip"
	try
	{
		RunWait, %cmd%, , Hide
	}
	catch e
	{
		return e.Message
	}
	return clipboard
}
